
module DM::Bevel

  class BevelData
    include GeometryMixin

    class CalculationState
      NONE = 0
      PARTIAL = 1
      FULL = 2

      EDGE_OFFSET_TOO_SMALL = 3
    end

    class TaubinParams
      IDEAL_ITERATIONS = 30
      IDEAL_LAMBDA = 0.6
      IDEAL_MU_RATIO = 0.955 # A percentage of the value of IDEAL_LAMBDA
    end

    # @return [Array<Sketchup::Edge>]
    def primary_edges
      @primary_edges.clone
    end

    # @return [Hash{Sketchup::Vertex => Corner}]
    def affected_corners
      @affected_corners
    end

    # @return [BevelLimiter]
    # def limiter
    #   @limiter
    # end

    # Describes the transition along a edge.
    # @return [Hash{EdgeVertex => TransitionSegment}]
    def edge_vert_to_transition_segment
      @edge_vert_to_transition_segment
    end

    # If there is a patch (requires at least three primary edges at corner) then this data describes that patch.
    # @return [Hash{Corner => CornerPatch}]
    def corner_to_patch
      @corner_to_patch
    end

    # @return [BevelParams]
    def params
      @params
    end

    # @return [Boolean]
    def full_calculation?
      @calculation_state == CalculationState::FULL
    end

    # @return [Boolean]
    def partial_calculation?
      @calculation_state == CalculationState::PARTIAL
    end

    # @return [Boolean]
    def none_calculated?
      @calculation_state == CalculationState::NONE
    end

    # @return [Boolean]
    def edge_offset_too_small?
      @calculation_state == CalculationState::EDGE_OFFSET_TOO_SMALL
    end

    def initialize
      # @type [Array<Sketchup::Edge>]
      @primary_edges = []
      # @type [BevelParams]
      @params = nil
      # # @type [BevelLimiter]
      # @limiter = BevelLimiter.new(self)
      # @type [CacheManager]
      @cache_manager = CacheManager.new

      self.clear
    end

    def clear
      # @type [CalculationState]
      @calculation_state = CalculationState::NONE
      # @type [Hash{Sketchup::Vertex => Corner}]
      @affected_corners = nil
      # @type [Hash{EdgeVertex => TransitionSegment}]
      @edge_vert_to_transition_segment = nil
      # @type [Hash{Corner => CornerPatch}]
      @corner_to_patch = nil

      # @limiter.clear
    end

    # @return [Array<Edge>]
    def all_connected_edges
      return [] if @primary_edges.empty?

      val = @cache_manager.retrieve(:all_connected_edges) do
        primary_edges.first.all_connected.grep(Sketchup::Edge)
      end
      val
    end

    # @param edge_or_edges [Sketchup::Edge, Array<Sketchup::Edge>]
    def add_edges(edge_or_edges)
      edge_or_edges = [edge_or_edges] unless edge_or_edges.is_a?(Enumerable)
      edge_or_edges.uniq.each do |edge|
        raise UnexpectedTypeError, "Value passed is not an Sketchup::Edge" unless edge.instance_of?(Sketchup::Edge)
        next if @primary_edges.include?(edge)
        raise UnexpectedStateError, "All edges in a BevelData instance must be connected (not necessarily touching)" if !@primary_edges.empty? && !self.all_connected_edges.include?(edge)

        @primary_edges << edge
      end
      self.clear
    end

    # @param edge_or_edges [Sketchup::Edge, Array<Sketchup::Edge>]
    def remove_edges(edge_or_edges)
      edge_or_edges = [edge_or_edges] unless edge_or_edges.is_a?(Enumerable)
      edge_or_edges.uniq.each do |edge|
        raise UnexpectedTypeError, "Value passed is not a Sketchup::Edge" unless edge.instance_of?(Sketchup::Edge)
        @primary_edges.delete(edge)
      end
      @cache_manager.clear if @primary_edges.empty?
      self.clear
    end

    # @param params [BevelParams]
    def params=(params)
      raise UnexpectedTypeError, "Must be an object of type BevelParams" unless params.instance_of?(BevelParams)

      @params = params
      self.clear
    end

    # @param partial_calculation [Boolean] Set to true to only calculate the edge offsets (no patches calculated).
    # partial_calculation was defaulted to false
    def calculate(partial_calculation)
      rCalculate(partial_calculation)
    rescue EdgeOffsetTooSmallError
      self.clear
      @calculation_state = CalculationState::EDGE_OFFSET_TOO_SMALL
      raise
    rescue BevelError
      raise # Let these errors pass through
    rescue StandardError => error
      # Wrap every other error as a BevelDebugError
      raise BevelDebugError, error.message, error.backtrace
    end

    # @return [Array<Sketchup::Face>]
    def affected_faces
      faces = []
      self.affected_corners.each_pair do |_, corner|
        this_corner_requires_patching = corner.requires_patching?
        corner.face_corners.each do |face_corner|
          next if face_corner.instance_of?(FakeFaceCorner)

          # if this face_corner is all_secondary, and the adjacent ones are too, then this isn't an affected face.
          if !this_corner_requires_patching && face_corner.all_secondary?
            num_all_secondary = 0
            face_corner.adjacent_face_corners.each do |adjacent_face_corner|
              num_all_secondary += 1 if adjacent_face_corner.all_secondary?
            end
            next if num_all_secondary == 2
          end

          if face_corner.respond_to?(:inner_face_corners)
            face_corner.inner_face_corners.each do |inner_face_corner|
              next if inner_face_corner.instance_of?(FakeFaceCorner)

              faces << inner_face_corner.face
            end
          else
            faces << face_corner.face
          end
        end
      end
      faces.uniq
    end

    # @param edge_vert [EdgeVertex]
    # @return [TransitionSegment]
    def transition_at_edge_vert(edge_vert)
      @edge_vert_to_transition_segment[edge_vert]
    end

    # @param face_corner [FaceCorner]
    # @return [Array<TransitionSegment>]
    def transitions_at_face_corner(face_corner)
      transition_segments = []
      transition_seg = self.transition_at_edge_vert(face_corner.edge_vert1)
      transition_segments << transition_seg unless transition_seg.nil?
      transition_seg = self.transition_at_edge_vert(face_corner.edge_vert2)
      transition_segments << transition_seg unless transition_seg.nil?
      transition_segments
    end

    # @param face_corner [FaceCorner]
    # @return [TransitionSegment]
    def get_three_point_transition_at_face_corner(face_corner)
      return nil unless face_corner.three_point_corner?

      self.transitions_at_corner(face_corner.corner).each do |transition_segment|
        next unless transition_segment.three_point_segment?

        return transition_segment if transition_segment.face_corner_offset1.face_corner == face_corner
      end
      nil
    end

    # @param corner [Corner]
    # @return [Array<TransitionSegment>]
    def transitions_at_corner(corner)
      transition_segments = []
      patch_data = @corner_to_patch[corner]
      if !patch_data.nil?
        transition_segments = patch_data.transition_loop.transition_segments
      else
        corner.edge_verts.each do |edge_vert|
          transition_segment = @edge_vert_to_transition_segment[edge_vert]
          transition_segments << transition_segment unless transition_segment.nil?
        end
      end
      transition_segments
    end

    # Note: This function will not return any Three Point Transition Segments.
    # @yieldparam transition_segment [TransitionSegment]
    # @return [Array<TransitionSegment>]
    def each_transition
      transition_segments = []
      @edge_vert_to_transition_segment.each_pair do |_, transition_segment|
        yield(transition_segment) if block_given?
        transition_segments << transition_segment
      end
      transition_segments
    end

    # @param corner [Corner]
    # @return [Hash{EdgeVertex => Array<Geom::Point3d>}]
    def get_edge_vert_to_transition_pts(corner)
      transition_segments = self.transitions_at_corner(corner)
      edge_vert_to_new_pts = Hash.new { |h,k| h[k] = [] }
      transition_segments.each do |transition_segment|
        curve = transition_segment.curve
        corner.edge_verts.each do |edge_vert|
          [curve.start_pt, curve.end_pt].each do |pt|
            # Check if it's on the line vector
            next unless self.point_on_line?(edge_vert.start_pt, edge_vert.end_pt, pt)

            pts_array = edge_vert_to_new_pts[edge_vert].concat([pt])
            edge_vert_to_new_pts[edge_vert] = unique_array(pts_array)
          end
        end
      end
      edge_vert_to_new_pts
    end

    # @param corner [Corner]
    # @return [Array<BezierCurve, BezierCurve>, Nil]
    def get_v_curves(corner)
      raise UnexpectedStateError.new("corner does not require a v-patch", corner) unless corner.v_corner?

      transition_segments = self.transitions_at_corner(corner)
      curve1 = transition_segments[0].curve
      curve2 = transition_segments[1].curve

      if curve1.start_pt == curve2.end_pt
        curve2 = curve2.reverse
      elsif curve1.end_pt == curve2.start_pt
        curve1 = curve1.reverse
      end
      raise UnexpectedStateError.new("v-patch must have the same start_pt", corner, [curve1, curve2]) if curve1[0] != curve2[0] # TODO: Logic check
      [curve1, curve2]
    end

    # @param corner [Corner]
    # @return [Boolean]
    def transition_segments_overlap?(corner)
      transition_segments = self.transitions_at_corner(corner)
      return false unless transition_segments.length == 2

      curve1 = transition_segments[0].curve
      curve2 = transition_segments[1].curve
      return false unless curve1.length == curve2.length

      # Check if the transition segments happen to overlap each other.
      unique_array(curve1.points + curve2.points).length == curve1.length
    end

    # Case where we are beveling one edge and the beveled edge goes 'into' the object.
    # @param corner [Corner]
    # @return [Boolean]
    def is_one_transition_curved_into_self?(corner)
      return false unless corner.mix_curve?

      self.one_transition_edge_verts(corner).length == 1
    end

    # If there's one transition at the corner, then return the edge_verts that the start/end pts of the transition fall on.
    # If the transition is 'curving into itself', there will be one transition. Otherwise, there should be two.
    # @param corner [Corner]
    # @return [Array<EdgeVertex>]
    def one_transition_edge_verts(corner)
      return [] unless corner.num_primary_edges == 1

      transition_segments = self.transitions_at_corner(corner)
      return [] unless transition_segments.length == 1

      curve = transition_segments.first.curve
      edge_vert = corner.primary_edge_verts.first
      adjacent_edge_verts = []
      corner.get_adjacent_edge_verts(edge_vert).each do |adjacent_edge_vert|
        [curve.start_pt, curve.end_pt].each do |pt|
          next unless pt.on_line?(adjacent_edge_vert.edge_line)

          vec = adjacent_edge_vert.vertex.position.vector_to(pt).normalize
          adjacent_edge_verts << adjacent_edge_vert if vec.samedirection?(adjacent_edge_vert.edge_vector)
        end
      end
      # If only one of the adjacent_edge_vert has a point on it, then this case is true.
      adjacent_edge_verts
    end

    private

    # @param primary_edges [Array<Sketchup::Edge>]
    # @return [Hash{Sketchup::Vertex => Corner}]
    def initialize_geometry_data(primary_edges)
      # @type [Hash{Sketchup::Vertex => Corner}]
      affected_corners = {}

      # Calculate a 'weight' for each affected face to determine which one is 'higher priority'
      # Used only in cases where edge.faces.length > 2
      face_to_weight = Hash.new {|h,k| h[k] = 0}
      unless all_connected_edges.find {|edge| edge.faces.length > 2 }.nil?
        primary_edges.each do |edge|
          [edge.start, edge.end].each do |vert|
            vert.faces.each do |face|
              (vert.edges & face.edges).each do |f_edge|
                # If prioritize faces that don't have edges with more than 3 faces
                face_to_weight[face] += 1 unless f_edge.faces.length > 2
                # Primary edges also take priority.
                face_to_weight[face] += 1 if primary_edges.include?(f_edge)
              end
            end
          end
        end
      end

      # Corner's to bevel with related edges.
      primary_edges.each do |edge|
        [edge.start, edge.end].each do |vert|
          next unless affected_corners[vert].nil?

          edge_contains_multiple_faces = false
          vert.edges.each do |v_edge|
            next unless v_edge.faces.length > 2

            edge_contains_multiple_faces = true
            break
          end

          if edge_contains_multiple_faces
            affected_corners[vert] = FragmentedCorner.new(vert, primary_edges, face_to_weight: face_to_weight)
          else
            affected_corners[vert] = Corner.new(vert, primary_edges)
          end
        end
      end

      affected_corners.each_value do |corner|
        # @type [Corner]
        corner = corner
        corner.get_all_edge_verts.each  do |edge_vert|
          opposite_vertex = edge_vert.opposite_vertex
          # @type [Corner]
          next_corner = affected_corners[opposite_vertex]
          next if next_corner.nil?

          # Add next corner reference
          corner.add_corner(next_corner)

          # Set the opposite edge_verts
          opposite_edge_vert = next_corner.get_all_edge_verts.find { |next_edge_vert| next_edge_vert.edge == edge_vert.edge }
          edge_vert.set_opposite_edge_vert(opposite_edge_vert)
          opposite_edge_vert.set_opposite_edge_vert(edge_vert)
        end
      end

      # Map the FaceCorner to next FaceCorners.
      affected_corners.each_value do |corner|
        # @type [Corner]
        corner = corner
        corner.face_corners.each do |face_corner|
          next_face_corners = [nil, nil]

          corner.next_corners.each do |next_corner|
            next_corner.face_corners.each do |next_face_corner|
              if face_corner.next?(next_face_corner)
                if face_corner.edge_vert1.opposite?(next_face_corner.edge_vert1) || face_corner.edge_vert1.opposite?(next_face_corner.edge_vert2)
                  next_face_corners[0] = next_face_corner
                elsif face_corner.edge_vert2.opposite?(next_face_corner.edge_vert1) || face_corner.edge_vert2.opposite?(next_face_corner.edge_vert2)
                  next_face_corners[1] = next_face_corner
                end
              end
            end
          end
          face_corner.set_next_face_corners(next_face_corners[0], next_face_corners[1])
        end
      end

      affected_corners
    end

    # @param affected_corners [Hash{Sketchup::Vertex => Corner}]
    # @param edge_offset [Float]
    # @return [Hash {FaceCorner => TransitionSegment}]
    def calculate_transition_data(affected_corners, edge_offset = @params.edge_offset)
      # @type [Hash {FaceCorner => Hash{EdgeVertex => Geom::Point3d}}]
      face_corner_to_edge_vert_to_offset_pt = Hash.new {|h,k| h[k] = {}}

      # Calculate the ideal offset pts for each face_corner
      affected_corners.each_pair do |vert, corner|
        corner.face_corners.each do |face_corner|
          edge_vert_to_offset_pt = {}
          face = face_corner.face
          case face_corner.state
          when FaceCorner::State::ALL_PRIMARY
            edge_vec = face_corner.edge_vert1.edge_vector
            other_edge_vec = face_corner.edge_vert2.edge_vector
            if !face_corner.three_point_corner?
              if face_corner.instance_of?(CoplanarFaceCorner) && corner.edge_verts.length < 3
                closest_edge_vert = face_corner.closest_coplanar_edge_vert(face_corner.edge_vert1)
                mid_vec = closest_edge_vert.edge_vector
              else
                mid_vec = face_corner.mid_vector
              end
              offset_vec =  face.normal * edge_vec
              offset_dist = (edge_offset / Math.cos(offset_vec.angle_between(mid_vec))).abs
              offset_pt = vert.position.offset(mid_vec, offset_dist)

              raise UnexpectedStateError.new('', face_corner, face_corner.edge_vert1) unless edge_vert_to_offset_pt[face_corner.edge_vert1].nil?
              raise UnexpectedStateError.new('', face_corner, face_corner.edge_vert2) unless edge_vert_to_offset_pt[face_corner.edge_vert2].nil?

              edge_vert_to_offset_pt[face_corner.edge_vert1] = offset_pt
              edge_vert_to_offset_pt[face_corner.edge_vert2] = offset_pt
            else
              # Three point case
              edge_offset_vec = face.normal * edge_vec
              edge_offset_vec.reverse! if edge_offset_vec.angle_between(other_edge_vec) < Math::PI / 2.0 # Reverse if it's not pointing towards the face.
              edge_offset_vec.length = edge_offset

              other_edge_offset_vec = face.normal * other_edge_vec
              other_edge_offset_vec.reverse! if other_edge_offset_vec.angle_between(edge_vec) < Math::PI / 2.0
              other_edge_offset_vec.length = edge_offset

              edge_pt = vert.position.offset(edge_offset_vec).offset(edge_vec, edge_offset)
              other_edge_pt = vert.position.offset(other_edge_offset_vec).offset(other_edge_vec, edge_offset)

              # Track how the edge_vert is tied to the offset_pts
              raise UnexpectedStateError.new('', face_corner, face_corner.edge_vert1) unless edge_vert_to_offset_pt[face_corner.edge_vert1].nil?
              raise UnexpectedStateError.new('', face_corner, face_corner.edge_vert2) unless edge_vert_to_offset_pt[face_corner.edge_vert2].nil?

              edge_vert_to_offset_pt[face_corner.edge_vert1] = edge_pt
              edge_vert_to_offset_pt[face_corner.edge_vert2] = other_edge_pt
            end
          when FaceCorner::State::PRIMARY_SECONDARY
            primary_secondary = face_corner.primary_secondary
            primary_edge_vec = primary_secondary[0].edge_vector
            secondary_edge_vec = primary_secondary[1].edge_vector

            if face_corner.instance_of?(ParallelFaceCorner)
              offset_pt = vert.position.offset(face_corner.mid_vector, edge_offset)
            else
              offset_vec =  face.normal * primary_edge_vec
              # If we have three or more edges (that are not coplanar), then ignore the coplanar edge
              if face_corner.instance_of?(CoplanarFaceCorner) && corner.edge_verts.length < 3
                closest_edge_vert = face_corner.closest_coplanar_edge_vert(primary_secondary[0])
                closest_edge_vert_vec = closest_edge_vert.edge_vector

                offset_dist = (edge_offset / Math.cos(offset_vec.angle_between(closest_edge_vert_vec))).abs
                offset_pt = vert.position.offset(closest_edge_vert_vec, offset_dist)
              elsif face_corner.instance_of?(QuadFaceCorner) && corner.edge_verts.length < 3
                quad_edge_vert_vec = face_corner.quad_diagonal_edge_vert.edge_vector
                offset_dist = (edge_offset / Math.cos(offset_vec.angle_between(quad_edge_vert_vec))).abs
                offset_pt = vert.position.offset(quad_edge_vert_vec, offset_dist)
              else
                offset_dist = (edge_offset / Math.cos(offset_vec.angle_between(secondary_edge_vec))).abs

                if !face_corner.face_between?
                  offset_pt = vert.position.offset(secondary_edge_vec.reverse, offset_dist)
                else
                  # Only apply the limit if it's going in the same direction as the secondary edge
                  offset_dist = primary_secondary[1].edge_length if offset_dist > primary_secondary[1].edge_length
                  offset_pt = vert.position.offset(secondary_edge_vec, offset_dist)
                end
              end
            end
            raise UnexpectedStateError.new('', face_corner, primary_secondary[1]) unless edge_vert_to_offset_pt[primary_secondary[1]].nil? # TODO: logic check

            edge_vert_to_offset_pt[primary_secondary[0]] = offset_pt
            edge_vert_to_offset_pt[primary_secondary[1]] = offset_pt
          else
            # ALL_SECONDARY
            edge_vert_to_offset_pt[face_corner.edge_vert1] = face_corner.vertex.position.offset(face_corner.edge_vert1.edge_vector, edge_offset)
            edge_vert_to_offset_pt[face_corner.edge_vert2] = face_corner.vertex.position.offset(face_corner.edge_vert2.edge_vector, edge_offset)
          end
          face_corner_to_edge_vert_to_offset_pt[face_corner] = edge_vert_to_offset_pt
        end
      end

      # This is to track if there are face_corners that share the same offset_pt. We will need to adjust them later.
      affected_face_corners_to_offset_pt = {}
      # For all PRIMARY_SECONDARY cases, check if there is a shared secondary edge and if the offset pts on that edge is different.
      # @type [Array<FaceCorner>]
      affected_corners.each_pair do |vert, corner|
        next if corner.all_primary?

        # @type [Hash{EdgeVertex => Array<Geom::Point3d>}]
        secondary_edge_vert_to_offset_pts = Hash.new {|h,k| h[k] = []}
        corner.face_corners.each do |face_corner|
          next unless face_corner.state == FaceCorner::State::PRIMARY_SECONDARY

          secondary_edge_vert = face_corner.primary_secondary.last
          secondary_edge_vert_to_offset_pts[secondary_edge_vert] << face_corner_to_edge_vert_to_offset_pt[face_corner][secondary_edge_vert]
        end

        secondary_edge_vert_to_offset_pts.each_pair do |secondary_edge_vert, offset_pts|
          # @type [EdgeVertex]
          secondary_edge_vert = secondary_edge_vert
          next unless offset_pts.length > 1

          raise UnexpectedStateError.new('', corner, secondary_edge_vert) if offset_pts.length != 2 # TODO Logic check. Don't think it'll ever be more than 2

          if offset_pts.first == offset_pts.last
            # Make sure the secondary edges share the same reference to the point
            # A bit dedundant with below, but for the sake of readability, leaving it as is
            secondary_edge_vert.face_corners.each do |face_corner|
              edge_verts = face_corner.primary_secondary

              raise UnexpectedStateError.new('', face_corner, secondary_edge_vert) if edge_verts[1] != secondary_edge_vert # Todo: Logic check

              face_corner_to_edge_vert_to_offset_pt[face_corner][edge_verts[0]] = offset_pts.first
              face_corner_to_edge_vert_to_offset_pt[face_corner][edge_verts[1]] = offset_pts.first
            end
          else
            vec_1 = vert.position.vector_to(offset_pts.first)
            vec_2 = vert.position.vector_to(offset_pts.last)
            next unless vec_1.samedirection?(vec_2) # If not in the same direction, than this is the MIX/transition case and they should be ignroed.

            # Pick an average between the two offset pts.
            average_pt = self.linear_combination(offset_pts.first, offset_pts.last)
            secondary_edge_vert.face_corners.each do |face_corner|
              # Since it's PRIMARY_SECONDARY, need to update the primary so they all hold the same reference to the Point3d object.
              edge_verts = face_corner.primary_secondary

              raise UnexpectedStateError.new('', face_corner, secondary_edge_vert) if edge_verts[1] != secondary_edge_vert # Todo: Logic check

              face_corner_to_edge_vert_to_offset_pt[face_corner][edge_verts[0]] = average_pt
              face_corner_to_edge_vert_to_offset_pt[face_corner][edge_verts[1]] = average_pt

              affected_face_corners_to_offset_pt[face_corner] = average_pt
            end
          end
        end
      end

      # Attempt to ensure the offset pts are parallel from corner to corner.
      # As of now, if there is a valid case where a offset is needed, then it will only do it once (rather than looping continuiously.)
      i = 0
      affected_face_corners = affected_face_corners_to_offset_pt.keys
      affected_face_corners.each do |face_corner|
        # @type [FaceCorner]
        face_corner = face_corner
        offset_pt = affected_face_corners_to_offset_pt[face_corner]

        i += 1
        raise UnexpectedStateError.new('', face_corner) if i > 10000 # TODO: Infinite loop check. Should be impossible as we're not continuously looping anymore.

        face_corner.next_face_corners.each do |next_face_corner|
          next if next_face_corner.nil?
          next if next_face_corner.instance_of?(FakeFaceCorner)
          # If one if a ParallelFaceCorner and the other is not, then ignore.
          next if face_corner.instance_of?(ParallelFaceCorner) != next_face_corner.instance_of?(ParallelFaceCorner)

          shared_edge_vert = face_corner.get_shared_edge_vert(next_face_corner)
          other_offset_pt = affected_face_corners_to_offset_pt[next_face_corner]

          if other_offset_pt.nil?
            other_offset_pt = face_corner_to_edge_vert_to_offset_pt[next_face_corner][shared_edge_vert.opposite_edge_vert]
            next if offset_pt.vector_to(other_offset_pt).parallel?(shared_edge_vert.edge_vector)

            if next_face_corner.all_primary?
              # Adjust the point along the other_edge_vert so that we don't break the 'parallelness' of it's neighbouring face_corner.
              next_other_edge_vert = next_face_corner.get_other_edge_vert(shared_edge_vert.opposite_edge_vert)
              if next_face_corner.edge_verts_parallel?
                next_line = [next_face_corner.vertex.position, next_face_corner.mid_vector]
              else
                next_line = [other_offset_pt, next_other_edge_vert.edge_vector]
              end
              new_pt = self.intersect_line_line([offset_pt, shared_edge_vert.edge_vector], next_line)
              next if new_pt.nil?

              other_offset_pt.set!(new_pt)

              # If it's a parallel joint, then we need to consider that as part of our loop.
              if next_face_corner.corner.is_parallel_joint?
                next if affected_face_corners.include?(next_face_corner)

                affected_face_corners << next_face_corner
                affected_face_corners_to_offset_pt[next_face_corner] = new_pt
              end
            elsif next_face_corner.primary_secondary?
              # Adjust the point on the secondary edge_vert.
              next_primary_secondary = next_face_corner.primary_secondary
              if next_face_corner.edge_verts_parallel?
                next_line = [next_face_corner.vertex.position, next_face_corner.mid_vector]
                next_edge_vert = nil
              else
                next_line = next_primary_secondary[1].edge_line
                next_edge_vert = next_primary_secondary[1]
              end
              new_pt = self.intersect_line_line([offset_pt, shared_edge_vert.edge_vector], next_line)
              next if new_pt.nil?
              next if new_pt == next_face_corner.vertex.position # Skip if it's on the vertex.

              unless next_edge_vert.nil?
                # IF the new offset happens to go in the opposite direction of hte original offset, then don't use it.
                orig_pt_vec = next_edge_vert.vertex.position.vector_to(other_offset_pt)
                new_pt_vec = next_edge_vert.vertex.position.vector_to(new_pt)
                next unless orig_pt_vec.samedirection?(new_pt_vec)

                # We know its going in the same direction as the edge. Now make sure new_pt doesn't go past the opposite_vertex
                vert_pt = next_face_corner.vertex.position
                next_vert_dist = vert_pt.distance(next_edge_vert.opposite_vertex.position)
                new_pt_dist = vert_pt.distance(new_pt)
                next if new_pt_dist > next_vert_dist
              end


              other_offset_pt.set!(new_pt)

              # If the adjacent face_corner hasn't already been added to the list, then add it.
              adjacent_face_corner = next_primary_secondary[1].get_other_face_corner(next_face_corner)
              next if affected_face_corners.include?(adjacent_face_corner)
              next if next_face_corner.corner.primary_edge_verts.length == 1

              affected_face_corners << adjacent_face_corner
              affected_face_corners_to_offset_pt[adjacent_face_corner] = new_pt
            else
              # All Secondary. Do nothing.
              next
            end
          else

            # Case where both are marked as affected face_corners
            # next if offset_pt.vector_to(other_offset_pt).parallel?(shared_edge_vert.edge_vector)

            # raise UnexpectedStateError unless face_corner.primary_secondary? && next_face_corner.primary_secondary?

            # TODO: Mar 14, 2023: For now, let's ignore. Don't change the offset pts if both face_corner and next_face_corner have been marked as 'affected_face_corners'
            #       Monitor and see if things still look terrible.
            # new_pt = self.intersect_line_line([offset_pt, shared_edge_vert.edge_vector], next_face_corner.primary_secondary[1].edge_line)
            # DB.draw_pts(new_pt)
          end
        end
      end

      # Calculate the TransitionSegments now (the parts that join the FaceCorner with the adjacent FaceCorner)
      # @type [Hash {EdgeVertex => TransitionSegment}]
      edge_vert_to_transition_segment = {}
      affected_corners.each_pair do |vert, corner|
        corner.face_corners.length.times do |i|
          face_corner1 = corner.face_corners[i]
          face_corner2 = corner.face_corners[i + 1]
          face_corner2 = corner.face_corners[0] if face_corner2.nil?

          next if face_corner1.all_secondary?
          next if face_corner2.all_secondary?

          shared_edge_vert = face_corner1.get_shared_edge_vert(face_corner2)
          next unless shared_edge_vert.primary?

          other_edge_vert1 = face_corner1.get_other_edge_vert(shared_edge_vert)
          other_edge_vert2 = face_corner2.get_other_edge_vert(shared_edge_vert)

          if face_corner1.three_point_corner?
            offset_pt1 = face_corner_to_edge_vert_to_offset_pt[face_corner1][shared_edge_vert]
          else
            offset_pt1 = face_corner_to_edge_vert_to_offset_pt[face_corner1][other_edge_vert1]
          end

          if face_corner2.three_point_corner?
            offset_pt2 = face_corner_to_edge_vert_to_offset_pt[face_corner2][shared_edge_vert]
          else
            offset_pt2 = face_corner_to_edge_vert_to_offset_pt[face_corner2][other_edge_vert2]
          end

          raise EdgeOffsetTooSmallError if offset_pt1 == offset_pt2 # Raise so it's captured by self.calculate

          if corner.requires_patching?
            if face_corner1.all_primary? && face_corner2.all_primary?
              # Grab an average in this case to get a consistent result. Otherwise it would be based on which face_corner comes first in the array
              # Find the exact spot they intersect. If that fails, find the closest point they intersect.
              t1_pt = Geom.intersect_line_line([offset_pt1, other_edge_vert1.edge_vector], [vert.position, shared_edge_vert.edge_vector])
              t1_pt = Geom.closest_points([offset_pt1, other_edge_vert1.edge_vector], [vert.position, shared_edge_vert.edge_vector]).last if t1_pt.nil?
              t1_pt = vert.position unless self.point_on_line?(shared_edge_vert.start_pt, shared_edge_vert.end_pt, t1_pt)

              t2_pt = Geom.intersect_line_line([offset_pt2, other_edge_vert2.edge_vector], [vert.position, shared_edge_vert.edge_vector])
              t2_pt = Geom.closest_points([offset_pt2, other_edge_vert2.edge_vector], [vert.position, shared_edge_vert.edge_vector]).last if t2_pt.nil?
              t2_pt = vert.position unless self.point_on_line?(shared_edge_vert.start_pt, shared_edge_vert.end_pt, t2_pt)

              transition_pt = Geom.linear_combination(0.5, t1_pt, 0.5, t2_pt)
            elsif face_corner1.all_primary?
              if face_corner1.instance_of?(ParallelFaceCorner)
                transition_pt = vert.position
              elsif face_corner1.instance_of?(QuadFaceCorner)
                transition_pt = Geom.closest_points([offset_pt1, other_edge_vert1.edge_vector],[vert.position, shared_edge_vert.edge_vector]).last
              else
                transition_pt = Geom.intersect_line_line([offset_pt1, other_edge_vert1.edge_vector], [vert.position, shared_edge_vert.edge_vector])
              end
            elsif face_corner2.all_primary?
              if face_corner2.instance_of?(ParallelFaceCorner)
                transition_pt = vert.position
              elsif face_corner2.instance_of?(QuadFaceCorner)
                transition_pt = Geom.closest_points([offset_pt2, other_edge_vert2.edge_vector], [vert.position, shared_edge_vert.edge_vector]).last
              else
                transition_pt = Geom.intersect_line_line([offset_pt2, other_edge_vert2.edge_vector], [vert.position, shared_edge_vert.edge_vector])
              end
            else
              transition_pt = vert.position
            end
          else
            transition_pt = vert.position
          end

          face_corner_offset1 = FaceCornerOffsetData.new(face_corner1, offset_pt1)
          face_corner_offset2 = FaceCornerOffsetData.new(face_corner2, offset_pt2)
          bezier_curve = BezierCurve.new(offset_pt1, offset_pt2, transition_pt, @params.segments)

          transition_segment = TransitionSegment.new(face_corner_offset1, face_corner_offset2, bezier_curve)
          edge_vert_to_transition_segment[shared_edge_vert] = transition_segment
          # DB.draw_bezier_curve(bezier_curve, 10, 2, 'purple')
        end
      end

      # Handle the case where a ParallelFaceCorner (as a Primary_Secondary) is a part of the patch.
      # Need to add a PointCurve TransitionSegment to join them together.
      affected_corners.each_pair do |_, corner|
        next unless corner.requires_patching?

        parallel_face_corners = corner.face_corners.select { |face_corner| face_corner.instance_of?(ParallelFaceCorner)}
        parallel_face_corners.each do |parallel_face_corner|
          next unless parallel_face_corner.primary_secondary?

          # Find the adjacent other_face_corner
          edge_verts = parallel_face_corner.primary_secondary
          primary_edge_vert = edge_verts[0]
          secondary_edge_vert = edge_verts[1]
          other_face_corner = (secondary_edge_vert.face_corners - parallel_face_corners).first

          # Get the offset pts
          start_pt = face_corner_to_edge_vert_to_offset_pt[parallel_face_corner][secondary_edge_vert]
          end_pt = face_corner_to_edge_vert_to_offset_pt[other_face_corner][secondary_edge_vert]

          # Build a TransitionSegment that goes from the start end offset pts.
          new_pts = self.fill_points_between(start_pt, end_pt, @params.segments)

          # Need to maintain point order. (go from start => end , start => end)
          start_transition_segment = edge_vert_to_transition_segment[primary_edge_vert]
          new_pts.reverse! if start_transition_segment.curve.start_pt == new_pts.first

          point_curve = PointCurve.new(new_pts)
          start_face_offset = FaceCornerOffsetData.new(parallel_face_corner, start_pt)
          end_face_offset = FaceCornerOffsetData.new(other_face_corner, end_pt)
          new_transition_segment = TransitionSegment.new(start_face_offset, end_face_offset, point_curve)

          edge_vert_to_transition_segment[secondary_edge_vert] = new_transition_segment
        end
      end

      # In certain patch scenarios (seems to be when 5 or more edges at the corner) we need to fill in the missing spaces
      # that occurs if not all edges at the corner are Primary
      affected_corners.each_pair do |vert, corner|
        next unless corner.requires_patching?

        # Start a primary edge_vert. This way we can make sure we capture the 'segments' in order.
        start_index = corner.edge_verts.find_index {|edge_vert| edge_vert.primary? }
        corner_edge_verts = corner.edge_verts

        # @type [Array<Array<EdgeVertex>>]
        segmented_edge_verts = []
        cur_segment = []
        corner_edge_verts.length.times do |i|
          ev_index = start_index + i
          ev_index -= corner_edge_verts.length if ev_index >= corner_edge_verts.length

          edge_vert = corner_edge_verts[ev_index]
          face_corner1 = edge_vert.face_corners[0]
          face_corner2 = edge_vert.face_corners[1]
          if ([FaceCorner::State::ALL_SECONDARY] & [face_corner1.state, face_corner2.state]).empty?
            # No ALL_SECONDARY states
            unless cur_segment.empty?
              segmented_edge_verts << cur_segment
              cur_segment = []
            end
          else
            # One of the face_corners must be an ALL_SECONDARY state
            cur_segment << edge_vert
          end
        end
        # Add the last one in if it hasn't already been added.
        segmented_edge_verts << cur_segment unless cur_segment.empty?
        next if segmented_edge_verts.empty? # No segments, so don't continue.

        segmented_edge_verts.each do |segment_e_v|
          # @type [EdgeVertex]
          prev_edge_vert = nil
          prev_start_face_corner = nil
          (segment_e_v.length - 1).times do |ev_i|
            # @type [EdgeVertex]
            edge_vert_secondary = segment_e_v[ev_i]
            # @type [EdgeVertex]
            next_edge_vert_secondary = segment_e_v[ev_i + 1]

            if ev_i == 0
              start_edge_vert = (edge_vert_secondary.face_corners.flat_map{|face_corner| face_corner.edge_verts } - segment_e_v).first
              start_face_corner = edge_vert_secondary.get_shared_face_corner(start_edge_vert)
            else
              start_edge_vert = edge_vert_secondary
              start_face_corner = edge_vert_secondary.get_shared_face_corner(prev_edge_vert)
            end

            if ev_i == segment_e_v.length - 2
              # When at this case, we'll cheat and grabbed the face_corner with the end offset_pt
              end_edge_vert = (next_edge_vert_secondary.face_corners.flat_map{|face_corner| face_corner.edge_verts } - segment_e_v).first
              end_face_corner = end_edge_vert.get_shared_face_corner(next_edge_vert_secondary)
              # While above will cheat and grab the next face_corner, we still need to be adjacent when creating the TransitionSegment.
              # Basically, it's a hack.
              ts_end_face_corner = next_edge_vert_secondary.get_shared_face_corner(edge_vert_secondary)
            else
              end_edge_vert = next_edge_vert_secondary
              end_face_corner = next_edge_vert_secondary.get_shared_face_corner(edge_vert_secondary)
              ts_end_face_corner = end_face_corner
            end

            # Get the offset pts
            start_pt = face_corner_to_edge_vert_to_offset_pt[start_face_corner][start_edge_vert]
            end_pt = face_corner_to_edge_vert_to_offset_pt[end_face_corner][end_edge_vert]

            # Build a TransitionSegment that goes from the start end offset pts.
            new_pts = self.fill_points_between(start_pt, end_pt, @params.segments)

            # Need to maintain point order. (go from start => end , start => end)
            start_transition_segment = edge_vert_to_transition_segment[(ev_i == 0 ? start_edge_vert : prev_edge_vert)]
            new_pts.reverse! if start_transition_segment.curve.start_pt == new_pts.first

            point_curve = PointCurve.new(new_pts)
            start_face_offset = FaceCornerOffsetData.new(start_face_corner, start_pt)
            end_face_offset = FaceCornerOffsetData.new(ts_end_face_corner, end_pt)
            new_transition_segment = TransitionSegment.new(start_face_offset, end_face_offset, point_curve)

            edge_vert_to_transition_segment[edge_vert_secondary] = new_transition_segment
            prev_edge_vert = edge_vert_secondary
            # DB.draw_bezier_curve(point_curve, 10, 2, 'green')
          end
        end
      end

      # Need to connect the TransitionSegments together now.
      edge_vert_to_transition_segment.each_pair do |edge_vert, transition_segment|
        # @type [EdgeVertex]
        edge_vert = edge_vert
        # @type [TransitionSegment]
        transition_segment = transition_segment

        opposite_transition_segment = edge_vert_to_transition_segment[edge_vert.opposite_edge_vert]
        unless opposite_transition_segment.nil?
          transition_segment.set_opposite_transition_segment(opposite_transition_segment)
        end

        valid_edge_verts = (edge_vert.corner.edge_verts - [edge_vert])
        valid_edge_verts.each do |other_edge_vert|
          adjacent_transition_segment = edge_vert_to_transition_segment[other_edge_vert]
          next if adjacent_transition_segment.nil?
          next if transition_segment == adjacent_transition_segment
          next unless transition_segment.connected?(adjacent_transition_segment)

          transition_segment.add_adjacent_transition_segments(adjacent_transition_segment)
        end
      end

      # For the 3 pt face corner cases, create the 'transition_segment' on that face.
      affected_corners.each_pair do |vert, corner|
        next unless corner.mix_curve?

        corner.face_corners.each do |face_corner|
          next unless face_corner.three_point_corner?

          start_pt = face_corner_to_edge_vert_to_offset_pt[face_corner][face_corner.edge_vert1]
          face_corner_offset1 = FaceCornerOffsetData.new(face_corner, start_pt)
          end_pt = face_corner_to_edge_vert_to_offset_pt[face_corner][face_corner.edge_vert2]
          face_corner_offset2 = FaceCornerOffsetData.new(face_corner, end_pt)

          mid_pt = Geom.intersect_line_line([start_pt, face_corner.edge_vert1.edge_vector], [end_pt, face_corner.edge_vert2.edge_vector])
          bezier_curve = BezierCurve.new(start_pt, end_pt, mid_pt, @params.segments)
          transition_segment = TransitionSegment.new(face_corner_offset1, face_corner_offset2, bezier_curve)

          corner.edge_verts.each do |edge_vert|
            adjacent_transition_segment = edge_vert_to_transition_segment[edge_vert]
            next if adjacent_transition_segment.nil?
            next unless transition_segment.connected?(adjacent_transition_segment)

            transition_segment.add_adjacent_transition_segments(adjacent_transition_segment)
            adjacent_transition_segment.add_adjacent_transition_segments(transition_segment)
          end
        end
      end

      edge_vert_to_transition_segment
    end

    # @param affected_corners [Hash{Sketchup::Vertex => Corner}]
    # @param edge_vert_to_transition_segment [Hash{FaceCorner => Array<Geom::Point3d>}]
    def calculate_corner_data(affected_corners, edge_vert_to_transition_segment)
      # Calculate the theorectical bezier curves that form the patch.
      # @type [Hash{Corner => TransitionLoop}]
      corner_to_transition_loop = {}
      affected_corners.each_pair do |_, corner|
        next unless corner.requires_patching?
        next if corner.is_parallel_joint?

        cur_transition_segment = edge_vert_to_transition_segment[corner.primary_edge_verts.first]
        transition_segments = [cur_transition_segment]

        while true
          did_add = false
          cur_transition_segment.adjacent_transition_segments.each do |adjacent_transition_segment|
            next if transition_segments.include?(adjacent_transition_segment)

            transition_segments << adjacent_transition_segment
            cur_transition_segment = adjacent_transition_segment
            did_add = true
            break
          end

          break unless did_add
        end
        corner_to_transition_loop[corner] = TransitionLoop.new(corner, transition_segments)
      end

      # @type [Hash{Corner => CenterPatchData}]
      corner_to_center_patch = {}
      corner_to_transition_segment_to_point_curves = {}
      corner_to_transition_loop.each_pair do |corner, transition_loop|
        # @type [Corner]
        corner = corner
        # @type [TransitionLoop]
        transition_loop = transition_loop

        transition_segment_to_bezier_curves = Hash.new { |h,k| h[k] = [] }
        transition_loop.transition_segments.each_with_index do |transition_segment, i|
          # Calculates the control point curve for the adjacent transition segment
          across_transition_segment = transition_loop.get_transition_segment_across(transition_segment)

          start_pt = transition_segment.curve.control_pt
          if transition_loop.transition_segments.length.odd?
            end_pt = across_transition_segment.curve.end_pt
          else
            end_pt = across_transition_segment.curve.control_pt
          end

          if start_pt == end_pt
            # Occurs when we have a transition_segment that is based off of PrimarySecondary state, and across_transition_segment that is also based on that.
            # The control_pt for those cases is the vertex.position. So for that case, just create a PointCurve that contains only vertex.position
            pts = []
            (@params.segments + 1).times do |_|
              pts << corner.vertex.position
            end
            curve_ct = PointCurve.new(pts)
          else
            curve_ct = BezierCurve.new(start_pt, end_pt, corner.vertex.position, @params.segments)
          end

          # now create the theorectical curve points (need to do some data massaging)
          adjacent_transition_segment = transition_segment.adjacent_to_end
          next_over_transition_segment = transition_loop.get_next_transition_segment_over(adjacent_transition_segment, true)
          adjacent_inner_pts = adjacent_transition_segment.inner_points_from_start
          next_over_inner_pts = next_over_transition_segment.inner_points_from_end

          adjacent_inner_pts.length.times do |pt_i|
            pt1 = adjacent_inner_pts[pt_i]
            pt2 = next_over_inner_pts[pt_i]
            c_pt = curve_ct[pt_i + 1]

            b_c = BezierCurve.new(pt1, pt2, c_pt, @params.segments)
            transition_segment_to_bezier_curves[transition_segment] << b_c
            # DB.draw_bezier_curve(b_c, 5, 2, 'orange')
          end
        end

        # set's a minimum distance from the "center" (the theoretical center)
        # TODO: Make this adjustable?
        min_center_offset = 0.5 # it's a percentage

        # # Made up logic to find if the closest to middle curve overlaps with another curve
        overlapping_transition_segment_to_overlap_ratio = {}
        transition_segment_to_bezier_curves.each_pair do |transition_segment, bezier_curves|
          last_bezier_curve = bezier_curves.last
          t_s_pt = transition_segment.curve.mid_pt
          cur_b_c_pt = last_bezier_curve.mid_pt
          line = [t_s_pt, t_s_pt.vector_to(cur_b_c_pt)]

          if bezier_curves.length == 1
            prev_b_c = transition_segment.curve
          else
            prev_b_c = bezier_curves[bezier_curves.length - 2]
          end

          dist_between_segments = prev_b_c.mid_pt.distance(last_bezier_curve.mid_pt) # We'll use the smallest segment... for now.
          cur_b_c_pt = cur_b_c_pt.offset(line[1], dist_between_segments * min_center_offset) # a percent of the curve dist
          cur_b_c_length = t_s_pt.distance(cur_b_c_pt)
          transition_segment_to_bezier_curves.each_pair do |other_transition_segment, other_bezier_curves|
            next if transition_segment == other_transition_segment

            other_last_bezier_curve = other_bezier_curves.last
            other_t_s_pt = other_transition_segment.curve.mid_pt
            other_b_c_pt = other_last_bezier_curve.mid_pt
            other_line = [other_b_c_pt, other_t_s_pt.vector_to(other_b_c_pt)]

            closest_pts = Geom.closest_points(line, other_line)
            closest_on_line = closest_pts.first # We are comparing to the first line so only need that one.
            if cur_b_c_length != t_s_pt.distance(closest_on_line) - cur_b_c_pt.distance(closest_on_line)
              # Need to add the min_center_offset to ensure the center remains at that size.
              num_segments_overlapped_ratio = last_bezier_curve.mid_pt.distance(closest_on_line) / dist_between_segments
              overlapping_transition_segment_to_overlap_ratio[transition_segment] = min_center_offset + num_segments_overlapped_ratio
            end
          end
        end

        # Take an average for a more consistent result.
        total = overlapping_transition_segment_to_overlap_ratio.values.inject(0, :+).to_f # Note: array.sum does not exist in Ruby versions lower than 2.4
        num_overlaps = overlapping_transition_segment_to_overlap_ratio.keys.length.to_f
        overlapping_transition_segment_to_overlap_ratio.each_pair do |t_s, _|
          overlapping_transition_segment_to_overlap_ratio[t_s] = total / num_overlaps
        end

        overlap_bezier_curve_to_index_offset_vec = Hash.new { |h,k| h[k] = [] }
        overlapping_transition_segment_to_overlap_ratio.each_pair do |transition_segment, overlap_ratio|
          # @type [TransitionSegment]
          transition_segment = transition_segment

          # @type [Array<BezierCurve>]
          all_bezier_curves = [transition_segment.curve.reverse] + transition_segment_to_bezier_curves[transition_segment]

          # Get a curve that makes up the "offset direction"
          pt_index_to_offset_curve = {}
          all_bezier_curves.first.length.times do |pt_i|
            curve_path = []
            all_bezier_curves.each do |b_c|
              curve_path << b_c[pt_i]
            end
            pt_index_to_offset_curve[pt_i] = PointCurve.new(curve_path, true)
          end

          (all_bezier_curves.length - 1).times do |b_c_i|

            b_c1 = all_bezier_curves[b_c_i]
            b_c2 = all_bezier_curves[b_c_i + 1]
            (1..b_c1.length - 2).each do |index|
              half_index = b_c1.length / 2
              # mod = maintains 'curve' of each segment. (Quadratic equation... sort of.)
              if index > half_index
                mod = (b_c1.length - 1 - index).to_f / half_index.to_f
              else
                mod = index.to_f / half_index.to_f
              end

              # c_ratio = how far to push it towards the transition segment
              c_ratio = ((b_c_i + 2).to_f / (all_bezier_curves.length).to_f)
              offset_ratio = overlap_ratio * ((mod * c_ratio) ** 2.0)

              # We have to offset it along a curve, so that's what's going on below.
              # @type [PointCurve]
              point_curve = pt_index_to_offset_curve[index]
              
              offset_dist = point_curve.segment_length(b_c_i + 1) * offset_ratio
              pt_dist = point_curve.distance_at_index(b_c_i + 1) - offset_dist
              t = pt_dist / point_curve.curve_length

              # if t > 1.0 || t < 0.0
              #   # TODO: Leaving this for now to remind myself why something may look stupid.
              #   DB.draw_bezier_curve(point_curve, 10, 4, 'red')
              # end
              # TODO: For now, ignore if the below case is true. Occurs in really obscure cases.
              next if t > 1.0 || t < 0.0

              offset_pt = point_curve.point_at_t(t)

              vec = b_c2[index].vector_to(offset_pt)
              overlap_bezier_curve_to_index_offset_vec[b_c2] << [index, vec]
            end
          end

        end

        # This is the adjusted curve points. They are technically no longer bezier curves after we modify them.
        transition_segment_to_point_curves = Hash.new { |h,k| h[k] = [] }
        transition_segment_to_bezier_curves.each_pair do |transition_segment, b_cs|
          b_cs.each do |b_c|
            offset_dists = overlap_bezier_curve_to_index_offset_vec[b_c]

            pts = b_c.points
            offset_dists.each do |index_vec|
              pts[index_vec[0]].offset!(index_vec[1])
            end

            transition_segment_to_point_curves[transition_segment] << PointCurve.new(pts)
          end
        end

        ## End of thearectical, and beginning of practical curves.

        ### Data correct/joining of curves at intersection points.

        # Do data correction of the curves to find actual pts.
        transition_segments = transition_loop.transition_segments
        point_curve_to_intersect_results = Hash.new { |h,k| h[k] = [] }
        transition_segments.length.times do |i|
          transition_segment1 = transition_segments[i]
          transition_segment2 = transition_segments[i + 1]
          transition_segment2 = transition_segments[0] if transition_segment2.nil?

          point_curves1 = transition_segment_to_point_curves[transition_segment1]
          point_curves2 = transition_segment_to_point_curves[transition_segment2]

          point_curves1.each_with_index do |point_curve, b_c_i|
            mid_index = point_curve.length / 2

            (1..(mid_index - 1)).each do |index|
              other_point_curve = point_curves2[index - 1]
              # Number of points must be the same.
              raise UnexpectedStateError.new('', point_curve, other_point_curve) if point_curve.length != other_point_curve.length

              other_index = other_point_curve.length - (b_c_i + 2) # Get the opposite end of hte other index, skipping the end_pt
              pts = [
                [point_curve[index], point_curve[index - 1]],
                [point_curve[index], point_curve[index + 1]]
              ]
              other_pts = [
                [other_point_curve[other_index], other_point_curve[other_index - 1]],
                [other_point_curve[other_index], other_point_curve[other_index + 1]]
              ]

              all_closest_pts = []
              pts.each do |pt_data|
                pt1 = pt_data[0]
                pt2 = pt_data[1]
                line = [pt1, pt1.vector_to(pt2)]
                other_pts.each do |other_pt_data|
                  other_pt1 = other_pt_data[0]
                  other_pt2 = other_pt_data[1]
                  other_line = [other_pt1, other_pt1.vector_to(other_pt2)]

                  closest_pts = Geom.closest_points(line, other_line)
                  closest_pt = Geom.linear_combination(0.5, closest_pts.first, 0.5, closest_pts.last)
                  total_dist = pt1.distance(closest_pt) + pt2.distance(closest_pt)
                  total_dist += other_pt1.distance(closest_pt) + other_pt2.distance(closest_pt)
                  all_closest_pts << [closest_pt, total_dist]
                end
              end

              closest_pt = all_closest_pts.min {|a,b| a[1] <=> b[1] }.first
              point_curve_to_intersect_results[point_curve] << [index, closest_pt]
              point_curve_to_intersect_results[other_point_curve] << [other_index, closest_pt]
            end
          end
        end

        ts_to_point_curves = Hash.new {|h,k| h[k] = []}
        transition_segments.each_with_index do |transition_segment, ts_i|
          point_curves = transition_segment_to_point_curves[transition_segment]
          ts_to_point_curves[transition_segment] << transition_segment.curve.reverse.to_point_curve
          point_curves.each do |point_curve|
            pts = point_curve.points.clone
            point_curve_to_intersect_results[point_curve].each do |result|
              index = result[0]
              new_pt = result[1]
              pts[index] = new_pt
            end

            ts_to_point_curves[transition_segment] << PointCurve.new(pts)
          end

        end
        corner_to_transition_segment_to_point_curves[corner] = ts_to_point_curves


        ################################################
        ### Center pt logic

        # Figure out the center pt for patching.
        # get centroid of center patch.
        mid_pts = []
        center_patch_pts = []
        transition_segment_to_center_patch_pts = {}
        transition_loop.transition_segments.each do |transition_segment|
          all_point_curves = ts_to_point_curves[transition_segment]

          last_curve = all_point_curves.last
          half_i = last_curve.length / 2
          if @params.single_segment?
            center_pts = [last_curve.points.first, last_curve.points.last]
          else
            center_pts = [last_curve[half_i - 1], last_curve[half_i], last_curve[half_i + 1]]
          end
          center_pts.reverse!

          center_patch_pts.concat(center_pts)
          mid_pts << last_curve[half_i] if @params.segments == 2

          transition_segment_to_center_patch_pts[transition_segment] = center_pts
        end

        center_patch_pts = unique_array(center_patch_pts)

        if @params.segments == 2
          # Special case for 2 segments.
          # If we use the below logic, the center pt sinks in. No amount of smoothing will bring that out.
          center_pt = centroid(mid_pts + [corner.vertex.position])
        else
          centroid_pt = centroid(center_patch_pts)
          # It doesn't matter where the center pt is as it'll just be smoothed out.
          center_pt = centroid_pt
        end

        corner_to_center_patch[corner] = CenterPatchData.new(center_pt, transition_segment_to_center_patch_pts)

        ############################################
        # Smoothing logic (post center pt)

        # Get the neighbouring pts
        point_to_neighbour_pts = Hash.new { |h,k| h[k] = [] }
        ts_to_point_curves.each_pair do |transition_segment, all_point_curves|
          # @type [TransitionSegment]
          transition_segment = transition_segment
          # @type [Array<PointCurve>]
          all_point_curves = all_point_curves

          next_transition_segment = transition_segment.adjacent_to_end
          # @type [Array<PointCurve>]
          next_all_point_curves = ts_to_point_curves[next_transition_segment]

          (1..all_point_curves.length - 1).each do |curve_i|
            point_curve = all_point_curves[curve_i]
            # DB.draw_pts(point_curve.points[1..point_curve.length / 2], 20, 6, 'green')

            (1..point_curve.length / 2).each do |pt_i|
              pt = point_curve[pt_i]
              if curve_i == all_point_curves.length - 1
                # The curve nearest to the center
                if pt_i  == point_curve.length / 2
                  # The mid pt adjacent ot the center
                  next_pt = center_pt
                else
                  next_point_curve = next_all_point_curves[pt_i]
                  next_pt = next_point_curve[next_point_curve.length / 2]
                end

                neighbor_pts = [
                  point_curve[pt_i - 1], point_curve[pt_i + 1],
                  all_point_curves[curve_i - 1][pt_i],
                  next_pt
                ]
              else

                next_p_c = all_point_curves[curve_i + 1]
                prev_p_c = all_point_curves[curve_i - 1]
                neighbor_pts = [
                  point_curve[pt_i - 1], point_curve[pt_i + 1],
                  next_p_c[pt_i],
                  prev_p_c[pt_i],
                ]
              end
              point_to_neighbour_pts[pt] = neighbor_pts
            end
          end

          # Add neighbors for center_pt
          last_point_curve = all_point_curves.last
          half_i = last_point_curve.length / 2
          point_to_neighbour_pts[center_pt] << last_point_curve[half_i]
        end

        # Attempt to smooth the patch.
        taubin_smoothing!(point_to_neighbour_pts, TaubinParams::IDEAL_ITERATIONS, best_lambda, best_mu)
      end # End of corner to transition loop lo`op.

      # Reorganize the date so it doesn't hurt the brain
      # @type [Hash{Corner => CornerPatch}]
      corner_to_patch = {}
      corner_to_transition_loop.each_pair do |corner, transition_loop|
        transition_segment_to_point_curves = corner_to_transition_segment_to_point_curves[corner]
        center_patch = corner_to_center_patch[corner]
        corner_to_patch[corner] = CornerPatch.new(transition_loop, transition_segment_to_point_curves, center_patch)
      end
      corner_to_patch
    end

    # @return [Float]
    def best_lambda
      lambda = TaubinParams::IDEAL_LAMBDA
      num_segments = @params.segments
      return lambda if num_segments > 30

      lambda = lambda * (num_segments.to_f / 30.to_f)
      lambda
    end

    # @return [Float]
    def best_mu
      (best_lambda * TaubinParams::IDEAL_MU_RATIO) * -1.0
    end

    # @param point_to_neighbour_pts [Hash{Geom::Point3d => Array<Geom::Point3d>}]
    # @param iterations [Integer]
    # @param lambda [Float]
    # @param mu [Float]
    def taubin_smoothing!(point_to_neighbour_pts, iterations = 10, lambda = 0.33, mu = -0.34)
      iterations.times do |_|
        pt_to_offset_vec = {}

        point_to_neighbour_pts.each_pair do |pt, neighbour_pts|
          vec = basic_smoothing_vector(pt, neighbour_pts, lambda)
          pt_to_offset_vec[pt] = vec
        end

        pt_to_offset_vec.each_pair do |pt, offset_vec|
          pt.set!(pt + offset_vec)
        end

        pt_to_offset_vec = {}
        point_to_neighbour_pts.each_pair do |pt, neighbour_pts|
          vec = basic_smoothing_vector(pt, neighbour_pts, mu)
          pt_to_offset_vec[pt] = vec
        end

        pt_to_offset_vec.each_pair do |pt, offset_vec|
          pt.set!(pt + offset_vec)
        end
      end
    end

    # @param pt [Geom::Point3d]
    # @param neighbour_pts [Array<Geom::Point3d>]
    # @param mod [Flaot]
    # @return [Array<Float>] A Vector3d in array format (most efficient)
    def basic_smoothing_vector(pt, neighbour_pts, mod)
      offset_pt = [0.0, 0.0, 0.0]
      neighbour_pts.each do |neighbour_pt|
        arr = neighbour_pt.to_a
        offset_pt[0] += arr[0]
        offset_pt[1] += arr[1]
        offset_pt[2] += arr[2]
      end

      num_edges = neighbour_pts.length.to_f
      avg_pt = [offset_pt[0] / num_edges, offset_pt[1] / num_edges, offset_pt[2] / num_edges]
      pt_arr = pt.to_a
      [mod * (avg_pt[0] - pt_arr[0]), mod * (avg_pt[1] - pt_arr[1]), mod * (avg_pt[2] - pt_arr[2])]
    end

  end

end