module DM::Bevel

  # Methods here are used in both the Bevel Tool and the non-destructive bevel
  module BevelMixin

    # Returns edges that are able to be beveled based on our tool validation
    # @param edges [Array<Sketchup::Edge>]
    # @param out_valid_edges [Array] Valid edges inserted here.
    # @param is_proxy [Boolean] Set to true to follow proxy rules.
    # @return [BevelableEdgeState]
    def validate_edges(edges, out_valid_edges = [], is_proxy = false)
      edges = [edges] unless edges.is_a?(Array)
      active_ents = Sketchup.active_model.active_entities
      overall_edge_state = BevelableEdgeState::VALID
      edges.each do |edge|
        next unless edge.valid? # Edge must be valid
        unless is_proxy
          next unless edge.parent.entities == active_ents # Only select edges/faces that are at the active_entities level
        end

        cur_state = BevelableEdgeState::VALID
        cur_state |= BevelableEdgeState::NOT_TWO_FACES if edge.faces.length != 2 # Edge must make exactly 2 faces to be an option for bevel

        # Don't go further unless the above are true
        unless cur_state == BevelableEdgeState::VALID
          overall_edge_state |= cur_state
          next
        end

        cur_state |= BevelableEdgeState::COPLANAR if self.coplanar_edge?(edge) # Only allow coplanar edges if enabled
        cur_state |= BevelableEdgeState::REVERSED_FACES if self.reversed_faces?(edge) # Dont allow edges that contain reversed faces

        unless is_proxy
          cur_state |= BevelableEdgeState::HIDDEN if edge.hidden? # Only allow hidden edges if enabled
        end
        cur_state |= BevelableEdgeState::QUAD_EDGE if QuadFace.is_quadface_diagonal?(edge)  # Soft, smooth, no shadows are quad edges -- these are not selectable

        # Only allow soft / smooth selection if enabled
        cur_state |= BevelableEdgeState::SOFT if edge.soft?
        # cur_state |= BevelableEdgeState::SMOOTH if edge.smooth?

        unless cur_state == BevelableEdgeState::VALID
          overall_edge_state |= cur_state
          next
        end

        out_valid_edges << edge
      end

      overall_edge_state
    end

    # Group the edges together based on if they are connected.
    # @param edges [Array<Sketchup::Edge>]
    # @return [Array<Array<Sketchup::Edge>>]
    def get_grouped_edges(edges)
      grouped_edges = []
      matched_edges = []
      edges.each do |edge|
        next if matched_edges.include?(edge)

        related_edges = edges & edge.all_connected
        matched_edges.concat(related_edges)
        grouped_edges << related_edges
      end
      grouped_edges
    end

    # @param picked_edges [Array<Sketchup::Edge>]
    def add_edges_to_selected(picked_edges, do_calc = true)
      get_grouped_edges(picked_edges).each do |edge_group|
        did_add = false
        @bevel_datas.each do |b_data|
          next if (b_data.all_connected_edges & edge_group).empty? # There is no related edges.

          b_data.add_edges(edge_group)
          did_add = true
          break
        end

        next if did_add

        bevel_data = BevelData.new
        bevel_data.add_edges(edge_group)
        bevel_data.params = @bevel_params
        @bevel_datas << bevel_data
      end
      recalculate_bevels if do_calc
    end

    # @param picked_edges [Array<Sketchup::Edge>]
    def remove_edges_from_selected(picked_edges)
      get_grouped_edges(picked_edges).each do |edge_group|
        # @type [BevelData]
        edited = nil
        @bevel_datas.each do |b_data|
          next if (b_data.all_connected_edges & edge_group).empty? # There is no related edges.

          b_data.remove_edges(edge_group)
          edited = b_data
          break
        end

        if !edited.nil? && edited.primary_edges.empty?
          @bevel_datas.delete(edited) # Check if edited is empty (delete / remove it)
        end
      end
      recalculate_bevels
    end

    def recalculate_bevels
      @bevel_datas.each do |b_data|
        begin
          b_data.calculate(true)
        rescue EdgeOffsetTooSmallError
          # ignore these
        end
      end
    end

    # smooth the perimeter edge based on the setting
    # unless the edge is not soft!
    # @param [Sketchup::Entities] entities
    def smooth_perimeter_edges(entities = Sketchup.active_model.active_entities)
      edges = entities.grep(Sketchup::Edge)
      edges.each do |edge|
        next if !PLUGIN.is_bevel_perimeter?(edge) || !edge.soft?

        edge.smooth = @bevel_params.smooth_perimeter?
      end
    end

  end

end