
module DM::Bevel

  class OffsetND < CoreTool

    OFFSET_REFERENCE_STATE = "Offset Reference"
    OFFSET_STATE = "Offset Action"

    def initialize(proxy_settings)
      # @type [ProxySettings]
      @proxy_settings = proxy_settings
      @tool_state = OFFSET_REFERENCE_STATE
      @bevel_previewer = BevelPreviewer.new
      @edge_extruder = EdgeExtruder.new
      @cursor = get_cursor("bevel_cursor", 1, 1)
      @last_bevel_params = @proxy_settings.bevel_params.clone
      super()
    end

    # Force view to re-draw so it's clear tool is active
    def activate
      super
      begin
        @bevel_datas = ProxySettings.active_bevel.bevel_datas
        @bevel_datas.each do |b_data|
          b_data.calculate(true)
        end
      rescue EdgeOffsetTooSmallError
        # Do nothing.
      end
      Sketchup.active_model.active_view.invalidate
    end

    def deactivate(view)
      super
      @proxy_settings.load_params # ensure settings are correct
    end

    def onCancel(reason, view)
      if @tool_state == OFFSET_REFERENCE_STATE
        @proxy_settings.load_params
        @last_bevel_params = nil
        view.invalidate
      end
      super
    end

    def set_cursor
      # Using default cursor for now
      #UI.set_cursor(@cursor)
    end

    def reset
      @tool_state = OFFSET_REFERENCE_STATE
      @edge_extruder.clear
      set_vcb
      set_status_text
      super
    end

    def set_status_text
      case @tool_state
      when OFFSET_REFERENCE_STATE
        status_text = LH["Pick offset reference point"]
      when OFFSET_STATE
        status_text = LH["Click to define offset or enter value."]
      else
        raise "Unexpected Bevel Tool State: #{@tool_state}" # Internal so no translation
      end
      Sketchup::set_status_text(status_text)
    end

    def set_vcb
      Sketchup.set_status_text(LH["Edge offset"], SB_VCB_LABEL)
      Sketchup.set_status_text("#{get_offset_length(@proxy_settings.bevel_params.edge_offset)}", SB_VCB_VALUE)
    end

    def get_offset_length(val)
      unit = @proxy_settings.modified_length_unit
      if unit == 'mm' # convert from " to mm
        val = val * 25.4
        val = val.round(1) # ensure rounding to 1
      else
        val = val.round(2) # ensure rounding to 2
      end
      "#{val} #{unit}" # manually format the val into mm OR "
    end

    # @param view [Sketchup::View]
    def onMouseMove(flags, x, y, view)
      @bevel_datas = ProxySettings.active_bevel.bevel_datas
      if @tool_state == OFFSET_REFERENCE_STATE
        @proxy_settings.show_bevel_surf = ProxySettings::ShowBevelsurf::BOTH if @proxy_settings.show_bevel_surf == ProxySettings::ShowBevelsurf::BEVELSURF # Ensure proxy surface is visible
        # (plane / face if needed) nearest point on Selected edge as well as adjacent plane
        refresh_edge_extruder(x, y, view)
      elsif @tool_state == OFFSET_STATE
        # distance from reference constrained to plane / face
        @proxy_settings.bevel_params.edge_offset = @edge_extruder.get_offset_length(x, y, view)
        set_vcb
      end
      super
    ensure
      view.invalidate
    end

    # @param view [Sketchup::View]
    def onLButtonDown(flags, x, y, view)
      if @tool_state == OFFSET_REFERENCE_STATE
        if @edge_extruder.valid?
          @tool_state = OFFSET_STATE
          @proxy_settings.bevel_params.edge_offset = @edge_extruder.get_offset_length(x, y, view)
        end
      elsif  @tool_state == OFFSET_STATE
        @proxy_settings.edge_offset = @edge_extruder.get_offset_length(x, y, view)
      end
      super
      set_vcb
      set_status_text
      view.invalidate
    end

    def onLButtonUp(flags, x, y, view)
      if @dragging && @edge_extruder.valid?
        @proxy_settings.edge_offset = @edge_extruder.get_offset_length(x, y, view)
      end
      super
    end

    def refresh_edge_extruder(x, y, view)
      edges = @bevel_datas.flat_map {|b_data| b_data.primary_edges }
      @edge_extruder.clear
      @edge_extruder.set_edge_reference_point(edges, x, y, view) unless edges.empty?
    end

    # @param view [Sketchup::View]
    def draw(view)
      return if @bevel_datas.nil?

      # Check if it should update.
      should_update = false
      @bevel_datas.each do |b_data|
        if b_data.none_calculated? || @last_bevel_params != @proxy_settings.bevel_params
          should_update = true
          @last_bevel_params = @proxy_settings.bevel_params.clone
          break
        end
      end

      @bevel_datas.each do |b_data|
        begin
          if should_update
            b_data.params = @proxy_settings.bevel_params
            b_data.calculate(true)
          end
        rescue EdgeOffsetTooSmallError
          view.tooltip = LH["Edge Offset too small."]
        end

        @bevel_previewer.data = b_data
        @bevel_previewer.draw(view, false)
      end

      if (@tool_state == OFFSET_REFERENCE_STATE || @tool_state == OFFSET_STATE) && @edge_extruder.valid?
        @edge_extruder.draw(view, @tool_state == OFFSET_STATE)
      end
    end

  end # end of class
end