
module DM::Bevel

  class BevelNonDestructive
    include GeometryMixin
    include BevelMixin

    CREASE_KEY = 'crease'

    attr_accessor :recalculate
    attr_accessor :bevel_datas

    # @param [Sketchup::Group or Sketchup::ComponentInstance] the object to be non-destructive beveled
    def initialize(object)
      unless object.is_a?(Sketchup::Group) || object.is_a?(Sketchup::ComponentInstance)
        raise ArgumentError.new(LH["ND Bevel must be initialized with a Group or Component."])
      end

      @coarse_mesh_instance = object
      @proxy_settings = ProxySettings.instance
      @proxy_settings.init_proxy(@coarse_mesh_instance)
      @active_proxymesh = nil
      @coarse_mesh = nil
      @ents = @coarse_mesh_instance.definition.entities
      @bevel_datas = []
      edges = @ents.grep(Sketchup::Edge)
      add_edges_to_selected(edges) # updates @bevel_datas
      @bevel_params = load_params
      @recalculate = false
    end

    # @return [BevelParams]
    def load_params
      params = BevelParams.new
      params.segments = @proxy_settings.segments
      params.edge_offset = @proxy_settings.edge_offset
      params.smooth_perimeter = @proxy_settings.smooth_perimeter?
      params
    end

    def set_params
      @bevel_params = load_params
    end

    # Applies the attribute to the group / component
    # Attempts to enter the proxy mode if version supports
    def self.non_destructive
      if ProxySettings.active_proxy.nil?
        # Start a new proxy
        object = get_object_for_modification(true)
        if object
          Sketchup.active_model.start_operation(LH["Enable Bevel Proxy"], true)
          case object
          when Sketchup::Group, Sketchup::ComponentInstance
            object = object.make_unique if object.is_a?(Sketchup::Group)
            beveler = BevelNonDestructive.new(object)
            beveler.create_proxy
            beveler.create_bevel_preview
            ProxySettings.active_bevel = beveler
            hide_hidden_edges
            ObserverManager.model_observer.new_proxy_transaction # Notify the observer that a new proxy has been created
            beveler.recalculate = true # Need to recalculate once we are inside the correct entities
            Sketchup.active_model.selection.clear
            notify_user_how_to_edit(object)
          else
            raise LH["Unexpected entity type"]
          end
          Sketchup.active_model.commit_operation
        end
      else
        # We are in an active proxy. So just reshow the dialog.
        object = get_object_for_modification(false)
        if object
          # If an object is return, assume user is trying to create another proxy within another proxy.
          UI.messagebox(LH["You cannot create a bevel proxy inside another proxy object."])
        else
          ProxySettings.instance.show
        end
      end
    end

    # Validate the selected entities and return a valid object that can be modified
    # @return [Sketchup::Group or Sketchup::ComponentInstance or false]
    def self.get_object_for_modification(show_msg)
      sel = Sketchup.active_model.selection
      if sel.count == 0
        UI.messagebox(LH["Selection is empty."] + " " + LH["Please select a group or component."]) if show_msg
        return false
      end

      selected_groups_or_components = sel.select {|e| e.instance_of?(Sketchup::Group) || e.instance_of?(Sketchup::ComponentInstance)}
      if selected_groups_or_components.length.zero?
        UI.messagebox(LH["Please select a group or component."]) if show_msg
        return false
      elsif selected_groups_or_components.length != 1
        UI.messagebox(LH["Please select only one group or component."]) if show_msg
        return false
      end

      selected_object = selected_groups_or_components.first
      if selected_object.locked?
        UI.messagebox(LH["Cannot edit a locked group or component."]) if show_msg
        return false
      end

      # Check if this object already has the proxy enabled
      if ProxySettings.is_proxy?(selected_object)
        notify_user_how_to_edit(selected_object, true) if show_msg
        return false
      end
      if selected_object.definition.entities.detect { |e| e.is_a?(Sketchup::Face) }
        return selected_object
      else
        notify_no_faces(selected_object) if show_msg
        return false
      end
      false
    end

    # @param [Sketchup::Group or Sketchup::ComponentInstance]
    def self.notify_user_how_to_edit(object, force_message = false)
      if Sketchup.version.to_i >= 20 # active path can only be set on SU2020+
        path_arr = Sketchup.active_model.active_path
        if path_arr
          path_arr << object
        else
          path_arr = [object]
        end
        ip = Sketchup::InstancePath.new(path_arr)
        Sketchup.active_model.active_path = ip
      else
        key = 'shown_edit_proxy_msg'
        has_shown = Sketchup.read_default(MAIN_DICTIONARY, key, false)
        unless has_shown && !force_message
          UI.messagebox(LH["Double-click to edit the bevel proxy surface"])
          Sketchup.write_default(MAIN_DICTIONARY, key, true)
        end
      end
    end

    def self.notify_no_faces(type)
      case type
      when Sketchup::Group
        message = LH["Selected Group must contain at least one face."]
      when Sketchup::ComponentInstance
        message = LH["Selected Component must contain at least one face."]
      else
        message = LH["Selection needs to be a single Group or Component, with at least one face."]
      end
      UI.messagebox(message)
    end

    # Used to give user visual confirmation that the Bevel operation worked
    def self.show_hidden_edges
      if Sketchup.version.to_i >= 20
        Sketchup.active_model.rendering_options["DrawHiddenGeometry"] = true
      else
        Sketchup.active_model.rendering_options["DrawHidden"] = true
      end
    end

    def self.hide_hidden_edges
      if Sketchup.version.to_i >= 20
        Sketchup.active_model.rendering_options["DrawHiddenGeometry"] = false
      else
        Sketchup.active_model.rendering_options["DrawHidden"] = false
      end
    end

    # Applies crease weight to the given edge or vertex
    # @param [Sketchup::Edge] e
    def self.set_crease(e)
      e.set_attribute(MAIN_DICTIONARY, CREASE_KEY, true)
    end

    # Gets the previously set crease boolean for an edge
    # @param [Sketchup::Edge] e
    # @return [Boolean]
    def self.get_crease_weight(e)
      e.get_attribute(MAIN_DICTIONARY, CREASE_KEY, false)
    end

    # Remove all crease data from this edge
    # @param [Sketchup::Edge] e
    def self.remove_crease(e)
      dicts = e.attribute_dictionaries
      if dicts && dicts[MAIN_DICTIONARY]
        e.delete_attribute(MAIN_DICTIONARY, CREASE_KEY)
      end
    end

    # Check if the given entity has crease attributes
    # @param [Sketchup::Edge] e
    # @return [Bool]
    def self.is_creased?(e)
      if e.get_attribute(MAIN_DICTIONARY, CREASE_KEY, false)
        return true
      else
        return false
      end
    end

    def coarse_entities
      @ents
    end

    def bevel_surface_entities
      @bevel_surface_ents
    end

    def bevel_surface_entities=(ents)
      @bevel_surface_ents = ents
    end

    # Regenerates the bevel surface geometry nested inside a proxy
    def update_bevelsurf
      # update the active proxy if it is currently open, otherwise update the selected group
      # guard against deleted or invalid proxy
      unless @coarse_mesh_instance.valid?
        UI.messagebox(LH["Invalid Proxy"])
        return
      end

      @proxy_settings.refresh_dialog(false) # Update the proxy settings dialog (needed for Undo / Redo)
      bevel_surf_inst = @proxy_settings.get_bevel_surf_instance
      if bevel_surf_inst
        @bevel_surface_ents = bevel_surf_inst.definition.entities
        # If the user has modified the axes of the proxy, the points for OSD mesh
        # will become out of sync with the points for the SketchUp bevel surface mesh so
        # below we ensure that the transformation of the bevel surface instance matches the
        # transform of the proxy. This only works if the user is currently editing a proxy.
        if @coarse_mesh_instance.definition.entities.parent == Sketchup.active_model.active_entities.parent
           bevel_surf_inst.transformation = @coarse_mesh_instance.transformation
        else
          bevel_surf_inst.transformation= Geom::Transformation.new # fixes a show proxy bug when opening and closing
         end
      end
      @bevel_surface_ents.clear!
      @recalculate = false
      create_bevel_preview
    end

    # Creates the bevel geometry, replacing the original mesh
    def commit_bevelsurf(bevel_datas = @bevel_datas)
      beveler = Beveler.new
      bevel_datas.each do |b_data|
        if @recalculate || !b_data.full_calculation? # Calc should never run as it was already done
          b_data.calculate(false)
        end
        beveler.data = b_data
        beveler.generate_geometry(@ents, false) # Create the data at the active entities
      end
      smooth_perimeter_edges(@ents) # Post operation apply perimeter smoothing
    rescue BevelEntitiesError
      UI.messagebox(PLUGIN_NAME + ": " + LH["Geometry is too small for SketchUp to create. Please scale the object up and try again."])
      Sketchup.active_model.abort_operation
      raise # TODO: raise error so above operation can be reverted?
    rescue StandardError => e
      # TODO: Write to logger, and sho w some generic message. Below is temp until this is implemented.
      DM::Bevel::Licensing::Log.error(e.message)
      DM::Bevel::Licensing::Log.error(e.backtrace)
      Sketchup.active_model.abort_operation
      raise # TODO: raise error so above operation can be reverted?
    ensure
      Sketchup.active_model.active_view.invalidate
    end

    # Creates the faces in the model based on the coarse mesh
    # @param [Sketchup::Entities] ents the entities collection to generate the mesh within
    # @return [Array<Sketchup::Edge>] edges the new edges that were added to the model
    def create_mesh_from_coarse(ents)
      proxymesh = @coarse_mesh
      smooth_flags = Geom::PolygonMesh::SOFTEN_BASED_ON_INDEX

      ents.clear! # Ensure the entities are empty

      if ents == Sketchup.active_model.active_entities
        trans = @coarse_mesh_instance.transformation * Sketchup.active_model.edit_transform
        if proxymesh.has_materials?
          generate_textured_mesh(ents, proxymesh, trans, smooth_flags)
        else
          ents.add_faces_from_mesh(proxymesh.mesh.transform!(trans), smooth_flags)
        end
        new_edges = ents.grep(Sketchup::Edge)
      else
        ents.clear!
        if proxymesh.has_materials?
          generate_textured_mesh(ents, proxymesh, IDENTITY, smooth_flags)
        else
          mesh = proxymesh.mesh
          ents.fill_from_mesh(mesh, true, smooth_flags)
        end
        new_edges = ents.grep(Sketchup::Edge)
      end
      new_edges
    end

    # This method is used when the mesh has materials applied to the individual faces.  It is slower
    # than the built-in SU methods for adding faces from mesh but it allows UVs and materials to be maintained.
    # We create an individual face PolygonMesh for each face / quad to be generated.  This allows us to set the materials
    # and uvs.  We can't generate all of the polygons at once because that method forces UVs to be shared (welded) for the
    # vertices which we don't want here.
    # @param [Sketchup::Entities] ents
    # @param [ProxyMesh] mesh
    # @param [Geom::Transformation] trans the transformation to apply to the mesh
    # @param [Number] smooth_flags flags that control how edge smoothing is handled when the mesh is generated
    def generate_textured_mesh(ents, mesh, trans, smooth_flags)
      mesh.mesh.transform!(trans)

      for poly in mesh.polygons
        face_mesh = Geom::PolygonMesh.new
        front_mat = poly.front_material

        points = poly.indices.collect { |i| mesh.point_at(i) }
        point_indices = points.collect { |p| face_mesh.add_point(p) }
        point_indices.each_index do |i|
          point_indices[i] = -point_indices[i] if poly[i] < 0
        end

        face_mesh.add_polygon(point_indices)
        fuvs = poly.front_uv_indices.collect { |f_index| mesh.front_uv_at(f_index) }
        buvs = poly.back_uv_indices.collect { |b_index| mesh.back_uv_at(b_index) }

        begin
          point_indices.each_index do |i|
            face_mesh.set_uv(point_indices[i], fuvs[i], true)
            #face_mesh.set_uv(point_indices[i], buvs[i], false)
          end
        rescue
          # TODO: log this instead of ruby console output
          puts "Front UVs: #{fuvs}"
          puts "Back UVs: #{buvs}"
          raise LH["Failed to apply UVs"]
        end
        ents.add_faces_from_mesh(face_mesh, smooth_flags, front_mat)
      end
    end

    def create_proxy
      bevel_surf_inst = @proxy_settings.create_proxy
      @bevel_surface_ents = bevel_surf_inst.entities
    end

    # Ensures the @bevel_datas are clear and reset
    def add_edges_to_selected(new_edges)
      valid_edges = []
      validate_edges(new_edges, valid_edges, true)
      @bevel_datas.clear
      @bevel_params = load_params
      super(valid_edges, false)
    end

    def create_bevel_preview
      @bevel_params = load_params
      new_edges = coarse_entities.grep(Sketchup::Edge)
      add_edges_to_selected(new_edges) # updates @bevel_datas
      affected_faces = []
      faces = @ents.grep(Sketchup::Face)
      beveler = Beveler.new
      @bevel_datas.each do |b_data|
        b_data.calculate(false) unless b_data.full_calculation?
        beveler.data = b_data
        affected_faces += b_data.affected_faces
        beveler.generate_geometry(@bevel_surface_ents, true)
      end
      unaffected = faces - affected_faces
      # Generate unaffected faces
      if unaffected && !unaffected.empty?
        data_manager = BevelerDataManager.new(unaffected)
        data_manager.face_to_face_loops.each_pair do |face, face_loops|
          face_loops.loop_to_loop_vertices.each_pair do |loop, loop_vertices|
            edge_create_datas = []
            loop_vertices.each do |loop_vertex|
              edge_create_datas << EdgeCreateData.new(loop_vertex.position, loop_vertex.next_vertex.position, { edge: loop_vertex.next_edge, copy_dictionary: true })
            end

            is_hole = face_loops.is_loop_a_hole?(loop)
            data_manager.add_face_create_datas(FaceCreateData.new(edge_create_datas, face.normal, face, is_hole: is_hole, copy_dictionary: true))
          end
        end
        EntitiesGenerator.bulk_add_faces(@bevel_surface_ents, data_manager)
      end
      smooth_perimeter_edges(@bevel_surface_ents) # Post operation apply perimeter smoothing
    rescue BevelEntitiesError
      UI.messagebox(PLUGIN_NAME + ": " + LH["Geometry is too small for SketchUp to create. Please scale the object up and try again."])
      Sketchup.active_model.abort_operation
    rescue StandardError => e
      # TODO: Write to logger, and sho w some generic message. Below is temp until this is implemented.
      DM::Bevel::Licensing::Log.error(e.message)
      DM::Bevel::Licensing::Log.error(e.backtrace)
      Sketchup.active_model.abort_operation
    ensure
      Sketchup.active_model.active_view.invalidate
    end

    # Copied from beveler.rb so it can be used to generate unaffected faces
    # @param points [Array<Geom::Point3d>]
    # @param soften_smooth [Boolean]
    # @param should_loop [Boolean]
    # @return [Array<EdgeCreateData>]
    def points_to_edge_creates(points, soften_smooth, should_loop = true)
      raise UnexpectedStateError.new("Must provide at least two points", points) if points.length <= 1
      raise UnexpectedStateError.new("Not enough points to loop (need at least 3)", points) if should_loop && points.length < 3

      edge_creates = []
      num_pts = points.length
      num_pts -= 1 unless should_loop
      num_pts.times do |pt_i|
        cur_pt = points[pt_i]
        next_pt = points[pt_i + 1]
        next_pt = points[0] if next_pt.nil?
        next if cur_pt == next_pt

        edge_creates << EdgeCreateData.new(cur_pt, next_pt, soft: soften_smooth, smooth: soften_smooth)
      end

      edge_creates
    end

    # Create the ProxyMesh data structure that will prepare the geometry for bevel operations
    def build_coarse_mesh(trans = Sketchup.active_model.edit_transform.inverse)
      mesh = ProxyMesh.new
      tw = Sketchup.create_texture_writer
      mesh.transformation = trans

      @ents.each { |e|
        verts = nil

        if e.class == Sketchup::Face
          loops = e.loops
          is_simple_convex_face = (loops.length == 1)
          verts = e.vertices
          front_uvs = nil #Common.get_front_uvs(e, verts, tw) # TODO: implement uvs
          back_uvs = nil #Common.get_back_uvs(e, verts, tw) # TODO: implement uvs
          # fix UVs for distorted textures
          # normalize_uvs(front_uvs) # TODO: implement uvs
          # normalize_uvs(back_uvs) # TODO: implement uvs
          if is_simple_convex_face
            if verts.length == 4
              mesh.add_poly(e, verts, front_uvs, back_uvs)
            elsif verts.length == 3
              # normalize_uvs(front_uvs) # TODO: implement uvs
              mesh.add_poly(e, verts, front_uvs, back_uvs)
            else
              # more than 4 verts but a simple convex face - OSD will automatically divide this type
              # of face into quads when refining the mesh
              mesh.add_poly(e, verts, front_uvs, back_uvs)
            end
          else
            fm = e.mesh 3
            polys = fm.polygons
            mesh_points = []
            face_verts = {}
            e.vertices.each { |v| face_verts[v] = v.position }
            polys.each { |poly|
              verts = [face_verts.key(fm.point_at(poly[0])),
                       face_verts.key(fm.point_at(poly[1])),
                       face_verts.key(fm.point_at(poly[2]))]
              front_uvs = [fm.uv_at(poly[0], true), fm.uv_at(poly[1], true), fm.uv_at(poly[2], true)]
              back_uvs = [fm.uv_at(poly[0], false), fm.uv_at(poly[1], false), fm.uv_at(poly[2], false)]
              mesh.add_poly(e, verts, front_uvs, back_uvs)
            }
          end
        elsif e.class == Sketchup::Edge
          edge_verts = e.vertices
          if BevelNonDestructive.is_creased?(e)
            # mesh.add_creased_edge(e, edge_verts) # TODO: enable creased edges (edges not to be beveled)
          end
        end
      }
      @coarse_mesh = mesh
    end

  end

end