module DM::Bevel

  class ProxySettings < Window
    extend Singleton
    include BevelMixin

    attr_accessor :dlg, :stored_rendering_options, :bevel_params, :coarse_mesh_instance

    ID_SEGMENTS = "segments"
    ID_OFFSET = "edge_offset"
    ID_SMOOTH_PERIMETER = "smooth_perimeter"
    ID_SHOW_BEVEL_SURF = "show_bevel_surf"
    ID_BEVEL_SURF_TRANSPARENCY = "bevel_surf_transparency"
    ID_BEVEL_SURF_TRANSPARENCY_SLIDER = "bevel_surf_transparency_slider"

    BEVEL_SURFACE_TAG_NAME = "Bevel surf"
    PROXY_MATERIAL_NAME = "Bevel Proxy"
    BEVEL_SURFACE_MATERIAL_NAME = "Bevel Surf"
    BEVEL_SURFACE_GROUP_NAME = "Bevel Surf"
    RENDER_MODE_MONOCHROME = 5
    RENDER_MODE_TEXTURED = 2
    EDGE_COLOR_MODE_COLORED = 0
    EDGE_COLOR_MODE_DEFAULT = 1
    PROXY_ALPHA = 0.0
    BEVEL_SURF_ALPHA = 0.98 # Alpha helps to prevent z-fighting
    BEVEL_SURFACE_COLOR = Sketchup::Color.new(160, 160, 160)
    PROXY_COLOR = Sketchup::Color.new(150, 255, 255)
    PROXY_HIGHLIGHT_COLOR = Sketchup::Color.new(255, 255, 0)

    class ShowBevelsurf
      BEVELSURF = 0
      PROXY = 1
      BOTH = 2
    end

    BEVEL_DEFAULT_OFFSET_PIXELS = 10.0 # offset default
    DEFAULTS = {
        ID_SEGMENTS => 4,
        ID_SMOOTH_PERIMETER => true,
        ID_SHOW_BEVEL_SURF => ShowBevelsurf::BOTH,
        ID_BEVEL_SURF_TRANSPARENCY => 100,
    }

    def self.destroy_dialog
      return unless self.exists?
      self.instance.close
      self.close
    end

    # @param [SketchUp::Group or Sketchup::ComponentInstance] inst
    def initialize
      reg_key = PLUGIN_NAME + "ProxySettings"
      options = {
          :dialog_title => LH["Live Bevel Settings"],
          :preferences_key => reg_key,
          :scrollable => true,
          :resizable => true,
          :width => 320,
          :height => 400,
          :left => 50,
          :top => 300,
          :min_width => 305,
          :min_height => 400,
          :max_width =>800,
          :max_height => 600,
          :style => UI::HtmlDialog::STYLE_UTILITY
      }
      @dialog_filename = "proxy_settings.html"
      super(options)

      # These are some of the rendering options that will get modified during proxy editing
      # mode which we will revert back to the previous settings for the user when proxy
      # mode is exited. Note that not all of the modified options will get reverted since
      # it may be unexpected if they are reverted (eg. render mode)
      @stored_rendering_options = {
          "EdgeColorMode" => nil,
          "InstanceFade" => nil,
          "DisplaySketchAxes" => nil,
          "DrawSilhouettes" => nil,
          "HighlightColor" => nil
      }
    end

    def init_proxy(inst)
      @coarse_mesh_instance = inst
      @bevel_params = BevelParams.new
      load_params
      @bevel_surf_instance = find_bevel_surf
    end

    def load_params
      @bevel_params.segments = segments
      @bevel_params.edge_offset = edge_offset
      @bevel_params.smooth_perimeter = smooth_perimeter?
    end

    def adjust_length_based_on_model(pixels = BEVEL_DEFAULT_OFFSET_PIXELS)
      val = Sketchup.active_model.active_view.pixels_to_model(pixels, @coarse_mesh_instance.bounds.center)
      val = 0.01 if val <= 0.0
      if modified_length_unit == 'mm'
        val *= 25.4
        val = val.round(1) # mm so round to 0.1
        val = 0.1 if val < 0.1 # ensure we have a value we can use
        val = val / 25.4 # convert back to "
      else
        val = val.round(2) # always inches so round to nearest 0.01
        val = 0.01 if val < 0.01 # ensure we have a value we can use
      end
      val
    end

    def add_callbacks
      super
      add_action_callback('change_segments') do |_, is_increment|
        seg = self.segments
        incr = seg == 1 ? 1 : 2 # We are only supporting even segments for now
        seg += is_increment == true ? incr : -incr
        self.segments = seg
      end
      add_action_callback('change_edge_offset') do |_, is_increment|
        edge_offset_callback(is_increment)
      end
      add_action_callback('edge_offset_tool') do |_|
        Sketchup.active_model.select_tool(OffsetND.new(self))
      end
      set_on_closed do
        if CoreTool.active_tool.is_a?(OffsetND)
          Sketchup.active_model.select_tool(nil)
        end
      end
    end

    # Method is used so the unit can update with changes and not be set once
    def edge_offset_callback(is_increment)
      offset = self.edge_offset
      incr = adjust_length_based_on_model(1)
      offset += is_increment == true ? incr : -incr
      offset = 0.01 if offset < 0.0
      self.edge_offset = offset
    end

    def show
      if visible?
        bring_to_front
      else
        initialize
        super
      end
    end

    # Get a reference to the active proxy if the user is currently editing a proxy - otherwise return nil
    # @return [Sketchup::Group or Sketchup::ComponentInstance or nil]
    def self.active_proxy
      ap = Sketchup.active_model.active_path
      return nil unless ap

      leaf = ap.last
      return leaf if is_proxy?(leaf)

      nil
    end

    # Check whether the given entity is a proxy
    def self.is_proxy?(object)
      if object.respond_to?(:definition)
        return object.definition.get_attribute(MAIN_DICTIONARY, 'proxy', false)
      end
      return false
    end

    def self.is_bevel_surf?(object)
      object.is_a?(Sketchup::Group) && object.name == BEVEL_SURFACE_GROUP_NAME
    end

    # Store a reference to the Beveler object that manages the bevel processing for the active proxy
    # This allows all of our tools to access this object so that our tools can trigger updates to the
    # bevel geometry
    # @param [Bevel] object
    def self.active_bevel=(object)
      @active_bevel = object
    end

    # @return [BevelNonDestructive]
    def self.active_bevel
      @active_bevel
    end

    # Gets the group containing the Bevel surface mesh for this proxy
    # @return [Sketchup::Group or nil]
    def get_bevel_surf_instance
      if valid_bevel_surf?
        return @bevel_surf_instance
      else
        return create_proxy
      end
    end

    # Check for Bevel surface group within entities
    # @return [Sketchup::ComponentDefinition or nil]
    def get_bevel_surf_definition
      bevel_surf = get_bevel_surf_instance
      if bevel_surf
        bevel_surf.definition
      else
        nil
      end
    end

    # Traverse the proxy entities to find the bevel_surf group.  This may be needed if
    # an undo or redo was performed which resulted in the bevel_surf_instance variable referencing
    # in invalid entity
    def find_bevel_surf
      @coarse_mesh_instance.definition.entities.find { |e| ProxySettings.is_bevel_surf?(e) }
    end

    # Check if the referenced bevel_surf instance is valid
    # @return [Boolean] returns true if it is a valid group
    def valid_bevel_surf?
      if @bevel_surf_instance && @bevel_surf_instance.valid?
        return true
      end
      false
    end

    def refresh_dialog(su_focus = true)
      set_value(ID_SEGMENTS, segments.to_s)
      set_dlg_offset(ID_OFFSET, edge_offset)
      set_checkbox(ID_SMOOTH_PERIMETER, smooth_perimeter?)
      set_radio(ID_SHOW_BEVEL_SURF, show_bevel_surf)
      refresh_bevel_surf_transparency
      focus_sketchup if su_focus
    end

    def refresh_bevel_surf_transparency
      set_value(ID_BEVEL_SURF_TRANSPARENCY, "#{bevel_surf_transparency}%")
      set_value(ID_BEVEL_SURF_TRANSPARENCY_SLIDER, bevel_surf_transparency)
    end

    # Sets up the proxy and bevel_surf objects for non-destructive editing
    # @return [Sketchup::Group] returns a new group that was created to contain the bevel geometry
    def create_proxy
      make_proxy_invisible
      @coarse_mesh_instance.definition.set_attribute(MAIN_DICTIONARY, 'proxy', true)
      @coarse_mesh_instance.definition.set_attribute(MAIN_DICTIONARY, ID_OFFSET, adjust_length_based_on_model) # Default offset is zoom based so need to set it to avoid unexpected results
      bevel_surf_group = @coarse_mesh_instance.definition.entities.add_group
      bevel_surf_group.name = BEVEL_SURFACE_GROUP_NAME
      bevel_surf_group.layer = get_bevel_surf_layer
      bevel_surf_group.locked = true
      @bevel_surf_instance = bevel_surf_group
      bevel_surf_group
    end

    def segments=(val)
      seg = val.to_i
      seg = 1 if seg < 1
      @bevel_params.segments = seg
      if @bevel_params.segments != segments # No need to create a transaction to store the value already stored
        Sketchup.active_model.start_operation(LH["Change Segments"], true)
        @coarse_mesh_instance.definition.set_attribute(MAIN_DICTIONARY, ID_SEGMENTS, @bevel_params.segments)
        Sketchup.active_model.commit_operation
      end

      set_value(ID_SEGMENTS, @bevel_params.segments)
    end

    def segments
      @coarse_mesh_instance.definition.get_attribute(MAIN_DICTIONARY, ID_SEGMENTS, DEFAULTS[ID_SEGMENTS])
    end

    # Special Parse length method that forces mm or "
    def parse_length(string)
      string.strip! # Trialing space will fail
      string.gsub!(/”|“/, '"') # Curly quotes will fail
      string.gsub!(/‘|’/, "'")
      string.gsub!(/-\s+/,'-')
      if string.match("'")
        string += '"' unless string[-1] == '"' # Since we found ' assume user is dealing with feet and inches
      end
      if string[-1].match(/[[:digit:]]/)
        # the string ends with a digit so assume no unit was provided and add the expected one
        unit = modified_length_unit
        string += " #{unit}"
      end
      return nil if Sketchup.parse_length(string) == nil
      Sketchup.parse_length(string) >= 0.0 ? Sketchup.parse_length(string) : -(Sketchup.parse_length(string.gsub('-', '')))
    end

    def edge_offset=(val)
      Sketchup.active_model.select_tool(nil) if CoreTool.active_tool.is_a?(OffsetND) # User changed offset using dialog so exit the tool
      if val.is_a?(String)
        begin
          val = parse_length(val)
        rescue
          # TODO: any action required
          # Error parsing the text
        end
      end
      # For display purposes and keeping value in sync with display
      if modified_length_unit == 'mm'
        val *= 25.4
        val = val.round(1) # mm so round to 0.1
        @bevel_params.edge_offset = val / 25.4 # convert back to "
      else
        @bevel_params.edge_offset = val.round(2) # always inches so round to nearest 0.01
      end

      if @bevel_params.edge_offset != edge_offset
        Sketchup.active_model.start_operation(LH["Change Edge Offset"], true)
        @coarse_mesh_instance.definition.set_attribute(MAIN_DICTIONARY, ID_OFFSET, @bevel_params.edge_offset)
        Sketchup.active_model.commit_operation
      end

      set_dlg_offset(ID_OFFSET, @bevel_params.edge_offset)
    end

    def edge_offset
      @coarse_mesh_instance.definition.get_attribute(MAIN_DICTIONARY, ID_OFFSET, adjust_length_based_on_model)
    end

    # @param [Boolean]
    def smooth_perimeter=(val)
      bool = val.downcase == 'true'
      ObserverManager.model_observer.needs_update = false # do NOT need to regenerate the bevel
      Sketchup.active_model.start_operation(LH["Change Smooth Perimeter"], true)
      @bevel_params.smooth_perimeter = bool
      @coarse_mesh_instance.definition.set_attribute(MAIN_DICTIONARY, ID_SMOOTH_PERIMETER, bool)
      smooth_perimeter_edges(@bevel_surf_instance.entities)
      Sketchup.active_model.commit_operation

      ObserverManager.model_observer.needs_update = true # resume watching for changes
      set_checkbox(ID_SMOOTH_PERIMETER, @bevel_params.smooth_perimeter?)
    end

    def smooth_perimeter?
      @coarse_mesh_instance.definition.get_attribute(MAIN_DICTIONARY, ID_SMOOTH_PERIMETER, DEFAULTS[ID_SMOOTH_PERIMETER])
    end

    def show_bevel_surf=(val)
      val = val.to_i
      # For optimization, the model observer is only updating the bevel surface when
      # the bevel surf is actually visible.  Therefore, if the user now switches
      # from proxy-only visibility mode to another mode, we now need to force
      # the observer to update the bevel surface geometry
      needs_update = false
      if show_bevel_surf == ShowBevelsurf::PROXY && val != ShowBevelsurf::PROXY
        needs_update = true if ObserverManager.model_observer.proxy_did_change?
      end
      ObserverManager.model_observer.needs_update = needs_update
      Sketchup.active_model.start_operation(LH["Toggle Bevel surface Visibility"], true)
      @coarse_mesh_instance.definition.set_attribute(MAIN_DICTIONARY, 'showBevelsurf', val)
      set_bevel_surf_proxy_visibility
      Sketchup.active_model.commit_operation
      ObserverManager.model_observer.start_watching_proxy_changes
      ObserverManager.model_observer.needs_update = true
      set_radio(ID_SHOW_BEVEL_SURF, show_bevel_surf)
    end

    def show_bevel_surf
      @coarse_mesh_instance.definition.get_attribute(MAIN_DICTIONARY, 'showBevelsurf', DEFAULTS[ID_SHOW_BEVEL_SURF])
    end

    def show_proxy=(val)
      ObserverManager.model_observer.needs_update = false
      Sketchup.active_model.start_operation(LH["Toggle Proxy Visibility"], true)
      @coarse_mesh_instance.definition.set_attribute(MAIN_DICTIONARY, 'showProxy', val)
      Sketchup.active_model.commit_operation
    end

    def bevel_surf_transparency=(val)
      val = val.to_i
      val = 0 if val < 0
      val = 100 if val > 100
      mat = get_bevel_surf_material
      Sketchup.active_model.start_operation(LH["Bevelsurf Transparency"], true)
      # Prevent bevel_surf refresh when simply changing transparency
      ObserverManager.model_observer.needs_update = false
      mat.alpha = val / 100.0
      Sketchup.active_model.commit_operation
      ObserverManager.model_observer.needs_update = true
      refresh_bevel_surf_transparency
    end

    def bevel_surf_transparency
      (get_bevel_surf_material.alpha * 100).to_i
    end

    # Sets up rendering styles and visibility settings based on the settings of the proxy
    def set_rendering_options
      store_rendering_options
      set_bevel_surf_proxy_visibility
    end

    def set_bevel_surf_proxy_visibility
      case show_bevel_surf
      when ShowBevelsurf::BEVELSURF
        make_proxysurf_visible
        make_proxy_invisible
        disable_edge_coloring
        # Ensure vertex selections do not remain visible
        # when proxy geometry is hidden
        Sketchup.active_model.selection.clear
      when ShowBevelsurf::PROXY
        make_bevel_surf_invisible
        make_proxy_visible
        disable_edge_coloring
      when ShowBevelsurf::BOTH
        make_proxysurf_visible
        make_proxy_transparent
      end
    end

    def disable_edge_coloring
      # Restore the user's previously used rendering options
      ro = Sketchup.active_model.rendering_options
      @stored_rendering_options.each do |key, val|
        ro[key] = val if val
      end
    end

    # Stores certain proxy mode rendering options so they can be reverted later
    def store_rendering_options
      ro = Sketchup.active_model.rendering_options
      @stored_rendering_options.each_key do |key|
        val = ro[key]
        @stored_rendering_options[key] = val
      end
    end

    def make_proxy_visible
      restore_proxy_material
      show_all_edges_and_faces(@coarse_mesh_instance.definition)
    end

    def make_proxy_invisible
      hide_all_edges_and_faces(@coarse_mesh_instance)
    end
    
    def restore_proxy_material
      @coarse_mesh_instance.definition.instances.each do |inst|
        mat_name = inst.get_attribute(MAIN_DICTIONARY, 'proxy_old_material', nil)
        if mat_name
          mat = Sketchup.active_model.materials[mat_name]
        else
          mat = nil
        end
        inst.material = mat
      end
    end

    def make_proxy_transparent
      mat = get_proxy_material
      if mat
        @coarse_mesh_instance.definition.instances.each do |inst|
          if inst.material && inst.material != mat
            inst.set_attribute(MAIN_DICTIONARY, 'proxy_old_material', inst.material.name)
          end
          inst.material = mat
        end
        ro = Sketchup.active_model.rendering_options
        ro["RenderMode"] = RENDER_MODE_TEXTURED
        ro["TransparencySort"] = 2
        ro["DisplaySketchAxes"] = false # To avoid z-fighting between proxy edges and axes
        ro["InstanceFade"] = 1.0 # Facilities easier modeling with mirrored components
        ro["DrawSilhouettes"] = false # For performance and because this usually looks bad or organic models
      end
      show_all_edges_and_faces(@coarse_mesh_instance.definition)
    end

    def make_bevel_surf_invisible
      layer = get_bevel_surf_layer
      if layer
        layer.visible = false
      else
        raise LH["Bevel surface Layer not found"]
      end
    end

    def make_proxysurf_visible(use_bevel_surf_material = true)
      layer = get_bevel_surf_layer
      if layer
        layer.visible = true
      else
        raise LH["Subsurface Layer not found"] # TODO: raise custom error
      end
      bevel_surf_inst = get_bevel_surf_instance
      if use_bevel_surf_material
        mat = get_bevel_surf_material
        if mat
          bevel_surf_inst.material = mat if bevel_surf_inst
        end
      else
        bevel_surf_inst.material = nil if bevel_surf_inst
      end
    end

    def get_proxy_material
      mats = Sketchup.active_model.materials
      proxy_mat_name = PROXY_MATERIAL_NAME
      proxy_mat = mats[proxy_mat_name]
      unless proxy_mat
        proxy_mat = mats.add(proxy_mat_name)
        proxy_mat.alpha = PROXY_ALPHA
        proxy_mat.color = PROXY_COLOR
      end
      proxy_mat
    end

    def get_bevel_surf_material
      mats = Sketchup.active_model.materials
      bevel_surf_mat_name = BEVEL_SURFACE_MATERIAL_NAME
      bevel_surf_mat = mats[bevel_surf_mat_name]

      unless bevel_surf_mat
        bevel_surf_mat = mats.add(bevel_surf_mat_name)
        bevel_surf_mat.color = BEVEL_SURFACE_COLOR
        bevel_surf_mat.alpha = BEVEL_SURF_ALPHA
      end
      bevel_surf_mat
    end

    def get_bevel_surf_layer
      layers = Sketchup.active_model.layers
      bevel_surf_layer = layers[BEVEL_SURFACE_TAG_NAME] || layers.add(BEVEL_SURFACE_TAG_NAME)
      bevel_surf_layer
    end

    # Destroys the proxy geometry and commits the bevel surface
    def commit_bevel_surface
      if @bevel_surf_instance.nil? || !@bevel_surf_instance.valid? || @bevel_surf_instance.definition.entities.to_a.empty? # ensure something exists that can be committed -- no empty entities
        UI.messagebox(LH["Error: No geometry to commit."])
        return # abort commit
      end

      Sketchup.active_model.start_operation(LH["Commit Bevel"], true)
      Sketchup.active_model.select_tool(nil) if CoreTool.active_tool.is_a?(OffsetND) # User committed bevel using dialog so exit the tool
      bevel_inst = get_bevel_surf_instance
      if bevel_inst
        beveler = ProxySettings.active_bevel
        make_proxy_visible # Ensure original geometry is visible
        beveler.set_params # ensure the correct settings are being used
        bevel_inst.erase! # Remove the 'preview' group
        # Remove the proxy material from all instances
        restore_proxy_material
        # Create the bevel using the original 'coarse mesh'
        beveler.commit_bevelsurf
        remove_bevel_surf_attributes
      else
        raise LH["No Bevel surface found!"]
      end
      # Cancel the bevel editing variables so that the model observer won't
      # trigger the 'cancel_edit_proxy_mode' method
      ObserverManager.model_observer.cancel_bevel_editing_variables
      disable_edge_coloring
      Sketchup.active_model.commit_operation
    end

    # Removes the bevel surface and reverts the proxy to basic geometry with no artisan attributes
    def remove_bevel_surface
      proxy = ProxySettings.active_proxy
      if proxy
        Sketchup.active_model.start_operation(LH["Remove Proxy"], true)
        if @bevel_surf_instance && @bevel_surf_instance.valid?
          @bevel_surf_instance.erase!
        end
        make_proxysurf_visible # added for older versions
        make_proxy_visible
        disable_edge_coloring
        ObserverManager.model_observer.cancel_bevel_editing_variables
        proxy.definition.attribute_dictionaries.delete(MAIN_DICTIONARY)
        Sketchup.active_model.commit_operation
      end
    end

    def hide_all_edges_and_faces(inst)
      inst.definition.entities.each do |e|
        if e.is_a?(Sketchup::Edge) || e.is_a?(Sketchup::Face)
          e.hidden = true
        end
      end
    end

    def show_all_edges_and_faces(defn)
      defn.entities.each do |e|
        if e.is_a?(Sketchup::Edge) || e.is_a?(Sketchup::Face)
          e.hidden = false
        end
      end
    end

    def remove_bevel_surf_attributes
      dicts = @coarse_mesh_instance.definition.attribute_dictionaries
      if dicts && dicts[MAIN_DICTIONARY]
        dicts.delete(MAIN_DICTIONARY)
      end
    end

    def modified_length_unit
      case get_model_length_unit
      when Length::Meter, Length::Millimeter, Length::Centimeter
        unit = "mm"
      else
        unit = '"'
      end
      unit
    end

    def set_dlg_offset(id, val)
      unit = modified_length_unit
      if unit == 'mm' # convert from " to mm
        val = val * 25.4
        val = val.round(1) # ensure rounding to 1
      else
        val = val.round(2) # ensure rounding to 2
      end
      length = "#{val} #{unit}" # manually format the val into mm OR "
      set_value(id, length)
    end

  end #class


end #module