
module DM::Bevel

  class BevelTool < BaseTool

    SELECTION_STATE = "Selection"
    OFFSET_REFERENCE_STATE = "Offset Reference"
    OFFSET_STATE = "Offset Action"

    DEFAULT_OFFSET = 0.0

    def initialize
      @tool_state = SELECTION_STATE
      @last_selected_edges_pids = []
      @bevel_previewer = BevelPreviewer.new
      @edge_extruder = EdgeExtruder.new
      @cursor = get_cursor("bevel_cursor", 1, 1)
      @select_add_cursor = get_cursor("bevel-add_cursor", 1, 1)
      @select_toggle_cursor = get_cursor("bevel-toggle_cursor", 1, 1)
      @select_subtract_cursor = get_cursor("bevel-remove_cursor", 1, 1)
      super
    end

  def set_cursor
    if @tool_state == SELECTION_STATE || (@tool_state == OFFSET_REFERENCE_STATE && add_remove_keys_down?)
      if @keys_down.include?(CONSTRAIN_MODIFIER_KEY) && @keys_down.include?(COPY_MODIFIER_KEY)
        UI.set_cursor(@select_subtract_cursor)
      elsif @keys_down.include?(CONSTRAIN_MODIFIER_KEY)
        UI.set_cursor(@select_toggle_cursor)
      elsif @keys_down.include?(COPY_MODIFIER_KEY)
        UI.set_cursor(@select_add_cursor)
      else
        UI.set_cursor(@cursor)
      end
    else
      UI.set_cursor(@cursor)
    end
  end

    # @return [BevelParams]
    def load_params
      params = BevelParams.new
      params.edge_offset = DEFAULT_OFFSET
      params.segments = Sketchup.read_default(MAIN_DICTIONARY, "segments", params.segments)
      params.smooth_perimeter = Sketchup.read_default(MAIN_DICTIONARY, "smooth_perimeter", params.smooth_perimeter?)
      params
    end

    def save_params
      Sketchup.write_default(MAIN_DICTIONARY, "segments", @bevel_params.segments)
    end

    # @return [Float]
    def last_edge_offset
      Sketchup.read_default(MAIN_DICTIONARY, "last_edge_offset", DEFAULT_OFFSET)
    end

    def last_edge_offset=(val)
      val = val.to_f.abs
      Sketchup.write_default(MAIN_DICTIONARY, "last_edge_offset", val)
    end

    def reset(hard_reset = false)
      super() if is_selection_states? || hard_reset

      if hard_reset
        @tool_state = SELECTION_STATE
      else
        if @tool_state == OFFSET_STATE
          @tool_state = OFFSET_REFERENCE_STATE
        else
          @tool_state = SELECTION_STATE
        end
      end
      @edge_extruder.clear
      set_vcb
      set_status_text
    end

    # Preselection logic
    def on_preselection
      model = Sketchup.active_model
      edges = collect_valid_edges(model.selection)
      model.selection.clear
      unless edges.empty?
        @tool_state = OFFSET_REFERENCE_STATE # Skip selection state go to state 2
        add_edges_to_selected(edges)
      end
    rescue InvalidEdgeError => err
      msgs = get_invalid_edge_tooltips(err)
      unless msgs.empty?
        msg = LH["Invalid Preselection: "]
        msgs.each {|m| msg += "\r\n- #{m}"}
        PLUGIN.notification(msg, 'bevel-tool_icon.png')
      end
      Sketchup.active_model.selection.clear
    rescue StandardError => err
      show_generic_error(err)
    end

    def set_status_text
      segments_hot_key_tip = LH["Up/Down = Change the number of segments"]
      case @tool_state
      when SELECTION_STATE
      status_text = LH["Pick edges or faces to bevel."] + " "
      status_text += rep_key("ALT", "=") + LH["Hold to use path select"]
      when OFFSET_REFERENCE_STATE
        status_text = LH["Pick offset reference point or use modifier keys to change selection. Double-Click to use last offset"] + " | " + segments_hot_key_tip
      when OFFSET_STATE
        status_text = LH["Click to define offset or enter value."] + " | " + segments_hot_key_tip
      else
        raise "Unexpected Bevel Tool State: #{@tool_state}" # Internal so no translation
      end
      Sketchup::set_status_text(status_text)
    end

    def set_vcb
      if @tool_state == SELECTION_STATE
        Sketchup.set_status_text(LH["Segments"], SB_VCB_LABEL)
        Sketchup.set_status_text("#{@bevel_params.segments}", SB_VCB_VALUE)
      else
        Sketchup.set_status_text(LH["Edge offset"] + "#{DELIMITER} " + LH["Segments"], SB_VCB_LABEL)
        Sketchup.set_status_text("#{@bevel_params.edge_offset.to_l}#{DELIMITER} #{@bevel_params.segments}", SB_VCB_VALUE)
      end
    end

    def onCancel(reason, view)
      @last_selected_edges_pids.clear
      super
      Sketchup.active_model.selection.clear if @tool_state == SELECTION_STATE
    end

    def onUserText(text, view)
      commit_operation = false
      offset_changed = false

      if text[-1] == 's'
        segments = text[0..-2].to_i
        @bevel_params.segments = segments
        commit_operation = !@last_selected_edges_pids.empty?
      elsif @tool_state == SELECTION_STATE || (@tool_state == OFFSET_REFERENCE_STATE && add_remove_keys_down?)
        segments = text.to_i
        @bevel_params.segments = segments
        commit_operation = !@last_selected_edges_pids.empty?
      else
        if text.include?(DELIMITER)
          user_inputs = text.split(DELIMITER)
          if !user_inputs[0].nil? && user_inputs[0] != ''
            edge_offset = get_length_from_vcb(user_inputs[0])
            @bevel_params.edge_offset = edge_offset
            offset_changed = true
          end
          if !user_inputs[1].nil? && user_inputs[1] != ''
            segments = user_inputs[1].strip.to_i
            unless segments.zero?
              @bevel_params.segments = segments
            end
          end
        else
          edge_offset = get_length_from_vcb(text)
          @bevel_params.edge_offset = edge_offset unless edge_offset.nil?
          offset_changed = true
        end
        commit_operation = true
      end

      save_params
      if commit_operation
        if @last_selected_edges_pids.empty?
          is_success = self.create_bevel
        else
          @bevel_params.edge_offset = self.last_edge_offset unless offset_changed
          is_success = recreate_bevel
        end
        reset(true) if is_success
      else
        self.recalculate_bevels
      end

      set_vcb
      set_status_text
      view.invalidate
    end

    def onKeyDown(key, repeat, flags, view)
      suppress = false
      if key == CONSTRAIN_MODIFIER_KEY || key == COPY_MODIFIER_KEY
        unless @tool_state == SELECTION_STATE
          @keys_down.push(key)
          self.onMouseMove(nil, @x, @y, view)
        end
      elsif key == ALT_MODIFIER_KEY
        @keys_down.push(key)
        self.onMouseMove(nil, @x, @y, view)
        suppress = true
      elsif key == VK_UP
        if @bevel_params.single_segment?
          @bevel_params.segments = 2
        else
          @bevel_params.segments += 2
        end
        save_params
        set_vcb
        if @last_selected_edges_pids.empty?
          recalculate_bevels
        else
          recreate_bevel
        end
        suppress = true
      elsif key == VK_DOWN
        if @bevel_params.segments == 2
          @bevel_params.segments = 1
        else
          @bevel_params.segments -= 2
        end
        save_params
        set_vcb
        if @last_selected_edges_pids.empty?
          recalculate_bevels
        else
          recreate_bevel
        end
        suppress = true
      end
      set_status_text
      set_cursor
      view.invalidate
      suppress
    end

    def onKeyUp(key, repeat, flags, view)
      @keys_down.delete(key)
      suppress = false
      if key == ALT_MODIFIER_KEY
        self.onMouseMove(nil, @x, @y, view)
        suppress = true
      end
      set_cursor
      set_status_text
      view.invalidate
      suppress
    end

    # @param view [Sketchup::View]
    def onMouseMove(flags, x, y, view)
      if @tool_state == SELECTION_STATE || (@tool_state == OFFSET_REFERENCE_STATE && add_remove_keys_down?)
        @hovered_edges.clear
        @hovered_edges.concat(get_edges_below_cursor(view, x, y, self.alt_key_down?))
        set_vcb
      end

      if @tool_state == OFFSET_REFERENCE_STATE
        # (plane / face if needed) nearest point on Selected edge as well as adjacent plane
        refresh_edge_extruder(x, y, view)
      end

      if @tool_state == OFFSET_STATE || (@dragging && @edge_extruder.valid?)
        # distance from reference constrained to plane / face
        @bevel_params.edge_offset = @edge_extruder.get_offset_length(x, y, view)
        set_vcb
      end
      super
      view.tooltip = ''
    rescue InvalidEdgeError => err
      msgs = get_invalid_edge_tooltips(err)
      view.tooltip = msgs.first unless msgs.empty?
    ensure
      view.invalidate
    end

    def alt_key_down?
      @keys_down.include?(ALT_MODIFIER_KEY)
    end

    def add_remove_keys_down?
      (@keys_down.include?(CONSTRAIN_MODIFIER_KEY) && @keys_down.include?(COPY_MODIFIER_KEY)) || @keys_down.include?(COPY_MODIFIER_KEY) || @keys_down.include?(CONSTRAIN_MODIFIER_KEY)
    end

    def is_selection_states?
      @tool_state == SELECTION_STATE || @tool_state == OFFSET_REFERENCE_STATE
    end

    # @param view [Sketchup::View]
    def onLButtonDown(flags, x, y, view)
      if @tool_state == SELECTION_STATE && !@hovered_edges.empty?
        @last_selected_edges_pids.clear
        @bevel_params.edge_offset = DEFAULT_OFFSET if @bevel_datas.empty?
        add_edges_to_selected(@hovered_edges.uniq)
        if @bevel_datas.length > 0
          @tool_state = OFFSET_REFERENCE_STATE
          refresh_edge_extruder(@x, @y, view)
        end
      elsif @tool_state == OFFSET_REFERENCE_STATE
        if add_remove_keys_down?
          return if @hovered_edges.empty?

          select_edges_by_keys(@hovered_edges)
          refresh_edge_extruder(@x, @y, view)
        elsif @edge_extruder.valid?
          @tool_state = OFFSET_STATE
          @bevel_params.edge_offset = @edge_extruder.get_offset_length(x, y, view)
        end
      elsif  @tool_state == OFFSET_STATE
        save_params
        reset(true) if self.create_bevel
      end
      super
      set_vcb
      set_status_text
    rescue StandardError => err
      show_generic_error(err)
    ensure
      view.invalidate
    end



    def onLButtonUp(flags, x, y, view)
      if @dragging
        save_params
        reset(true) if self.create_bevel
      end
      super
    end

    def onLButtonDoubleClick(flags, x, y, view)
      if is_selection_states? && add_remove_keys_down?
        return if @hovered_edges.empty?

        all_edges = []
        @hovered_edges.each do |edge|
          next if all_edges.include?(edge)

          all_edges.concat(collect_valid_edges(edge.all_connected))
        end

        # A bit of a hack. OnLbuttonDown always gets called before onLButtonDoubleClick. So we have to undo what the
        # onLButtonDown selected before checking all edges. (Applies only on the toggle command)
        select_edges_by_keys(@hovered_edges) if @keys_down.length == 1  && @keys_down.include?(CONSTRAIN_MODIFIER_KEY)
        select_edges_by_keys(all_edges)
      elsif @tool_state == OFFSET_STATE
        # onLButtonDown is always called before onLbuttonDoubleClick.
        # So @tool_state = OFFSET_REFERENCE_STATE (when this action is called.)
        unless self.last_edge_offset.zero?
          @bevel_params.edge_offset = self.last_edge_offset
          save_params
          reset(true) if self.create_bevel
        end
      end
      set_vcb
      set_status_text
    rescue InvalidEdgeError => err
      msgs = get_invalid_edge_tooltips(err)
      view.tooltip = msgs.first unless msgs.empty?
    rescue StandardError => err
      show_generic_error(err)
    ensure
      view.invalidate
    end

    # @param picked_edges [Array<Sketchup::Edge>]
    def add_edges_to_selected(picked_edges, do_calc = true)
      super
      update_selection if @last_selected_edges_pids.empty?
    end

    # @param picked_edges [Array<Sketchup::Edge>]
    def remove_edges_from_selected(picked_edges)
      super
      Sketchup.active_model.selection.remove(picked_edges) # Update the SketchUp selection
      update_selection
    end

    # Needed for some strange SketchUp bug related to all_connected and selection visibility maybe?
    def update_selection
      edges = []
      @bevel_datas.each do |b_data|
        b_data.primary_edges.each do |edge|
          edges.push(edge)
        end
      end
      Sketchup.active_model.selection.add(edges) # Update the SketchUp selection
    end

    # @param view [Sketchup::View]
    def draw(view)
      view.line_width = 4

      if @tool_state == SELECTION_STATE || (@tool_state == OFFSET_REFERENCE_STATE && add_remove_keys_down?)
        if @keys_down.include?(CONSTRAIN_MODIFIER_KEY) && @keys_down.include?(COPY_MODIFIER_KEY)
          selected_edges = []
          removed_edges = @hovered_edges
        elsif @keys_down.include?(CONSTRAIN_MODIFIER_KEY)
          current_edges = @bevel_datas.flat_map {|b_data| b_data.primary_edges }
          selected_edges = @hovered_edges - current_edges
          removed_edges = current_edges & @hovered_edges
        else
          # COPY_MODIFIER_KEY (Add)
          selected_edges = @hovered_edges
          removed_edges = []
        end
        view.drawing_color = 'green'
        selected_edges.each do |edge|
          view.draw2d(GL_LINE_STRIP, [edge.start.position, edge.end.position].map {|pt| view.screen_coords(pt)})
        end

        view.drawing_color = 'red'
        removed_edges.each do |edge|
          view.draw2d(GL_LINE_STRIP, [edge.start.position, edge.end.position].map {|pt| view.screen_coords(pt)})
        end
      end

      @bevel_datas.each do |b_data|
        begin
          b_data.calculate(true) if @tool_state == OFFSET_STATE || (@dragging && @edge_extruder.valid?)
        rescue EdgeOffsetTooSmallError
          view.tooltip = LH["Edge Offset too small"]
        end

        @bevel_previewer.data = b_data
        low_detail = @tool_state == SELECTION_STATE
        @bevel_previewer.draw(view, low_detail)
      end

      if ((@tool_state == OFFSET_REFERENCE_STATE && !add_remove_keys_down?) || @tool_state == OFFSET_STATE) && @edge_extruder.valid?
        @edge_extruder.draw(view, @tool_state == OFFSET_STATE || @dragging)
      end
    end

    def refresh_edge_extruder(x, y, view)
      edges = @bevel_datas.flat_map {|b_data| b_data.primary_edges }
      @edge_extruder.clear
      @edge_extruder.set_edge_reference_point(edges, x, y, view) unless edges.empty?
    end

    # @param picked_edges [Array<Sketchup::Edge>]
    def select_edges_by_keys(picked_edges)
      picked_edges = picked_edges.uniq
      if @keys_down.include?(CONSTRAIN_MODIFIER_KEY) && @keys_down.include?(COPY_MODIFIER_KEY)
        remove_edges_from_selected(picked_edges)
        @tool_state = SELECTION_STATE if @bevel_datas.empty?
      elsif @keys_down.include?(COPY_MODIFIER_KEY)
        add_edges_to_selected(picked_edges)
      elsif @keys_down.include?(CONSTRAIN_MODIFIER_KEY)
        current_edges = @bevel_datas.flat_map {|b_data| b_data.primary_edges }
        selected_edges = picked_edges - current_edges
        removed_edges = current_edges & picked_edges

        add_edges_to_selected(selected_edges) unless selected_edges.empty?
        remove_edges_from_selected(removed_edges) unless removed_edges.empty?
        @tool_state = SELECTION_STATE if @bevel_datas.empty?
      end
    end

    # @return [Boolean]
    def recreate_bevel
      return if @last_selected_edges_pids.empty?

      Sketchup.undo
      edges = Sketchup.active_model.find_entity_by_persistent_id(@last_selected_edges_pids)
      add_edges_to_selected(edges, false)
      is_success = self.create_bevel(true)
      @bevel_datas.clear if is_success
      is_success
    end

    # @param is_post_edit [Boolean]
    # @return [Boolean]
    def create_bevel(is_post_edit = false)
      model = Sketchup.active_model
      model.start_operation(LH["Apply Bevel"], true)

      @last_selected_edges_pids.clear
      beveler = Beveler.new
      @bevel_datas.each do |b_data|
        b_data.primary_edges.map do |edge|
          @last_selected_edges_pids << edge.persistent_id
        end

        begin
          b_data.calculate(false) unless b_data.full_calculation?
        rescue EdgeOffsetTooSmallError
          raise BevelEntitiesError # Reraise to show the "too small" error msg.
        end
        beveler.data = b_data
        beveler.generate_geometry(model.active_entities)
      end

      self.last_edge_offset = @bevel_params.edge_offset
      model.commit_operation
      true
    rescue BevelEntitiesError
      @last_selected_edges_pids.clear unless is_post_edit
      UI.messagebox(PLUGIN_NAME + ": " + LH["Geometry is too small for SketchUp to create. Please scale the object up and try again."])
      model.abort_operation
      false
    rescue StandardError => e
      @last_selected_edges_pids.clear
      show_generic_error(e)
      model.abort_operation
      false
    end

    # @param err [StandardError]
    def show_generic_error(err)
      DM::Bevel::Licensing::Log.error(err.message)
      DM::Bevel::Licensing::Log.error(err.backtrace)
      UI.messagebox(LH["Unable to bevel the selected edge(s). Please check adjacent geometry for reversed faces or edges with more than 2 faces."])
      reset(true) # Force a hard reset.
    end

  end
end