
module DM::Bevel

  class Beveler
    include GeometryMixin

    def initialize
      # @type [BevelData]
      @bevel_data = nil
    end

    # @param bevel_data [BevelData]
    def data=(bevel_data)
      raise UnexpectedTypeError, "Must be of type BevelData" unless bevel_data.instance_of?(BevelData)
      @bevel_data = bevel_data
    end

    # @param ents [Sketchup::Entities]
    def generate_geometry(ents, no_erase = false)
      raise UnexpectedStateError, "@bevel_data not set" if @bevel_data.nil?

      rGenerate_geometry(ents, no_erase)
    rescue BevelError
      raise # Let these errors pass through
    rescue StandardError => error
      # Wrap every other error as a BevelDebugError
      raise BevelDebugError, error.message, error.backtrace
    end

    private

    # @param bevel_data [BevelData]
    # @return [BevelerDataManager]
    def calculate(bevel_data)
      data_manager = BevelerDataManager.new(bevel_data.affected_faces)

      # Normal/basic border cases.
      bevel_data.each_transition.each do |transition_segment|
        next if transition_segment.point_curve?

        transition_segment.face_corner_offsets.each do |face_corner_offset|
          cur_face_corner = face_corner_offset.face_corner
          next if cur_face_corner.three_point_corner? # Handle this separately
          next if cur_face_corner.all_secondary?
          next if [CoplanarFaceCorner, ParallelFaceCorner, QuadFaceCorner].include?(cur_face_corner.class)

          face_loops = data_manager.face_to_face_loops[cur_face_corner.face]
          loop_vertex = face_loops.get_loop_vertex(cur_face_corner)
          face_loops.set_override_pts(loop_vertex, face_corner_offset.offset_pt)
        end
      end

      # Parallel joint cases.
      bevel_data.each_transition.each do |transition_segment|
        next if transition_segment.point_curve?

        transition_segment.face_corner_offsets.each do |face_corner_offset|
          face_corner = face_corner_offset.face_corner
          next if face_corner.instance_of?(CoplanarFaceCorner)
          next if face_corner.instance_of?(QuadFaceCorner)
          next if face_corner.all_secondary?
          next if face_corner.three_point_corner? # Handle this separately

          # @type [FaceLoops]
          face_loops = data_manager.face_to_face_loops[face_corner.face]
          loop_vertex = face_loops.get_loop_vertex(face_corner)
          corner = face_corner.corner
          next unless face_corner.instance_of?(ParallelFaceCorner)

          # Have to check if this face_corner is being used as part of a PointCurve.
          contains_point_curve = false
          bevel_data.transitions_at_face_corner(face_corner).each do |other_transition_segment|
            next unless other_transition_segment.point_curve?

            contains_point_curve = true
            break
          end
          next if contains_point_curve

          # Handle normal parallel joint cases.
          offset_pts = [face_corner_offset.offset_pt]

          # If this is the start/end of parallel joint, and it's not all_primary,
          # we need to add in the closing vertex to the list of offset pts.
          unless face_corner.all_primary?
            edge_vert = face_corner.primary_secondary[0]
            if loop_vertex.next_vertex == edge_vert.opposite_vertex
              # This means prev_vertex is the secondary_edge
              offset_pts.insert(0, corner.vertex.position)
            else
              # This means next_vertex is the secondary_edge
              offset_pts << corner.vertex.position
            end
          end
          face_loops.set_override_pts(loop_vertex, offset_pts)
        end
      end

      # PointCurve cases
      # A PointCurve case is drawing a straight line across a all_secondary face_corner. (when there is patching)
      # So apply that point only to the face_corners that are all_secondary
      bevel_data.each_transition.each do |transition_segment|
        next unless transition_segment.point_curve?

        edge_vert_to_offset_pt = {}
        transition_segment.face_corner_offsets.each do |face_corner_offset|
          # The pts always lie on the edge_verts (except for ParallelFaceCorners). So organize that way.
          face_corner = face_corner_offset.face_corner
          face_corner.edge_verts.each do |edge_vert|
            # Always add the offset pt if it's a ParallelFaceCorner
            next if !face_corner.instance_of?(ParallelFaceCorner) && !face_corner_offset.offset_pt.on_line?(edge_vert.edge_line)

            edge_vert_to_offset_pt[edge_vert] = face_corner_offset.offset_pt
          end
        end

        face_corner_to_offset_pts = Hash.new {|h,k| h[k] = []}
        edge_vert_to_offset_pt.each_pair do |edge_vert, offset_pt|
          # @type [EdgeVertex]
          edge_vert = edge_vert
          edge_vert.face_corners.each do |face_corner|
            next if face_corner.instance_of?(FakeFaceCorner)
            next if face_corner.instance_of?(QuadFaceCorner) # handled in quad section

            if face_corner.instance_of?(CoplanarFaceCorner)
              # Have to grab the inner face_corner
              face_corner = face_corner.closest_coplanar_face_corner(edge_vert)
            end
            face_corner_to_offset_pts[face_corner] << offset_pt
          end
        end

        face_corner_to_offset_pts.each_pair do |face_corner, offset_pts|
          raise UnexpectedStateError.new('', face_corner) if offset_pts.length > 2 # TODO: Logic check. Shouldn't be possible but can't remember.
          next unless offset_pts.length == 2
          # @type [FaceCorner]
          face_corner = face_corner
          face_loops = data_manager.face_to_face_loops[face_corner.face]
          loop_vertex = face_loops.get_loop_vertex(face_corner)
          # Sort the pts.
          if face_corner.instance_of?(ParallelFaceCorner)
            # Find the offset_pt that is on the edge_vert.
            if offset_pts.first.on_line?([face_corner.vertex.position, face_corner.mid_vector])
              offset_pts.reverse! unless self.point_on_line?(loop_vertex.position, loop_vertex.next_vertex.position, offset_pts.last)
            else
              offset_pts.reverse! unless self.point_on_line?(loop_vertex.position, loop_vertex.prev_vertex.position, offset_pts.first)
            end
          else
            offset_pts.reverse! unless loop_vertex.are_start_end_ordered?(offset_pts)
          end

          face_loops.set_override_pts(loop_vertex, offset_pts)
        end
      end

      # For the three_point corners, we need to add that to our border.
      bevel_data.corner_to_patch.each_pair do |corner, patch_data|
        patch_data.transition_loop.transition_segments.each do |transition_segment|
          next unless transition_segment.three_point_segment?

          face_corner = transition_segment.face_corner_offset1.face_corner
          next if face_corner.instance_of?(CoplanarFaceCorner)
          next if face_corner.instance_of?(QuadFaceCorner)

          curve_pts = transition_segment.curve.points

          face_loops = data_manager.face_to_face_loops[face_corner.face]
          loop_vertex = face_loops.get_loop_vertex(face_corner)

          curve_pts.reverse! unless face_corner.edge_vert1.opposite_vertex == loop_vertex.prev_vertex
          face_loops.set_override_pts(loop_vertex, curve_pts)
        end
      end

      # Determine how each vertex/edge will be affected based on the bevel_data
      bevel_data.affected_corners.each_pair do |vertex, corner|
        next if corner.all_primary?
        next if corner.requires_patching?

        if corner.num_primary_edges == 1
          next if corner.edge_verts.length == 2 && corner.contains_quad? # Ultra specific case that is handled separately.

          transition_segment = bevel_data.transitions_at_corner(corner).first
          curve_pts = transition_segment.curve.points.clone

          # Force these cases to be hardened.
          edge_creates = points_to_edge_creates(curve_pts, false, false)
          data_manager.add_force_smoothness_edge_create(edge_creates)

          adjacent_edge_verts = bevel_data.one_transition_edge_verts(corner)
          if corner.edge_verts.length > 3 || corner.is_parallel_joint? || (corner.contains_quad? && bevel_data.one_transition_edge_verts(corner).length == 1)
            # Check if we can merge with a face instead
            merged_with_face = false
            corner.face_corners.each do |face_corner|
              next if face_corner.instance_of?(QuadFaceCorner)
              next if face_corner.instance_of?(CoplanarFaceCorner)
              next if face_corner.instance_of?(FakeFaceCorner)

              common_edge_verts = face_corner.edge_verts - adjacent_edge_verts
              next if common_edge_verts.empty?

              # We check the plane and not the face as it's possible to 'extend' a face in these cases.
              face_plane = face_corner.face.plane
              next unless points_on_plane?(curve_pts, face_plane)

              face_loops = data_manager.face_to_face_loops[face_corner.face]
              next if face_loops.nil?

              loop_vertex = face_loops.get_loop_vertex(face_corner)

              curve_pts.reverse! if curve_pts.last.on_line?(loop_vertex.prev_line)
              face_loops.set_override_pts(loop_vertex, curve_pts)
              merged_with_face = true
              break
            end

            unless merged_with_face
              # We need to create a extra face
              primary_edge_vert = corner.primary_edge_verts.first
              normal = primary_edge_vert.edge_vector
              normal.reverse! if primary_edge_vert.concave?

              edge_creates = points_to_edge_creates((curve_pts + [vertex.position]), false, false)
              data_manager.add_face_create_datas(FaceCreateData.new(edge_creates, normal))
            end
          else
            next if adjacent_edge_verts.empty? # This happens when there are edges with >2 faces.

            # Figure out which face we're merging into
            shared_edge_vert = transition_segment.shared_edge_vert

            # We know one of the offset pts is off the adjacent edge_vert. So pick the edge_vert that has a pt on its edge_vert and get the face.
            if adjacent_edge_verts.length == 1
              adjacent_edge_vert = adjacent_edge_verts.first
              adjacent_face_corner = (adjacent_edge_vert.face_corners - shared_edge_vert.face_corners).first
              if adjacent_face_corner.instance_of?(CoplanarFaceCorner)
                if bevel_data.is_one_transition_curved_into_self?(corner)
                  # In this case, get the closest coplanar_face_corner to the adjacent_edge_vert
                  adjacent_face_corner = adjacent_face_corner.closest_coplanar_face_corner(adjacent_edge_vert)
                  # We need to add the corner.vertex to the curvepts so the coplanar edges don't snap to the curve.
                  if self.point_on_line?(adjacent_edge_vert.start_pt, adjacent_edge_vert.end_pt, curve_pts.first)
                    curve_pts << corner.vertex.position
                  else
                    curve_pts.insert(0, corner.vertex.position)
                  end
                end
              end
            else
              adjacent_edge_vert = adjacent_edge_verts.last
              adjacent_face_corner = adjacent_edge_verts.first.get_shared_face_corner(adjacent_edge_verts.last)
            end
            next if adjacent_face_corner.instance_of?(FakeFaceCorner)

            if adjacent_face_corner.instance_of?(CoplanarFaceCorner) && adjacent_edge_verts.length == 2
              if !adjacent_edge_verts.first.coplanar? && !adjacent_edge_verts.last.coplanar? # If any are coplanar, skip this.
                if (!adjacent_edge_verts.first.concave? && !adjacent_edge_verts.last.concave? && transition_segment.shared_edge_vert.concave?) ||
                  (adjacent_edge_verts.first.concave? && adjacent_edge_verts.last.concave? && !transition_segment.shared_edge_vert.concave?)
                  # Case where the one concave edge is 'extending' off the face. Recreate a face to fill in the void.
                  # TODO: This will create coplanar edges. Should we delete those edges to merge onto the adjacent faces?
                  #       Concern is having to rely on Sketchup to merge the faces, and we also lose control of which material
                  #       takes priority (if any).
                  face_pts = [corner.vertex.position] + curve_pts
                  edge_creates = points_to_edge_creates(face_pts, false, true)
                  edge_creates.first.soften_smooth = true
                  edge_creates.last.soften_smooth = true
                  data_manager.add_face_create_datas(FaceCreateData.new(edge_creates, adjacent_face_corner.face.normal))
                end
              end
            else
              next if adjacent_face_corner.instance_of?(QuadFaceCorner) && adjacent_edge_verts.length == 2 # This case handled in QuadFaceCorner section.

              # We are merging into a face in this case.
              adjacent_face_corner = adjacent_face_corner.closest_face_corner(adjacent_edge_vert) if adjacent_face_corner.instance_of?(QuadFaceCorner)
              face_loops = data_manager.face_to_face_loops[adjacent_face_corner.face]
              loop_vertex = face_loops.get_loop_vertex(adjacent_face_corner)
              curve_pts.reverse! unless loop_vertex.are_start_end_ordered?(curve_pts)
              face_loops.set_override_pts(loop_vertex, curve_pts)
            end
          end
        else
          # The 2 primary edge_vert case
          raise UnexpectedStateError.new('', corner) unless corner.num_primary_edges == 2

          if corner.v_corner?
            v_curves = bevel_data.get_v_curves(corner)
            curve1 = v_curves[0]
            curve2 = v_curves[1]
            # TODO: Sort of a temporary hack. Issues occur with reversed faces, which mess up the edge_vert.concave? method, which then messes up corner.v_corner?
            next if curve1.are_ends_touching?(curve2, true)

            (1..curve1.length - 1).each do |c_i|
              cur_pt1 = curve1[c_i]
              prev_pt1 = curve1[c_i - 1]
              cur_pt2 = curve2[c_i]
              prev_pt2 = curve2[c_i - 1]

              face_pts = [cur_pt1, cur_pt2, prev_pt2]
              face_pts << prev_pt1 if c_i - 1 != 0

              face_vec = calculate_expected_face_normal(corner, face_pts)
              if points_planar?(face_pts)
                edge_create_data = points_to_edge_creates(face_pts, true)
                data_manager.add_face_create_datas(FaceCreateData.new(edge_create_data, face_vec))
              else
                # triangulate the pts
                edge_create_data = points_to_edge_creates(face_pts[0..2], true)
                data_manager.add_face_create_datas(FaceCreateData.new(edge_create_data, face_vec))
                edge_create_data = points_to_edge_creates([face_pts[2], face_pts[3], face_pts[0]], true)
                data_manager.add_face_create_datas(FaceCreateData.new(edge_create_data, face_vec))
              end
            end

            # # Force the last edge to not be soft
            data_manager.add_force_smoothness_edge_create(EdgeCreateData.new(v_curves[0].points.last, v_curves[1].points.last, soft: false, smooth: false))

            _edge_vert_to_transition_pts = bevel_data.get_edge_vert_to_transition_pts(corner)
            _secondary_edge_verts = _edge_vert_to_transition_pts.keys
            if _secondary_edge_verts.length > 1
              # Check if we're merging into a QuadFaceCorner. If so, ignore that as that is handled later.
              related_face_corners = _secondary_edge_verts[0].face_corners & _secondary_edge_verts[1].face_corners
              if related_face_corners.length == 1 && (related_face_corners.first.instance_of?(QuadFaceCorner) || related_face_corners.first.instance_of?(CoplanarFaceCorner))
                # Don't do these cases here. That is handled below.
              else
                # If the secondary_edge_verts are adjacent, then we need to update that adjacent faces vertice data.
                _is_adjacent = false
                _secondary_edge_verts.length.times do |ev_i|
                  cur_edge_vert = _secondary_edge_verts[ev_i]
                  next_edge_vert = _secondary_edge_verts[ev_i + 1]
                  next_edge_vert = _secondary_edge_verts[0] if next_edge_vert.nil?

                  next unless cur_edge_vert.adjacent?(next_edge_vert)

                  _is_adjacent = true
                  face_corner = cur_edge_vert.get_shared_face_corner(next_edge_vert)
                  next if face_corner.nil?
                  next if face_corner.instance_of?(FakeFaceCorner)

                  face_loops = data_manager.face_to_face_loops[face_corner.face]
                  loop_vertex = face_loops.get_loop_vertex(face_corner)

                  face_pts = _edge_vert_to_transition_pts[cur_edge_vert] + _edge_vert_to_transition_pts[next_edge_vert]
                  face_pts.reverse! if loop_vertex.prev_vertex == next_edge_vert.opposite_vertex # Make sure they're in the right order

                  face_loops.set_override_pts(loop_vertex, face_pts)
                  data_manager.add_force_smoothness_edge_create(EdgeCreateData.new(face_pts[0], face_pts[1], soft: !bevel_data.params.single_segment?, smooth: !bevel_data.params.single_segment?))
                end

                # create a face to fill in the hole.
                unless _is_adjacent
                  # Check if the v-curve will merge into a face.
                  merged_with_face = false                  
                  corner.face_corners.each do |face_corner|
                    next if face_corner.instance_of?(QuadFaceCorner)
                    next if face_corner.instance_of?(CoplanarFaceCorner)
                    next if face_corner.instance_of?(FakeFaceCorner)

                    # v_curves always 'wrap' onto a face, so check that the points fall on the face.
                    next unless self.point_on_face?(face_corner.face, curve1.end_pt)
                    next unless self.point_on_face?(face_corner.face, curve2.end_pt)

                    face_loops = data_manager.face_to_face_loops[face_corner.face]
                    next if face_loops.nil?

                    loop_vertex = face_loops.get_loop_vertex(face_corner)

                    # Clear whatever override exists here.
                    face_loops.clear_override_pts(loop_vertex)
                    override_pts = [curve1.end_pt, curve2.end_pt]
                    if self.point_on_line?(loop_vertex.position, loop_vertex.next_loop_vertex.position, override_pts[0]) ||
                        self.point_on_line?(loop_vertex.position, loop_vertex.prev_loop_vertex.position, override_pts[1])
                      override_pts.reverse!
                    end

                    face_loops.set_override_pts(loop_vertex, override_pts)
                    data_manager.add_force_smoothness_edge_create(EdgeCreateData.new(override_pts[0], override_pts[1], soft: !bevel_data.params.single_segment?, smooth: !bevel_data.params.single_segment?))
                    merged_with_face = true
                    break
                  end

                  unless merged_with_face
                    face_pts = [corner.vertex.position, curve1.end_pt, curve2.end_pt]
                    normal = Geom::Vector3d.new
                    corner.primary_edge_verts.each { |edge_vert| normal += edge_vert.edge_vector }
                    edge_create_datas = points_to_edge_creates(face_pts, false, true)
                    data_manager.add_face_create_datas(FaceCreateData.new(edge_create_datas, normal.normalize))
                  end
                end

              end
            end

          else
            transition_segments = bevel_data.transitions_at_corner(corner)
            next if transition_segments[0].curve.are_ends_touching?(transition_segments[1].curve)

            # Nothing is touching, so fill in the gaps. (Case with 6+ edge_verts and you pick the two 'across' from each other.)
            transition_segments.each do |transition_segment|
              face_pts = [corner.vertex.position] + transition_segment.curve.points
              normal = transition_segment.shared_edge_vert.edge_vector
              edge_create_datas = points_to_edge_creates(face_pts, false, true)
              data_manager.add_face_create_datas(FaceCreateData.new(edge_create_datas, normal.normalize))

              data_manager.add_force_smoothness_edge_create(points_to_edge_creates(transition_segment.curve.points, false, false))
            end
          end
        end
      end

      # Quad FaceCorner cases
      # We have to move the diagonal to maintain the quad.
      # FYI: QuadFaceCorners cannot be three_point_corners.
      # @type [Hash {Sketchup::Edge => Array<QuadFaceCorner, Array<Geom::Point3d>, FaceCorner, FaceCorner>}]
      diagonal_edge_to_quad_curve_start_end_face_corners = Hash.new {|h,k| h[k] = []}
      bevel_data.affected_corners.each do |_, corner|
        next unless corner.contains_quad?

        _edge_vert_to_transition_pts = bevel_data.get_edge_vert_to_transition_pts(corner)
        _secondary_edge_verts = _edge_vert_to_transition_pts.keys
        corner.face_corners(QuadFaceCorner).each do |quad_face_corner|
          if corner.num_primary_edges == 1 && _secondary_edge_verts.length == 2 && (quad_face_corner.edge_verts & _secondary_edge_verts).length == 2
            # Single Primary edge case
            # Case where we have one curve merging onto the quad.
            # We need to set the face_corners to the start/end of the curve, and then fill in the middle.
            transition_segment = bevel_data.transitions_at_corner(corner).first
            curve = transition_segment.curve
            if curve.start_pt.on_line?(quad_face_corner.edge_vert1.edge_line)
              start_face_corner = quad_face_corner.face_corner1
              end_face_corner = quad_face_corner.face_corner2
            else
              start_face_corner = quad_face_corner.face_corner2
              end_face_corner = quad_face_corner.face_corner1
            end

            face_loops = data_manager.face_to_face_loops[start_face_corner.face]
            loop_vertex = face_loops.get_loop_vertex(start_face_corner)
            face_loops.set_override_pts(loop_vertex, curve.start_pt)

            face_loops = data_manager.face_to_face_loops[end_face_corner.face]
            loop_vertex = face_loops.get_loop_vertex(end_face_corner)
            face_loops.set_override_pts(loop_vertex, curve.end_pt)

            # We need to track if both ends of the diagonal_edge is being affected by this case.
            diagonal_edge_to_quad_curve_start_end_face_corners[quad_face_corner.quad_diagonal_edge_vert.edge] << [quad_face_corner, curve.points.clone, start_face_corner, end_face_corner]
          elsif corner.num_primary_edges == 2 && corner.v_corner? && _secondary_edge_verts.length == 2 && (quad_face_corner.edge_verts & _secondary_edge_verts).length == 2
            # Two primary edge case, and it's a v_corner where the v is being merged onto the quad.
            start_face_corner = quad_face_corner.face_corner1
            end_face_corner = quad_face_corner.face_corner2
            start_pt = _edge_vert_to_transition_pts[quad_face_corner.edge_vert1].first
            end_pt = _edge_vert_to_transition_pts[quad_face_corner.edge_vert2].first

            face_loops = data_manager.face_to_face_loops[start_face_corner.face]
            loop_vertex = face_loops.get_loop_vertex(start_face_corner)
            face_loops.set_override_pts(loop_vertex, start_pt)

            face_loops = data_manager.face_to_face_loops[end_face_corner.face]
            loop_vertex = face_loops.get_loop_vertex(end_face_corner)
            face_loops.set_override_pts(loop_vertex, end_pt)

            data_manager.add_force_smoothness_edge_create(EdgeCreateData.new(start_pt, end_pt, soft:true, smooth: true ))

            # We need to track if both ends of the diagonal_edge is being affected by this case.
            diagonal_edge_to_quad_curve_start_end_face_corners[quad_face_corner.quad_diagonal_edge_vert.edge] << [quad_face_corner, [start_pt, end_pt], start_face_corner, end_face_corner]
          else
            # Should always be one offset pt on a FaceCorner
            transition_segment = bevel_data.transitions_at_face_corner(quad_face_corner).first
            next if transition_segment.nil?

            if transition_segment.point_curve?
              # POint curve case. Similar to Single edge case.
              curve = transition_segment.curve
              if curve.start_pt.on_line?(quad_face_corner.edge_vert1.edge_line)
                start_face_corner = quad_face_corner.face_corner1
                end_face_corner = quad_face_corner.face_corner2
              else
                start_face_corner = quad_face_corner.face_corner2
                end_face_corner = quad_face_corner.face_corner1
              end

              face_loops = data_manager.face_to_face_loops[start_face_corner.face]
              loop_vertex = face_loops.get_loop_vertex(start_face_corner)
              face_loops.set_override_pts(loop_vertex, curve.start_pt)

              face_loops = data_manager.face_to_face_loops[end_face_corner.face]
              loop_vertex = face_loops.get_loop_vertex(end_face_corner)
              face_loops.set_override_pts(loop_vertex, curve.end_pt)

              # We need to track if both ends of the diagonal_edge is being affected by this case.
              diagonal_edge_to_quad_curve_start_end_face_corners[quad_face_corner.quad_diagonal_edge_vert.edge] << [quad_face_corner, curve.points.clone, start_face_corner, end_face_corner]
            elsif quad_face_corner.primary_secondary? && corner.edge_verts.length < 3
              face_corner_offset = transition_segment.get_offset_on_face_corner(quad_face_corner)
              offset_pt = face_corner_offset.offset_pt

              quad_face_corner.inner_face_corners.each do |inner_face_corner|
                # @type [FaceCorner]
                inner_face_corner = inner_face_corner

                face_loops = data_manager.face_to_face_loops[inner_face_corner.face]
                loop_vertex = face_loops.get_loop_vertex(inner_face_corner)

                if inner_face_corner.edge_verts.include?(_secondary_edge_verts.first)
                  # The curve falls on that inner_face_corner, so add the curve.
                  offset_pts = transition_segment.curve.points.clone
                  offset_pts.reverse! if self.point_on_line?(loop_vertex.position, loop_vertex.prev_loop_vertex.position, offset_pts.last)
                else
                  offset_pts = [offset_pt]
                end

                face_loops.set_override_pts(loop_vertex, offset_pts)
              end
            else
              face_corner_offset = transition_segment.get_offset_on_face_corner(quad_face_corner)
              offset_pt = face_corner_offset.offset_pt

              # Shift the diagonal by setting the loop_vertex at this corner to be the same as the offset pt
              quad_face_corner.inner_face_corners.each do |inner_face_corner|
                # @type [FaceCorner]
                inner_face_corner = inner_face_corner

                face_loops = data_manager.face_to_face_loops[inner_face_corner.face]
                loop_vertex = face_loops.get_loop_vertex(inner_face_corner)
                face_loops.set_override_pts(loop_vertex, offset_pt)
              end
            end
          end
        end
      end

      # For the case where a curve is being merged onto the quad,
      diagonal_edge_to_quad_curve_start_end_face_corners.each_pair do |_, quad_curve_start_end_face_corners|
        if quad_curve_start_end_face_corners.length == 1
          # Convert the faces to the opposite side of the diagonal.
          quad_face_corner, curve_pts, start_face_corner, end_face_corner = *quad_curve_start_end_face_corners[0]
          expected_normal = quad_face_corner.face.normal
          (curve_pts.length - 1).times do |c_i|
            pt1 = curve_pts[c_i]
            pt2 = curve_pts[c_i + 1]
            face = c_i < curve_pts.length / 2 ? start_face_corner.face : end_face_corner.face

            end_pt = quad_face_corner.quad_diagonal_edge_vert.end_pt
            opposite_edge_vert = quad_face_corner.quad_diagonal_edge_vert.opposite_edge_vert
            unless opposite_edge_vert.nil?
              # If the oppsite end of the edge vert is offset due to a edge being beveled, we have to grab that override point.
              oppsite_face_corner = opposite_edge_vert.face_corners.first
              face_loops = data_manager.face_to_face_loops[oppsite_face_corner.face]
              loop_vertex = face_loops.get_loop_vertex(oppsite_face_corner)
              override_pts = face_loops.loop_vertex_to_override_pts[loop_vertex]
              end_pt = override_pts.first unless override_pts.empty?
            end

            edge_creates = points_to_edge_creates([pt1, pt2, end_pt], true, true)
            face_create = FaceCreateData.new(edge_creates, expected_normal, face)
            data_manager.add_face_create_datas(face_create)
          end
        elsif quad_curve_start_end_face_corners.length == 2
          # We have to join both ends of the diagonal
          quad_face_corner1, curve_pts1, start_face_corner1, end_face_corner1 = *quad_curve_start_end_face_corners[0]
          quad_face_corner2, curve_pts2, start_face_corner2, end_face_corner2 = *quad_curve_start_end_face_corners[1]

          curve_pts2.reverse! if start_face_corner1.same_face?(end_face_corner2)
          expected_normal = quad_face_corner1.face.normal
          (curve_pts1.length - 1).times do |c_i|
            curve1_pt1 = curve_pts1[c_i]
            curve1_pt2 = curve_pts1[c_i + 1]
            curve2_pt1 = curve_pts2[c_i]
            curve2_pt2 = curve_pts2[c_i + 1]
            face = c_i < curve_pts1.length / 2 ? start_face_corner1.face : end_face_corner1.face

            face_pts = [curve1_pt1, curve1_pt2, curve2_pt2, curve2_pt1]
            if points_planar?(face_pts)
              edge_creates = points_to_edge_creates(face_pts, true, true)
              face_create = FaceCreateData.new(edge_creates, expected_normal, face)
              data_manager.add_face_create_datas(face_create)
            else
              tri_pts = [curve1_pt1, curve1_pt2, curve2_pt2]
              edge_creates = points_to_edge_creates(tri_pts, true, true)
              face_create = FaceCreateData.new(edge_creates, expected_normal, face)
              data_manager.add_face_create_datas(face_create)

              tri_pts = [curve2_pt2, curve2_pt1, curve1_pt1]
              edge_creates = points_to_edge_creates(tri_pts, true, true)
              face_create = FaceCreateData.new(edge_creates, expected_normal, face)
              data_manager.add_face_create_datas(face_create)

              # make it a quad
              edge_create = EdgeCreateData.new(curve1_pt1, curve2_pt2, smooth: true, soft: true, casts_shadows: false)
              data_manager.add_force_smoothness_edge_create(edge_create)
            end
          end
        end
      end


      # Coplanar cases for the border related cases and regular patch corners.
      bevel_data.affected_corners.each do |_, corner|
        # TODO: Jun 9 2023: Bug where the coplanar clean up will not clean up the edges of a v-curve (if the v-curve falls on the coplanar edge)
        _edge_vert_to_transition_pts = bevel_data.get_edge_vert_to_transition_pts(corner)
        _secondary_edge_verts = _edge_vert_to_transition_pts.keys

        corner.face_corners.each do |cur_face_corner|
          next unless cur_face_corner.instance_of?(CoplanarFaceCorner)

          # Get the border pts related to this face_corner. Always start from cur_face_corner.edge_vert1 and end at edge_vert2
          border_pts = []
          ignore_border_start_end = false # Used at the end to enforce edge smoothness
          is_point_curve = false
          if cur_face_corner.all_secondary? && corner.edge_verts.length == 3 && corner.num_primary_edges == 1
            next if bevel_data.is_one_transition_curved_into_self?(corner)

            # In a three edge corner, we select one edge to be beveled. This face corner is the one that doesn't have the any primary edges.
            transition_segment = bevel_data.transitions_at_corner(corner).first
            curve_pts = transition_segment.curve.points
            curve_pts.reverse! unless curve_pts.first.on_line?(cur_face_corner.edge_vert1.edge_line)

            border_pts << cur_face_corner.edge_vert1.opposite_vertex.position
            border_pts.concat(curve_pts)
            border_pts << cur_face_corner.edge_vert2.opposite_vertex.position
            ignore_border_start_end = true
          elsif cur_face_corner.all_secondary? && corner.v_corner? &&  _secondary_edge_verts.length == 2 && (_secondary_edge_verts[0].face_corners & _secondary_edge_verts[1].face_corners).first == cur_face_corner
            # A v-curve is merging into this coplanar_face_corner. Add the very last edge of the v-curve
            line_pts = _edge_vert_to_transition_pts.values.flatten.reverse
            line_pts.reverse! unless line_pts.first.on_line?(cur_face_corner.edge_vert1.edge_line)
            border_pts.concat(line_pts)
          else
            # Normal border case.
            transition_segment1 = bevel_data.transition_at_edge_vert(cur_face_corner.edge_vert1)
            unless transition_segment1.nil?
              if transition_segment1.point_curve?
                if cur_face_corner.all_secondary?
                  # Only time we need to consider the border of a point_curve is if its an all_secondary face corner.
                  border_pts.concat(transition_segment1.curve.points)
                  is_point_curve = true
                end
              else
                border_pts.concat(transition_segment1.face_to_opposite_start_end_pts[cur_face_corner.face].reverse) # Note the .reverse
              end
            end

            if cur_face_corner.three_point_corner?
              # Add the three point curve to the border.
              curve_pts = bevel_data.get_three_point_transition_at_face_corner(cur_face_corner).curve.points.clone
              curve_pts.reverse! unless curve_pts.first == border_pts.last
              border_pts.concat(curve_pts)
            end

            transition_segment2 = bevel_data.transition_at_edge_vert(cur_face_corner.edge_vert2)
            unless transition_segment2.nil?
              if transition_segment2.point_curve?
                # If we already joined with a pointcurve, don't add the other.
                if !is_point_curve && cur_face_corner.all_secondary?
                  border_pts.concat(transition_segment2.curve.points)
                  is_point_curve = true
                end
              else
                border_pts.concat(transition_segment2.face_to_opposite_start_end_pts[cur_face_corner.face])
              end
            end
          end
          border_pts = unique_array(border_pts)

          next if border_pts.empty?

          # DB.draw_line2d(border_pts, 5) # TODO: Debug

          # Find the intersection pts.
          index_to_int_pts = Hash.new {|h,k| h[k] = []} # Used later to enforce the softness of the border (yes, it's a hack.)
          cur_face_corner.inner_face_corners.each do |coplanar_face_corner|
            # @type [FaceCorner]
            coplanar_face_corner = coplanar_face_corner
            next if coplanar_face_corner.instance_of?(FakeFaceCorner)
            next if cur_face_corner.edge_verts_parallel? && cur_face_corner.primary_secondary? && coplanar_face_corner.all_secondary?

            edge_vert_to_int_pt_with_index = {}
            coplanar_face_corner.edge_verts.each do|coplanar_edge_vert|
              next unless coplanar_edge_vert.coplanar?

              (border_pts.length - 1).times do |c_i|
                cur_pt = border_pts[c_i]
                next_pt = border_pts[c_i + 1]
                curve_line = [cur_pt, cur_pt.vector_to(next_pt).normalize]

                int_pt = Geom.intersect_line_line(coplanar_edge_vert.edge_line, curve_line)
                next if int_pt.nil?
                next unless self.point_on_line?(coplanar_edge_vert.start_pt, coplanar_edge_vert.end_pt, int_pt)
                next unless self.point_on_line?(cur_pt, next_pt, int_pt)

                edge_vert_to_int_pt_with_index[coplanar_edge_vert] = [int_pt, c_i]
                index_to_int_pts[c_i] << int_pt

                break unless int_pt == coplanar_face_corner.vertex.position # Keep checking for other int_pts if it falls on the vertex.position.
              end
            end

            # Figure out if this coplanar_edge_vert has any edge_verts that are parellel with it's parent's face_corner's edge_verts (cur_face_corner)
            # and if we had any interception pts on those parellel edge_verts.
            # Only applies to all_secondary face_corners at this moment.
            parallel_edge_vert_to_parent_edge_vert = {}
            if coplanar_face_corner.all_secondary?
              edge_vert_to_int_pt_with_index.each_pair do |edge_vert, _|
                cur_face_corner.edge_verts.each do |parent_edge_vert|
                  next unless parent_edge_vert.edge_vector.parallel?(edge_vert.edge_vector)

                  parallel_edge_vert_to_parent_edge_vert[edge_vert] = parent_edge_vert
                end
              end
            end
            face_loops = data_manager.face_to_face_loops[coplanar_face_corner.face]
            loop_vertex = face_loops.get_loop_vertex(coplanar_face_corner)

            single_segment_border_pt = nil # A super specific case for single_segment bevels
            if edge_vert_to_int_pt_with_index.length == 1
              # Start/end of coplanar face_corners
              int_pt, pt_index = *edge_vert_to_int_pt_with_index.values.first
              if edge_vert_to_int_pt_with_index[coplanar_face_corner.edge_vert1].nil?
                # This is the start (so edge_vert2 would have the int_pt)
                if border_pts.length <= 2
                  offset_pts = [int_pt]
                  if transition_segment1.nil? && int_pt != border_pts.first
                    offset_pts << border_pts.first
                  elsif !transition_segment1.nil? && transition_segment1.point_curve?
                    offset_pts << border_pts.first
                  end
                  if !parallel_edge_vert_to_parent_edge_vert[coplanar_face_corner.edge_vert2].nil?
                    # int_pt lies on the parallel edge, so need to rejoin with the corner.vertex.
                    offset_pts << corner.vertex.position
                  elsif coplanar_face_corner.all_secondary? && self.point_on_face?(coplanar_face_corner.face, border_pts.first, false)
                    # the border ends inside this coplanar_face_corner's face. So join with corner.vertex.
                    offset_pts << corner.vertex.position
                  end
                else
                  offset_pts = []
                  if is_point_curve
                    offset_pts.concat(border_pts[0..pt_index])
                  else
                    unless pt_index == 0
                      offset_pts.concat(border_pts[1..pt_index])
                      single_segment_border_pt = border_pts[0] if bevel_data.params.single_segment?
                    end
                  end
                  offset_pts << int_pt
                end
              elsif edge_vert_to_int_pt_with_index[coplanar_face_corner.edge_vert2].nil?
                # End of coplanar face_corners (so edge_vert1 would have the int_pt)
                if border_pts.length <= 2
                  offset_pts = [int_pt]
                  if transition_segment2.nil? && int_pt != border_pts.last
                    offset_pts << border_pts.last
                  elsif !transition_segment2.nil? && transition_segment2.point_curve?
                    offset_pts << border_pts.last
                  end
                  if !parallel_edge_vert_to_parent_edge_vert[coplanar_face_corner.edge_vert1].nil?
                    # int_pt lies on the parallel edge, so need to rejoin with the corner.vertex.
                    offset_pts.insert(0, corner.vertex.position)
                  elsif coplanar_face_corner.all_secondary? && self.point_on_face?(coplanar_face_corner.face, border_pts.last, false)
                    # the border ends inside this coplanar_face_corner's face. So join with corner.vertex.
                    offset_pts << corner.vertex.position
                  end
                else
                  offset_pts = [int_pt]
                  if is_point_curve
                    offset_pts += border_pts[(pt_index + 1)..(border_pts.length - 1)]
                  else
                    offset_pts += border_pts[(pt_index + 1)..(border_pts.length - 2)]
                    single_segment_border_pt = border_pts.last if bevel_data.params.single_segment?
                  end
                end
              else
                raise UnexpectedStateError.new("", loop_vertex)  # TODO: Logic check. Is this possible?
              end

              if !single_segment_border_pt.nil? && bevel_data.params.single_segment? && offset_pts.length > 1 && offset_pts.include?(loop_vertex.position)
                if loop_vertex.parallel?
                  # In this case, if the offset_pt.first is the pt at the vertex, then it's not in order.
                  vec = single_segment_border_pt.vector_to(offset_pts.first).normalize
                  offset_pts.reverse! unless loop_vertex.prev_line[1].parallel?(vec)
                else
                  # Case for single segment bevels where there are two offset pts but one lies on the corner.vertex. So need a way to sort them properly.
                  if offset_pts.first == loop_vertex.position
                    vec = single_segment_border_pt.vector_to(offset_pts.last).normalize
                    offset_pts.reverse! if loop_vertex.prev_line[1].parallel?(vec)
                  else
                    vec = single_segment_border_pt.vector_to(offset_pts.first).normalize
                    offset_pts.reverse! if loop_vertex.next_line[1].parallel?(vec)
                  end
                end
              else
                offset_pts.reverse! if offset_pts.length > 1 && !loop_vertex.are_start_end_ordered?(offset_pts)
              end
            elsif edge_vert_to_int_pt_with_index.length == 2
              int_pt1, pt_index1 = *edge_vert_to_int_pt_with_index[coplanar_face_corner.edge_vert1]
              int_pt2, pt_index2 = *edge_vert_to_int_pt_with_index[coplanar_face_corner.edge_vert2]
              offset_pts = [int_pt1]
              if pt_index2 > pt_index1
                offset_pts.concat(border_pts[(pt_index1 + 1)..pt_index2])
              end
              offset_pts << int_pt2

              offset_pts.reverse! unless loop_vertex.are_start_end_ordered?(offset_pts)
            else
              # No collision
              offset_pts = []

              # At the start/end face_corners of a CoplanarFaceCorner, we need to add these pts if nothing collided.
              unless coplanar_face_corner.all_secondary?
                if transition_segment1.nil?
                  offset_pts = [corner.vertex.position, border_pts.first]
                  offset_pts.reverse! if loop_vertex.prev_vertex == transition_segment2.shared_edge_vert.opposite_vertex
                elsif transition_segment2.nil?
                  offset_pts = [border_pts.last, corner.vertex.position]
                  offset_pts.reverse! if loop_vertex.next_vertex == transition_segment1.shared_edge_vert.opposite_vertex
                else
                  if bevel_data.params.single_segment?
                    # TODO: May 9, 2023: Still a eztra edge here due to limitations of how i'm recreating faces. Leaving for nw.
                    # No Collision, yet the bevel is 'pushed' away from the coplanar edges.
                    int_pt_to_edge_vert_with_index = {}
                    (border_pts.length - 1).times do |b_i|
                      b_pt1 = border_pts[b_i]
                      b_pt2 = border_pts[b_i + 1]
                      border_line = [b_pt1, b_pt1.vector_to(b_pt2)]
                      # Note we are using cur_face_corner, which is the parent of coplanar_face_corner...
                      cur_face_corner.edge_verts.each do |edge_vert|
                        next unless edge_vert.primary?

                        int_pt = Geom.intersect_line_line(border_line, edge_vert.edge_line)
                        next if int_pt.nil?
                        next unless self.point_on_line?(b_pt1, b_pt2, int_pt)
                        next unless self.point_on_line?(edge_vert.start_pt, edge_vert.end_pt, int_pt)

                        offset_pts << int_pt
                        int_pt_to_edge_vert_with_index[int_pt] = [edge_vert, b_i]
                      end
                    end

                    if offset_pts.length == 2
                      # We should have two points. So get the mid point and use that as our 'joining' point.
                      edge_vert, b_i = *int_pt_to_edge_vert_with_index[offset_pts.first]
                      other_edge_vert, other_b_i = *int_pt_to_edge_vert_with_index[offset_pts.last]

                      mid_pt = self.linear_combination(offset_pts[0], offset_pts[1])
                      offset_pts = [mid_pt]
                      if coplanar_face_corner.edge_verts.include?(edge_vert)
                        if edge_vert.edge == loop_vertex.prev_edge
                          offset_pts.insert(0, border_pts[b_i])
                          offset_pts << corner.vertex.position
                        else
                          offset_pts.insert(0, corner.vertex.position)
                          offset_pts << border_pts[b_i + 1]
                        end
                      else
                        if other_edge_vert.edge == loop_vertex.prev_edge
                          offset_pts.insert(0, border_pts[other_b_i])
                          offset_pts << corner.vertex.position
                        else
                          offset_pts.insert(0, corner.vertex.position)
                          offset_pts << border_pts[other_b_i + 1]
                        end
                      end
                    else
                      # Not sure this is even possible
                    end
                  else
                    # This seems to occur when the edge_offset is too large and we end up with some messed up geometry.
                    # So just ignore for now?
                  end
                end
              end
            end

            # Nothing's changed, so move on.
            next if offset_pts.empty?

            face_loops.set_override_pts(loop_vertex, offset_pts)
          end

          if !index_to_int_pts.empty? && !bevel_data.params.single_segment?
            # Make sure all the boundary edges (with coplanar) are softened.
            all_border_pts = border_pts.clone
            # Since we are inserting into an array, the index will change after the first insert. So reverse it and
            # start from the highest index to avoid any issues.
            index_to_int_pts.to_a.reverse.each do |index_with_int_pts|
              index = index_with_int_pts[0]
              int_pts = unique_array(index_with_int_pts[1])
              all_border_pts.insert(index + 1, *int_pts)
            end
            all_border_pts = unique_array(all_border_pts)
            all_border_pts = all_border_pts[1..all_border_pts.length - 2] if ignore_border_start_end
            data_manager.add_force_smoothness_edge_create(points_to_edge_creates(all_border_pts, true, false))
          end
        end
      end

      # Create the reconstructed faces based off the above.
      # Note, the override_pts are sorted based on the logic above. So all we need to do is collect them and create the FaceCreateData
      data_manager.face_to_face_loops.each_pair do |face, face_loops|
        face_loops.get_loop_to_new_face_pts.each_pair do |loop, new_face_pts|
          # If there are not three or more unique pts, then skip this face (IE: don't recreate it)
          next if unique_array(new_face_pts).length < 3
          # Must be planar to recreate. And If not, then tessallate it
          if self.points_planar?(new_face_pts)
            # If the face 'collaspes' into itself, then don't recreate it.
            next unless self.pts_form_valid_face?(new_face_pts, face.normal)

            edge_create_datas = points_to_edge_creates(new_face_pts, false, true)
            is_hole = face_loops.is_loop_a_hole?(loop)
            data_manager.add_face_create_datas(FaceCreateData.new(edge_create_datas, face.normal, face, is_hole: is_hole, copy_dictionary: true))
          else
            # This issue (as of June 21, 2023) only happens with quads. And Quads can't have holes in them. So, keep it simple and don't worry about inner_loops.
            # SU 2020 and up support only.
            next unless Geom.respond_to?(:tesselate)
            next if face_loops.is_loop_a_hole?(loop)

            tess_pts = Geom.tesselate(new_face_pts)
            tess_pts.each_slice(3) do |tri_pts|
              # If the face 'collaspes' into itself, then don't recreate it.
              next unless self.pts_form_valid_face?(new_face_pts)

              edge_create_datas = points_to_edge_creates(tri_pts, true, true)
              data_manager.add_face_create_datas(FaceCreateData.new(edge_create_datas, face.normal, face, is_hole: false, copy_dictionary: true))
            end
          end

        end
      end

      # All regenerated faces become hardened (default SU Edge values). Below will copy over any properties (hidden, visible, layer, material, etc)
      # that existing edges might have.
      data_manager.face_to_face_loops.each_pair do |face, face_loops|
        face_loops.loop_to_loop_vertices.each_pair do |loop, loop_vertices|
          loop_vertices.length.times do |lv_i|
            cur_loop_vertex = loop_vertices[lv_i]
            next_edge = cur_loop_vertex.next_edge
            next if bevel_data.primary_edges.include?(next_edge)

            next_loop_vertex = loop_vertices[lv_i + 1]
            next_loop_vertex = loop_vertices[0] if next_loop_vertex.nil?

            cur_override_pts = face_loops.loop_vertex_to_override_pts[cur_loop_vertex]
            next_override_pts = face_loops.loop_vertex_to_override_pts[next_loop_vertex]

            cur_pt = cur_override_pts.empty? ? cur_loop_vertex.position : cur_override_pts.last
            next_pt = next_override_pts.empty? ? next_loop_vertex.position : next_override_pts.first
            next if cur_pt == next_pt

            data_manager.add_force_smoothness_edge_create(EdgeCreateData.new(cur_pt, next_pt, edge: next_edge, copy_dictionary: true ))
          end
        end
      end

      # Create the corner to corner faces
      # Track which corner_to_corner we've created faces for to avoid duplicating the faces.
      _handled_corner_to_corner = {}
      bevel_data.each_transition.each do |transition_segment|
        next if transition_segment.overlapping_with_opposite?

        shared_edge_vert = transition_segment.shared_edge_vert
        cur_corner = shared_edge_vert.corner
        if shared_edge_vert.primary?
          opposite_corner = shared_edge_vert.opposite_edge_vert.corner
          next unless _handled_corner_to_corner[[cur_corner, opposite_corner]].nil?
        end

        line_pts = []
        transition_segment.loop_with_opposite do |pt1, pt2|
          line_pts << [pt1, pt2]
        end

        if bevel_data.params.single_segment?
          raise UnexpectedStateError.new("", cur_corner) if line_pts.length != 2 # TODO: Logic chceck

          # We must harden the edges in single_segment cases.
          # Note: Can't set the smoothness in the below as the non-planar case will have the diagonal harden. (or different texture case.)
          line_pts1 = line_pts[0]
          line_pts2 = line_pts[1]
          pts = [line_pts1[0], line_pts1[1], line_pts2[1], line_pts2[0]]
          edge_create_datas = points_to_edge_creates(pts, false, true)
          data_manager.add_force_smoothness_edge_create(edge_create_datas)

          if transition_segment.face_properties_different?
            # If the face properties (materials/visible/etc.) are different between the faces, then split the face in the middle.
            mid_line = [self.linear_combination(line_pts1[0], line_pts2[0]), self.linear_combination(line_pts1[1], line_pts2[1])]
            line_pts.insert(1, mid_line)
          end
        end

        expected_normal = Geom::Vector3d.new
        shared_edge_vert.faces.each { |face| expected_normal += face.normal }
        expected_normal.normalize!

        faces = transition_segment.faces
        (line_pts.length - 1).times do |i|
          line1 = line_pts[i]
          line2 = line_pts[i + 1]
          pts = [line1[0], line1[1], line2[1], line2[0]]
          # The face to get the material from (if any)
          face = (i < line_pts.length / 2) ? faces[0] : faces[1]
          face = get_valid_face(face)

          if points_planar?(pts)
            edge_create_datas = points_to_edge_creates(pts, true)
            data_manager.add_face_create_datas(FaceCreateData.new(edge_create_datas, expected_normal, face))
          else
            # Triangulate it
            tri_pts = [line1[0], line1[1], line2[1]]
            edge_create_datas = points_to_edge_creates(tri_pts, true)
            edge_create_datas.last.properties.make_quad_diagonal!
            data_manager.add_face_create_datas(FaceCreateData.new(edge_create_datas, expected_normal, face))
            tri_pts = [line1[0], line2[0], line2[1]]
            edge_create_datas = points_to_edge_creates(tri_pts, true)
            edge_create_datas.last.properties.make_quad_diagonal!
            data_manager.add_face_create_datas(FaceCreateData.new(edge_create_datas, expected_normal, face))
          end
        end

        next unless shared_edge_vert.primary?

        _handled_corner_to_corner[[cur_corner, opposite_corner]] = true
        _handled_corner_to_corner[[opposite_corner, cur_corner]] = true
      end

      # Create the patch now .
      bevel_data.corner_to_patch.each_pair do |corner, corner_patch|
        corner_patch.transition_segment_to_point_curves.each_pair do |ts, all_pts|
          # @type [TransitionSegment]
          ts = ts
          faces = ts.faces
          (all_pts.length - 1).times do |i|
            pts1 = all_pts[i]
            pts2 = all_pts[i + 1]
            inner_pts = ts.inner_points_from_start
            (inner_pts.length + 2).times do |pt_i|
              face = (pt_i < pts1.length / 2) ? faces[1] : faces[0]
              face = get_valid_face(face)

              is_last = pt_i == (inner_pts.length + 1)
              if is_last
                pts = [pts1[pt_i + 1], pts2[pt_i] , pts2[pt_i + 1]]
              else
                pts = [pts1[pt_i], pts2[pt_i + 1] , pts1[pt_i + 1]]
              end

              expected_normal = calculate_expected_face_normal(corner, pts)
              edge_create_datas = points_to_edge_creates(pts, true)
              edge_create_datas.first.properties.make_quad_diagonal! # Make them quads
              data_manager.add_face_create_datas(FaceCreateData.new(edge_create_datas, expected_normal, face))

              if is_last
                pts = [pts1[pt_i], pts2[pt_i], pts1[pt_i + 1]]
              else
                pts = [pts1[pt_i], pts2[pt_i], pts2[pt_i + 1]]
              end
              edge_create_datas = points_to_edge_creates(pts, true)
              # Make them quads
              if is_last
                edge_create_datas[1].properties.make_quad_diagonal!
              else
                edge_create_datas.last.properties.make_quad_diagonal!
              end
              data_manager.add_face_create_datas(FaceCreateData.new(edge_create_datas, expected_normal,face))
            end
          end
        end

        center_pt = corner_patch.center_patch.center_pt
        surrounding_pts = corner_patch.center_patch.surrounding_pts

        if bevel_data.params.single_segment?
          # Need to harden the edges that surround the patch.
          edge_create_datas = points_to_edge_creates(surrounding_pts, false)
          data_manager.add_force_smoothness_edge_create(edge_create_datas)
        end

        if bevel_data.params.single_segment? && !corner_patch.center_patch.face_properties_different? && self.points_planar?(surrounding_pts)
          # Very specific single segment case.
          expected_normal = calculate_expected_face_normal(corner, surrounding_pts)
          edge_create_datas = points_to_edge_creates(surrounding_pts, true)
          data_manager.add_face_create_datas(FaceCreateData.new(edge_create_datas, expected_normal))
        else
          is_single_segment_and_different_materials = bevel_data.params.single_segment? && corner_patch.center_patch.face_properties_different?
          ts_i = 0
          corner_patch.center_patch.transition_segment_to_surrounding_pts.each_pair do |ts, surrounding_pts|
            surrounding_pts = surrounding_pts.clone

            if is_single_segment_and_different_materials
              # We need to split the polygon so the material gets applied evenly.
              mid_pt = self.linear_combination(surrounding_pts.first, surrounding_pts.last)
              surrounding_pts.insert(1, mid_pt)
            end

            (surrounding_pts.length - 1).times do |i|
              pt1 = surrounding_pts[i]
              pt2 = surrounding_pts[i + 1]
              pt2 = surrounding_pts[0] if pt2.nil?

              face = (i >= surrounding_pts.length / 2) ? ts.faces[1] : ts.faces[0]
              face = get_valid_face(face)
              polygon = [pt1, center_pt, pt2]
              expected_normal = calculate_expected_face_normal(corner, polygon)
              edge_create_datas = points_to_edge_creates(polygon, true)
              data_manager.add_face_create_datas(FaceCreateData.new(edge_create_datas, expected_normal, face))

              # Make every even one a quad_diagonal
              if bevel_data.params.single_segment?
                # different counter for single segment.
                if ts_i.even?
                  edge_create_datas.first.properties.make_quad_diagonal!
                else
                  edge_create_datas[1].properties.make_quad_diagonal!
                end
              else
                if i.even?
                  edge_create_datas.first.properties.make_quad_diagonal!
                else
                  edge_create_datas[1].properties.make_quad_diagonal!
                end
              end
            end
            ts_i += 1
          end
        end

      end

      # Dont soften if single bevel.
      # If smooth_perimeter is false, then we will set the border (perimeter) edges to smooth = false (soft = true)
      should_soft = !bevel_data.params.single_segment?
      should_smooth = should_soft
      bevel_data.each_transition.each do |transition_segment|
        transition_segment.face_to_opposite_start_end_pts.values.each do |pt, opposite_pt|
          next if pt.distance(opposite_pt).round(PLUGIN::FLOAT_PRECISION).zero?

          edge_create = EdgeCreateData.new(pt, opposite_pt, soft: should_soft, smooth: should_smooth)
          edge_create.properties.dictionary_props[:perimeter_edge] = true
          data_manager.add_force_smoothness_edge_create(edge_create)
        end

        transition_segment.adjacent_transition_segments.each do |adjacent_transition_segment|
          next unless adjacent_transition_segment.three_point_segment?

          edge_creates = points_to_edge_creates(adjacent_transition_segment.curve.points, should_soft, false)
          edge_creates.each do |edge_create|
            edge_create.properties.smooth = should_smooth
            edge_create.properties.dictionary_props[:perimeter_edge] = true
          end
          data_manager.add_force_smoothness_edge_create(edge_creates)
        end
      end

      data_manager
    # rescue
    #   # TODO: Debug
    #   draw_data_manager_preview(data_manager) unless data_manager.nil?
    #   raise
    end

    # @param ents [Sketchup::Entities]
    # @param data_manager [BevelerDataManager]
    def create_geometry(ents, data_manager, no_erase = false)
      # Need to create the object first as we have to copy the face materials (if any).
      grp = ents.add_group
      edge_to_properties = EntitiesGenerator.bulk_add_faces(grp.entities, data_manager)

      unless no_erase
        edges_faces_to_delete = []
        # Sometimes, when deleting a coplanar edge along with non-coplanar edges, SU will delete the coplanar and 'merge' the non-coplanar edges
        # together. So you are left with an edge that should be there. To avoid this stupid issue, Collect all the edges that are to be deleted,
        # delete as normal, and then do another check to make sure there are no dangling edges left.
        coplanar_edges = []
        data_manager.face_to_face_loops.each_pair do |face, face_loops|
          next unless face_loops.is_override?

          face_loops.get_loop_to_affected_edges.each do |_, affected_edges|
            affected_edges.each do |edge|
              # Sometimes, SU will explode the curve but sometimes it wont.
              # For a consistent result, always explode the curves we are beveling into.
              edge.explode_curve unless edge.curve.nil?
              if self.coplanar_edge?(edge)
                # We have to grab all the edges that join with that coplanar edge at the vertices.
                coplanar_edges.concat(edge.vertices.flat_map{|vert|vert.edges})
              else
                edges_faces_to_delete << edge
              end
            end
          end
          edges_faces_to_delete << face
        end
        ents.erase_entities(edges_faces_to_delete)

        # Special coplanar delete logic (described above.)
        coplanar_edges.select! { |edge| edge.valid? } # Somehow, some of these edges might get deleted.
        unless coplanar_edges.empty?
          beveler_ents = BevelerEntities.new(coplanar_edges)
          coplanar_edge_pts = coplanar_edges.map {|edge| [edge.start.position, edge.end.position]}

          # Keep deleting the coplanar edges until we are certain they are all gone.
          i = 0
          while true
            i += 1
            raise UnexpectedStateError.new("Infinite Loop hit when deleting coplanar edges") if i > 100

            dangling_edges = []
            coplanar_edge_pts.each do |edge_pts|
              potential_edges = beveler_ents.find_edges_by_pt_vec(edge_pts[0], edge_pts[1])
              potential_edges.each do | edge|
                dangling_edges << edge if edge.faces.length.zero?
              end
            end
            break if dangling_edges.empty?

            ents.erase_entities(dangling_edges.uniq)
          end
        end
      end

      border_ents = grp.explode
      EntitiesGenerator.cleanup_quads(border_ents, edge_to_properties.values)
    end

    # @param points [Array<Geom::Point3d>]
    # @param soften_smooth [Boolean]
    # @param should_loop [Boolean]
    # @return [Array<EdgeCreateData>]
    def points_to_edge_creates(points, soften_smooth, should_loop = true)
      raise UnexpectedStateError.new("Must provide at least two points", points) if points.length <= 1
      raise UnexpectedStateError.new("Not enough points to loop (need at least 3)", points) if should_loop && points.length < 3

      edge_creates = []
      num_pts = points.length
      num_pts -= 1 unless should_loop
      num_pts.times do |pt_i|
        cur_pt = points[pt_i]
        next_pt = points[pt_i + 1]
        next_pt = points[0] if next_pt.nil?
        next if cur_pt == next_pt

        edge_creates << EdgeCreateData.new(cur_pt, next_pt, soft: soften_smooth, smooth: soften_smooth)
      end

      edge_creates
    end

    # Use the center_vector and the calculated plane vector to get an expected_vector.
    # @param corner [Corner]
    # @param face_pts [Array<Geom::Point3d>]
    # @return [Geom::Vector3d]
    def calculate_expected_face_normal(corner, face_pts)
      center_vec = corner.center_vector
      f_vec = get_plane_from_points(face_pts).last
      f_vec.reverse! unless similar_direction?(f_vec, center_vec)
      (center_vec +  f_vec).normalize
    end

    # @param face [Sketchup::Face, FakeFace, QuadFace]
    # @return [Sketchup::Face, Nil]
    def get_valid_face(face)
      return face if face.instance_of?(Sketchup::Face)
      return face.sub_faces.first if face.instance_of?(QuadFace)

      nil
    end

  end

end