
module DM::Bevel

  class BevelCurve

    def reverse
      raise NotImplementedError
    end

    # @return [Geom::Point3d]
    def control_pt
      raise NotImplementedError
    end

    # @return [Geom::Point3d]
    def start_pt
      @points.first
    end

    # @return [Geom::Point3d]
    def end_pt
      @points.last
    end

    # @return [Array<Geom::Point3d>]
    def points
      @points
    end

    def initialize
      # @type [Array<Geom::Point3d>]
      @points = []
    end

    # @return [Integer]
    def length
      @points.length
    end

    # @param index [Integer]
    # @return [Geom::Point3d]
    def [](index)
      @points[index]
    end

    # @return [Geom::Point3d]
    def mid_pt
      raise UnexpectedStateError.new("Number of points is not an odd number", self) if @points.length.even?

      @points[@points.length / 2]
    end

    # @param other_curve [BevelCurve]
    # @param both_are_touching [Boolean]
    # @return [Boolean]
    def are_ends_touching?(other_curve, both_are_touching = false)
      start_touching = self.start_pt == other_curve.start_pt || self.start_pt == other_curve.end_pt
      end_touching = self.end_pt == other_curve.start_pt || self.end_pt == other_curve.end_pt
      if both_are_touching
        return true if start_touching && end_touching
      else
        return true if start_touching || end_touching
      end
      false
    end

    # @return [PointCurve]
    def to_point_curve
      PointCurve.new(self.points)
    end

    def inspect
      self.to_s
    end

  end


  class PointCurve < BevelCurve


    # @param points [Array<Geom::Point3d>]
    # @param use_cache [Boolean] Since this class is tracking points by reference, if the points were changed via pt.set! then that would
    #                            would invalidate this cache. We have no way of tracking if pt.set! was called so for now only cache if
    #                            explicitly told to.
    def initialize(points, use_cache = false)
      # @type [Array<Geom::Point3d>]
      @points = points
      # @type [Boolean]
      @use_cache = use_cache
      # @type [Hash{Integer => Float}] Total Distance at Index
      @cached_index_to_total_distance = {}
    end

    # @return [PointCurve]
    def reverse
      PointCurve.new(@points.reverse)
    end

    # @return [Geom::Point3d]
    def control_pt
      if @points.length == 2
        Geom.linear_combination(0.5, @points.first, 0.5, @points.last)
      else
        # Not really a control point, but needed for stuff to work.
        self.mid_pt
      end
    end

    # @return [Float]
    def curve_length
      distance_at_index(@points.length - 1)
    end

    # @param index [Integer]
    # @return [Float]
    def distance_at_index(index)
      raise UnexpectedStateError.new('Index must be a value of 0 or @points.length - 1', self) if index < 0 || index >= @points.length
      if @use_cache
        return @cached_index_to_total_distance[index] unless @cached_index_to_total_distance[index].nil?
      end

      dist = 0.0
      @cached_index_to_total_distance[0] = dist if @use_cache
      (0..index - 1).each do |i|
        dist += @points[i].distance(@points[i + 1]).to_f
        @cached_index_to_total_distance[i + 1] = dist if @use_cache
      end
      dist
    end

    # @param index [Integer]
    # @return [Float]
    def segment_length(index)
      raise UnexpectedStateError.new('Index must be a value of 0 or @points.length - 1', self) if index < 0 || index >= @points.length

      if index.zero?
        @points[0].distance(@points[1]).to_f
      else
        @points[index].distance(@points[index - 1]).to_f
      end
    end

    # @param t_val [Float]
    # @return [Geom::Point3d]
    def point_at_t(t_val)
      raise UnexpectedStateError.new("t_val must be between 0 and 1: (#{t_val})", self) if t_val < 0.0 || t_val > 1.0
      return @points.last if t_val == 1.0

      pt_dist = self.curve_length * t_val
      dist = 0.0
      (@points.length - 1).times do |i|
        pt1 = @points[i]
        pt2 = @points[i + 1]
        vec = pt1.vector_to(pt2)
        vec_dist = vec.length.to_f

        if pt_dist < (dist + vec_dist)
          vec.length = pt_dist - dist
          return pt1 + vec
        else
          dist += vec_dist
        end
      end
      nil
    end
  end

  class BezierCurve < BevelCurve

    # @return [Geom::Point3d]
    def start_pt
      @start_pt
    end

    # @return [Geom::Point3d]
    def end_pt
      @end_pt
    end

    # @return [Geom::Point3d]
    def control_pt
      @control_pt
    end

    # @return [Array<Geom::Point3d>]
    def points
      # Make a copy to prevent the the points from changing if modified from the outside.
      @points.map {|pt| pt.clone }
    end

    # @param index [Integer]
    # @return [Geom::Point3d]
    def [](index)
      @points[index].clone
    end

    # @param start_pt [Geom::Point3d]
    # @param end_pt [Geom::Point3d]
    # @param control_pt1 [Geom::Point3d]
    # @param segments [Integer]
    def initialize(start_pt, end_pt, control_pt1, segments)
      raise UnexpectedStateError.new("pt1 and pt2 cannot be the same.", self) if start_pt == end_pt

      @start_pt = start_pt
      @end_pt = end_pt
      @control_pt = control_pt1
      @segments = segments.to_i

      # @type [Array<Geom::Point3d>]
      @points = calculate_bezier(@start_pt, @end_pt, @control_pt, @segments)
    end

    # @return [Geom::Point3d]
    def mid_pt
      super.clone
    end

    # @return [BezierCurve]
    def reverse
      BezierCurve.new(@end_pt, @start_pt, @control_pt, @segments)
    end

    private

    # @param pt1 [Geom::Point3d]
    # @param pt2 [Geom::Point3d]
    # @param c_pt [Geom::Point3d]
    # @param segments [Integer]
    # @return [Array<Geom::Point3d>]
    def calculate_bezier(pt1, pt2, c_pt, segments)
      # https://javascript.info/bezier-curve
      # bezier curve
      pt1_arr = pt1.to_a
      pt2_arr = pt2.to_a
      c_arr = c_pt.to_a
      bezier_cb = lambda do |t|
        x_t = (((1.0 - t) ** 2.0) * pt1_arr.x) + (2.0 * (1.0 - t) * t * c_arr.x) + ((t ** 2.0) * pt2_arr.x)
        y_t = (((1.0 - t) ** 2.0) * pt1_arr.y) + (2.0 * (1.0 - t) * t * c_arr.y) + ((t ** 2.0) * pt2_arr.y)
        z_t = (((1.0 - t) ** 2.0) * pt1_arr.z) + (2.0 * (1.0 - t) * t * c_arr.z) + ((t ** 2.0) * pt2_arr.z)
        Geom::Point3d.new(x_t, y_t, z_t)
      end

      curve_pts = []
      t_seg = 1.0 / segments
      (segments + 1).times do |i|
        t = t_seg * i
        curve_pts << bezier_cb.call(t)
      end
      curve_pts
    end

  end


end
