
module DM::Bevel

  class BaseCorner

    # @return [Sketchup::Vertex]
    attr_reader :vertex

    # @return [Array<EdgeVertex>]
    def edge_verts
      @edge_verts.clone
    end

    # @param class_filter [Class]
    # @return [Array<FaceCorner]
    def face_corners(class_filter = nil)
      if class_filter.nil?
        @face_corners.clone
      else
        @face_corners.grep(class_filter)
      end
    end

    # @return [Array<Corner>] A reference to the corners next to this corner.
    def next_corners
      @next_corners.clone
    end

    class CurveState
      CONCAVE = 1
      CONVEX = 2
      MIX = 3
    end


    # @param vertex [Sketchup::Vertex]
    # @param primary_edges [Array<Sketchup::Edge>]
    # @param data [Hash]
    def initialize(vertex, primary_edges = [], **data)
      @vertex = vertex
      edge_verts = get_ordered_edge_verts(vertex, primary_edges, data)
      raise UnexpectedStateError.new("No edge_verts found", vertex) if edge_verts.empty? # TODO: Logic check

      # Create the face_corners array in a loop that "circles" around a corner. IE, the next face_corner is adjacent the the current face_corner, etc.
      # @type [Array<FaceCorner>]
      face_corners = [] # FYI: this is ordered as edge_verts was already ordered.
      _tracked_parallel_faces = []
      edge_verts.each_with_index do |edge_vert, ev_i|
        # @type [EdgeVertex]
        edge_vert = edge_vert
        next_edge_vert = edge_verts[ev_i + 1]
        next_edge_vert = edge_verts[0] if next_edge_vert.nil?

        if edge_vert.shares_fake_face?(next_edge_vert)
          fake_face = edge_vert.faces.find { |face| face.instance_of?(FakeFace) }
          face_corner = FakeFaceCorner.new(edge_vert, next_edge_vert, fake_face)
        elsif !edge_vert.parallel?(next_edge_vert)
          face_corner = FaceCorner.new(edge_vert, next_edge_vert)
          # TODO: Logic check. Make sure this face_corner actually has a face.
          raise UnexpectedStateError.new("FaceCorner created but EdgeVertexs do not share a face", vertex, face_corner) if face_corner.faces.empty?
        else
          related_faces = ((edge_vert.faces & next_edge_vert.faces) - _tracked_parallel_faces)
          face = nil
          if related_faces.length == 1 || edge_verts.length == 2
            face = related_faces.first
          else
            # Make sure we're grabbing the right face by checking if the edge_vert is adjacent to the next_edge_vert on that face.
            related_faces.each do |related_face|
              edgeuse = edge_vert.edgeuses(related_face).first
              next unless [edgeuse.next.edge, edgeuse.previous.edge].include?(next_edge_vert.edge)

              face = related_face
              break
            end
          end
          raise UnexpectedStateError.new("Attempted to make ParallelFaceCorner but failed to find face.", vertex) if face.nil?

          face_corner = ParallelFaceCorner.new(edge_vert, next_edge_vert, face)
          _tracked_parallel_faces << face
        end

        face_corners << face_corner
        face_corner.set_corner(self)  # Assign this Corner to the FaceCorner
        face_corner.edge_vert1.add_face_corner(face_corner)
        face_corner.edge_vert2.add_face_corner(face_corner)
      end

      # Find Quad FaceCorners
      quad_face_corners = []
      face_corners.each do |face_corner|
        # @type [FaceCorner]
        face_corner = face_corner
        quad_edge_vert = face_corner.edge_vert2
        next unless quad_edge_vert.quadface_diagonal?

        other_face_corner = quad_edge_vert.get_other_face_corner(face_corner)
        quad_face_corner = QuadFaceCorner.new(face_corner, other_face_corner)
        quad_face_corner.set_corner(face_corner.corner)

        # Override the referenced adjacent face_corners of the edge_verts
        quad_face_corner.edge_vert1.add_face_corner(quad_face_corner)
        quad_face_corner.edge_vert2.add_face_corner(quad_face_corner)
        quad_face_corners << quad_face_corner
      end

      # If there were any found, then insert them into face_corners array and delete the face_corners it encapsulates
      quad_face_corners.each do |quad_face_corner|
        fc_index = face_corners.find_index {|face_corner| face_corner.edge_vert2 == quad_face_corner.quad_diagonal_edge_vert}
        face_corners.insert(fc_index, quad_face_corner)
        quad_face_corner.inner_face_corners.each do |inner_face_corner|
          face_corners.delete(inner_face_corner)
        end
      end

      # face_corners.each_with_index do |face_corner, i|
      #   DB.draw_face_corner(face_corner) { |pt|
      #     DB.draw_pt_text(pt, i)
      #   }
      # end

      # Handle coplanar faces
      # Find the collection of FaceCorners that are coplanar and on a face that has the same normal (meaning they are 'the same')
      coplanar_face_corner_with_index_groups = []
      _handled_face_corners = {}
      face_corners.each_with_index do |face_corner, fc_i|
        next unless _handled_face_corners[face_corner].nil?
        _handled_face_corners[face_corner] = true

        next unless face_corner.contains_coplanar?

        coplanar_face_corners_with_index = [[face_corner, fc_i]]
        face_corners.each_with_index do |other_face_corner, o_fc_i|
          next unless _handled_face_corners[other_face_corner].nil?
          next unless other_face_corner.contains_coplanar?
          next unless face_corner.coplanar?(other_face_corner)

          coplanar_face_corners_with_index << [other_face_corner, o_fc_i]
          _handled_face_corners[other_face_corner] = true
        end

        coplanar_face_corner_with_index_groups << coplanar_face_corners_with_index if coplanar_face_corners_with_index.length > 1
      end

      unless coplanar_face_corner_with_index_groups.empty?
        # Although face_corners is sorted, that doesn't mean the sub coplanar face_corners will be sorted. It is possible that face_corner[0]
        # is the middle of a coplanar face_corner collection. So it could be something like indices [0, 1, 5, 6], where 1 is the end and 5 is the start.
        # In that case, we need to reorder so that the array looks like this [5, 6, 0, 1]
        coplanar_face_corner_with_index_groups.each_with_index do |face_corners_with_index, arr_index|
          last_index = nil
          face_corners_with_index.each_with_index do |face_corner_with_index, cur_index|
            fc_index = face_corner_with_index[1]

            if last_index.nil?
              last_index = fc_index
            else
              if fc_index - last_index > 1
                coplanar_face_corner_with_index_groups[arr_index] = face_corners_with_index[cur_index..face_corners_with_index.length - 1] + face_corners_with_index[0..cur_index - 1]
                break
              else
                last_index = fc_index
              end
            end
          end
        end

        coplanar_face_corner_with_index_groups.each do |face_corners_with_index|
          # @type [Array<FaceCorner>]
          face_corners_with_coplanar = face_corners_with_index.map { |face_corner_with_index| face_corner_with_index[0] }

          # If we're in a coplanar case, there should always be two or more FaceCorners
          # Reversed faces will mess with the coplanar check.
          next if face_corners_with_coplanar.length < 2

          # Set the adjacency of these coplanar FaceCorners. (They are only adjacent to each other, but not outside their Face)
          start_face_corner = face_corners_with_coplanar.first
          face_corners_with_coplanar.length.times do |fc_i|
            face_corner1 = face_corners_with_coplanar[fc_i - 1] unless fc_i == 0
            face_corner2 = face_corners_with_coplanar[fc_i + 1]

            start_face_corner.set_adjacent_face_corners(face_corner1, face_corner2)
            start_face_corner = face_corner2
          end

          # Nest these face_corners_with_coplanar in a CoplanarFaceCorner object.
          start_face_corner = face_corners_with_coplanar.first
          end_face_corner = face_corners_with_coplanar.last

          coplanar_face_corner = CoplanarFaceCorner.new(start_face_corner.edge_vert1, end_face_corner.edge_vert2, face_corners_with_coplanar)
          coplanar_face_corner.set_corner(self)
          # Override the referenced adjacent face_corners of the edge_verts
          coplanar_face_corner.edge_vert1.add_face_corner(coplanar_face_corner)
          coplanar_face_corner.edge_vert2.add_face_corner(coplanar_face_corner)

          # Now update the list of face_corners to remove the coplanar ones and add the CoplanarFaceCorner one
          cur_face_corner_index = face_corners.find_index(start_face_corner)
          face_corners.insert(cur_face_corner_index, coplanar_face_corner)

          face_corners_with_coplanar.each do |face_corner|
            face_corners.delete(face_corner)
          end

          # TODO: logic check to make sure the face_corners are still adjacent.
          (face_corners.length).times do |i|
            fc_1 = face_corners[i]
            fc_2 = face_corners[i + 1]
            fc_2 = face_corners[0] if fc_2.nil?

            raise UnexpectedStateError.new("FaceCorners are not adjacent", vertex, [fc_1, fc_2]) unless fc_1.adjacent?(fc_2)
          end
        end
      end
      raise UnexpectedStateError.new("Cannot have <2 FaceCorners.", vertex, face_corners) if face_corners.length < 2 # TODO: Logic check

      # face_corners.each_with_index do |face_corner, i|
      #   DB.draw_face_corner(face_corner) { |pt|
      #     DB.draw_pt_text(pt, face_corner.class)
      #   }
      # end

      # @edge_verts should not include any coplanar ones. The coplanar edge_verts can be accessed from the CoplanarFaceCorners
      # @edge_verts should also not include quad_diagonals. Those are accessed through the QuadFaceCorners.
      face_corners.each do |face_corner|
        if face_corner.instance_of?(CoplanarFaceCorner)
          face_corner.inner_edge_verts.each do |edge_vert|
            edge_verts.delete(edge_vert)
          end
        elsif face_corner.instance_of?(QuadFaceCorner)
          edge_verts.delete(face_corner.quad_diagonal_edge_vert)
        end
      end

      @edge_verts = edge_verts

      # Set the 'adjacency' of the face_corners
      @face_corners = face_corners
      start_face_corner = @face_corners.first
      @face_corners.length.times do |i|
        face_corner1 = @face_corners[i - 1]
        face_corner2 = @face_corners[i + 1]
        face_corner2 = @face_corners[0] if face_corner2.nil?
        start_face_corner.set_adjacent_face_corners(face_corner1, face_corner2)
        start_face_corner = face_corner2
      end

      @next_corners = []
      @cache_manager = CacheManager.new
    end

    # @param corner [Corner]
    def add_corner(corner)
      is_valid = false
      self.get_all_edge_verts.each do |edge_vert|
        corner.get_all_edge_verts.each do |other_edge_vert|
          if edge_vert.opposite?(other_edge_vert)
            is_valid = true
            break
          end
        end
        break if is_valid
      end

      raise UnexpectedStateError.new("corner is not next (across) to this corner", self, corner) unless is_valid

      @next_corners << corner
    end

    # Returns all faces at this corner (including the coplanar faces, if any.)
    # @return [Array<Sketchup::Face>]
    def faces
      self.face_corners.flat_map {|face_corner| face_corner.faces }
    end

    # @return [Integer]
    def num_primary_edges
      @edge_verts.count { |edge_vert| edge_vert.primary? }
    end

    # @return [Boolean]
    def all_primary?
      @edge_verts.each do |edge_vert|
        return false unless edge_vert.primary?
      end
      true
    end

    # @return [Array<EdgeVertex>]
    def primary_edge_verts
      @edge_verts.select { |edge_vert| edge_vert.primary? }
    end

    # @param edge_vert [EdgeVertex]
    # @return [Array<EdgeVertex>]
    def get_adjacent_edge_verts(edge_vert)
      raise UnexpectedStateError.new("edge_vert does not belong to this corner", self, edge_vert) if edge_vert.corner != self

      # FYI: Recall that @edge_verts is ordered.
      edge_vert_index = @edge_verts.find_index(edge_vert)
      adjacent_edge_verts = []
      adjacent_edge_verts << @edge_verts[edge_vert_index - 1]
      adjacent_edge_verts << @edge_verts[(edge_vert_index == @edge_verts.length - 1 ? 0 : edge_vert_index + 1)]
      adjacent_edge_verts
    end

    # If true, then a CornerPatch is needed for this corner
    # @return [Boolean]
    def requires_patching?
      self.num_primary_edges >= 3
    end

    # @return [Boolean]
    def is_parallel_joint?
      val = @cache_manager.retrieve(:is_parallel_joint) do
        non_coplanar_edge_verts = self.edge_verts.select { |edge_vert| !edge_vert.coplanar? }
        next false unless non_coplanar_edge_verts.length == 2

        non_coplanar_edge_verts.first.parallel?(non_coplanar_edge_verts.last)
      end
      val
    end

    # @return [Boolean]
    def contains_coplanar?
      self.face_corners.each do |face_corner|
        return true if face_corner.contains_coplanar?
      end
      false
    end

    # @return [Boolean]
    def contains_quad?
      self.face_corners.each do |face_corner|
        return true if face_corner.instance_of?(QuadFaceCorner)
      end
      false
    end

    # @return [Array<EdgeVertex>]
    def coplanar_edge_verts
      co_edge_verts = []
      self.face_corners.each do |face_corner|
        next unless face_corner.instance_of?(CoplanarFaceCorner)

        co_edge_verts.concat(face_corner.inner_edge_verts)
      end
      co_edge_verts.uniq
    end

    # Gets an unordered list of all EdgeVertex at this corner.
    # @return [Array<EdgeVertex>]
    def get_all_edge_verts
      val = @cache_manager.retrieve(:get_all_edge_verts) do
        edge_verts = []
        self.face_corners.each do |face_corner|
          if face_corner.respond_to?(:inner_face_corners)
            face_corner.inner_face_corners.each do |inner_face_corner|
              edge_verts.concat(inner_face_corner.edge_verts)
            end
          else
            edge_verts.concat(face_corner.edge_verts)
          end
        end
        edge_verts.uniq
      end
      val.clone
    end

    # @return [Geom::Vector3d]
    def center_vector
      val = @cache_manager.retrieve(:center_vector) do
        normal = Geom::Vector3d.new
        _used_normals = []
        self.face_corners.each do |face_corner|
          face_normal = face_corner.face.normal
          if _used_normals.empty?
            normal += face_normal
          else
            is_similar_vec = false
            _used_normals.each do |used_normal|
              ang = face_normal.angle_between(used_normal)
              if ang < 5.degrees # TODO: Degree tolerance
                is_similar_vec = true
                break
              end
            end
            next if is_similar_vec

            normal += face_normal
          end

          _used_normals << face_normal
        end

        normal.normalize
      end
      val.clone
    end

    # @return [Boolean]
    def v_corner?
      num_edge_verts = self.edge_verts.length
      return false if num_edge_verts < 3
      return false unless self.num_primary_edges == 2

      primary_edge_verts = self.primary_edge_verts
      edge_vert1 = primary_edge_verts[0]
      edge_vert2 = primary_edge_verts[1]
      if num_edge_verts == 3
        if edge_vert1.concave? == edge_vert2.concave?
          return false
        else
          return true
        end
      elsif num_edge_verts == 4
        if edge_vert1.get_shared_face_corner(edge_vert2).nil?
          return self.mix_curve?
        else
          return true
        end
      else
        # In this case, if the two primary edge_verts are more than two edge_verts away, then it is not a v-shape scenario.
        looped_edge_verts = self.edge_verts + self.edge_verts
        indices1 = []
        looped_edge_verts.each_with_index { |edge_vert, i| indices1 << i if edge_vert == edge_vert1 }
        indices2 = []
        looped_edge_verts.each_with_index { |edge_vert, i| indices2 << i if edge_vert == edge_vert2 }

        indices1.length.times do |i|
          index1 = indices1[i]
          indices2.length.times do |j|
            index2 = indices2[j]
            if (index1 - index2).abs < 3
              return true
            end
          end
        end
        return false
      end
    end

    # @return [Boolean]
    def mix_curve?
      self.curve_state == CurveState::MIX
    end

    # @return [Boolean]
    def all_concave?
      self.curve_state == CurveState::CONCAVE
    end

    # @return [Boolean]
    def all_convex?
      self.curve_state == CurveState::CONVEX
    end

    # @return [CurveState]
    def curve_state
      val = @cache_manager.retrieve(:curve_state) do
        next CurveState::MIX if self.contains_coplanar?

        all_concaves = []
        @edge_verts.each do |edge_vert|
          all_concaves << true if edge_vert.concave?
        end

        if all_concaves.length == @edge_verts.length
          CurveState::CONCAVE
        elsif all_concaves.length.zero?
          CurveState::CONVEX
        else
          CurveState::MIX
        end
      end
      val
    end

    # @param next_corner [Corner]
    # @return [EdgeVertex or Nil]
    def get_shared_edge_vert(next_corner)
      self.edge_verts.each do |edge_vert|
        next_corner.edge_verts.each do |next_edge_vert|
          return edge_vert if edge_vert.opposite?(next_edge_vert)
        end
      end
      nil
    end

    def inspect
      self.to_s
    end

    private

    # @param vertex [Sketchup::Vertex]
    # @param primary_edges [Array<Sketchup::Edge>]
    # @param data [Hash]
    # @return [Array<EdgeVertex>]
    def get_ordered_edge_verts(vertex, primary_edges, **data)
      raise NotImplementedError
    end

  end




  class Corner < BaseCorner

    private

    def get_ordered_edge_verts(vertex, primary_edges, **data)
      edge_verts = []
      vertex.edges.each do |edge|
        unless edge.faces.length == 2 || edge.faces.length == 1
          # FOR now, must have exactly one or two faces to be bevel-able.
          # puts "Ignoring edge: Only has #{edge.faces.length} faces."
          next
        end

        edge_vert = EdgeVertex.new(edge, @vertex, primary_edges.include?(edge))
        edge_vert.set_corner(self)
        edge_verts << edge_vert
      end

      # Order the edge_verts by adjacency.
      # Need to account for any edge_verts that have a missing face.
      start_edge_vert = edge_verts.first
      edge_verts.each do |edge_vert|
        # If there are any edge_verts with a missing face, always start at that edge_vert.
        next unless edge_vert.faces.length == 1

        start_edge_vert = edge_vert
      end

      # If there are any edge_verts with a missing face, it's possible that there are many 'fragmented edge_verts.
      fragmented_edge_verts = []
      ordered_edge_verts = [start_edge_vert]
      (edge_verts.length - 1).times do |_|
        remaining_edge_verts = (edge_verts - ordered_edge_verts)
        fragmented_edge_verts.each do |fragmented_evs|
          # These have been accounted for, so also remove them.
          remaining_edge_verts -= fragmented_evs
        end
        break if remaining_edge_verts.empty?

        if remaining_edge_verts.length == 1
          ordered_edge_verts << remaining_edge_verts.last
          fragmented_edge_verts << ordered_edge_verts
        else
          remaining_edge_verts.each do |other_edge_vert|
            next unless start_edge_vert.adjacent?(other_edge_vert)

            ordered_edge_verts << other_edge_vert
            start_edge_vert = other_edge_vert
            break
          end

          if start_edge_vert.faces.length == 1 && remaining_edge_verts.length > 1
            # If the next edge_vert (which is start_edge_vert in this context) only contains one face,
            # then that is the end of this fragment and the start of a new one.
            fragmented_edge_verts << ordered_edge_verts
            start_edge_vert = (remaining_edge_verts - [start_edge_vert]).find {|ev| ev.faces.length == 1}
            ordered_edge_verts = [start_edge_vert]
          end
        end
      end

      # fragmented_edge_verts.each do |evs|
      #   evs.each_with_index do |edge_vert, i|
      #     DB.draw_edge_vert(edge_vert) { |pt| DB.draw_pt_text(pt, i)}  #if edge_verts.length > 2
      #   end
      # end

      if fragmented_edge_verts.length == 1
        all_ordered_edge_verts = fragmented_edge_verts.first
      else
        # If there are multiple fragments, below will do its best to order them.
        # Below is made up logic. Use the closest fragment based on the smallest angle between the last edge_vert.
        cur_fragment_evs = fragmented_edge_verts.first
        ordered_fragmnet_evs = [cur_fragment_evs]
        (fragmented_edge_verts.length - 1).times do |_|
          remaining_fragment_evs = fragmented_edge_verts - ordered_fragmnet_evs
          break if remaining_fragment_evs.empty?

          end_edge_vert = cur_fragment_evs.last
          all_angle_with_startend_with_fragment_evs = []
          remaining_fragment_evs.each do |fragment_ev|
            # TODO: Fails if the 'visual' angle is greater than 180
            start_ang = end_edge_vert.edge_vector.angle_between(fragment_ev.first.edge_vector)
            all_angle_with_startend_with_fragment_evs << [start_ang, true, fragment_ev]
            end_ang = end_edge_vert.edge_vector.angle_between(fragment_ev.last.edge_vector)
            all_angle_with_startend_with_fragment_evs << [end_ang, false, fragment_ev]
          end

          closest_data = all_angle_with_startend_with_fragment_evs.sort {|a, b| a[0] <=> b[0]}.first
          closest_fragment = closest_data[2]
          closest_fragment.reverse! unless closest_data[1] # If we are joining with the end (false), then we need to reverse it.
          ordered_fragmnet_evs << closest_fragment

          cur_fragment_evs = closest_fragment
        end
        all_ordered_edge_verts = ordered_fragmnet_evs.flat_map {|evs| evs}
      end

      # all_ordered_edge_verts.each_with_index do |edge_vert, i|
      #   DB.draw_edge_vert(edge_vert) { |pt| DB.draw_pt_text(pt, i)} # if edge_verts.length > 2
      # end

      # TODO: Logic check
      raise UnexpectedStateError.new("Failed to sort edge_verts at corner", vertex, edge_verts) if all_ordered_edge_verts.length != edge_verts.length

      edge_verts = all_ordered_edge_verts

      # Check if any edge_verts have only one Face. If so, add a fake one.
      edge_verts.length.times do |ev_i|
        cur_edge_vert = edge_verts[ev_i]
        next_edge_vert = edge_verts[ev_i + 1]
        next_edge_vert = edge_verts[0] if next_edge_vert.nil?

        next if cur_edge_vert.faces.length == 2 || next_edge_vert.faces.length == 2
        next unless (cur_edge_vert.faces & next_edge_vert.faces).empty? # If they share a face, then skip.

        normal = cur_edge_vert.edge_vector * next_edge_vert.edge_vector

        cur_normal = cur_edge_vert.faces.first.normal
        next_normal = next_edge_vert.faces.first.normal
        if cur_normal.parallel?(normal)
          # if parallel, then that means this fake face is coplanar. So make the normal the same direction
          normal.reverse! unless normal.samedirection?(cur_normal)
        elsif next_normal.parallel?(normal)
          normal.reverse! unless normal.samedirection?(next_normal)
        else
          # Attempt to get this calculated normal to go in a similar direction as the other faces.
          # Grab the other edge_verts that connect this corner, and if this normal is moving in a similar direction as those,
          # then we assume its pointing inwards. Reverse it in that case.
          # TODO: If it's a concave or mixed corner, this normal will be reveresed.
          summed_vec = Geom::Vector3d.new
          (edge_verts - [cur_edge_vert, next_edge_vert]).each do |ev|
            summed_vec += ev.edge_vector
          end

          ang = summed_vec.angle_between(normal)
          normal.reverse! if ang <= Math::PI / 2.0
        end

        fake_face = FakeFace.new(cur_edge_vert.vertex, normal)
        cur_edge_vert.add_fake_face(fake_face)
        next_edge_vert.add_fake_face(fake_face)
      end

      edge_verts.each do |edge_vert|
        if edge_vert.faces.length != 2 # TODO: Logic check.
          # DB.draw_edge_vert(edge_vert)
          raise UnexpectedStateError.new("edge_vert does not have two faces", vertex, edge_vert)
        end
      end

      edge_verts
    end

  end



  class FragmentedCorner < BaseCorner

    private

    # @param data [Hash]
    #   @option data [Hash{Sketchup::Face => Integer}] :face_to_weight
    def get_ordered_edge_verts(vertex, primary_edges, **data)
      multi_face_edges = []
      vertex.edges.each do |v_edge|
        next unless v_edge.faces.length > 2

        multi_face_edges << v_edge
      end

      raise UnexpectedStateError.new("Must have an edge that contains >2 faces.", vertex) if multi_face_edges.empty?

      # Find the adjacent edges based on the edges that DON'T have more than two faces
      normal_edges = vertex.edges - multi_face_edges
      fragmented_edges = []
      normal_edges.each do |edge|
        sorted_edge = []
        edge.faces.each do |face|
          vertex.edges.each do |other_v_edge|
            next if edge == other_v_edge
            next unless face.edges.include?(other_v_edge)

            sorted_edge << other_v_edge
          end
        end
        sorted_edge.insert(1, edge)
        fragmented_edges << sorted_edge
      end

      # Order the edges now.
      ordered_edges = fragmented_edges.first.clone
      unconnected_fragments = []
      fragmented_edges[1..fragmented_edges.length - 1].each do |next_fragment|
        next if (next_fragment - ordered_edges).empty?

        if ordered_edges.first == next_fragment.first
          if ordered_edges.last == next_fragment.last
            ordered_edges.insert(0, next_fragment[1])
          else
            next_fragment[1..2].each do |n_edge|
              ordered_edges.insert(0, n_edge)
            end
          end
        elsif ordered_edges.first == next_fragment.last
          if ordered_edges.last == next_fragment.first
            ordered_edges.insert(0, next_fragment[1])
          else
            next_fragment[0..1].reverse.each do |n_edge|
              ordered_edges.insert(0, n_edge)
            end
          end

        elsif ordered_edges.last == next_fragment.first

          next_fragment[1..2].each do |n_edge|
            ordered_edges << n_edge
          end
        elsif ordered_edges.last == next_fragment.last
          next_fragment[0..1].reverse.each do |n_edge|
            ordered_edges << n_edge
          end
        else
          unconnected_fragments << next_fragment
        end

      end

      raise UnexpectedStateError.new('Unconnected fragment found.', vertex) unless unconnected_fragments.empty? # TODO: deal with later

      face_to_weight = data[:face_to_weight]
      raise UnexpectedStateError.new("Must set :face_to_weight option when using FragmentedCorner", vertex) if face_to_weight.nil? # TODO: Logic check

      edge_verts = []
      ordered_edges.each_with_index do |edge, e_i|
        # Sort the
        faces = edge.faces
                    .map { |face| [face, face_to_weight[face]]}
                    .sort { |a,b| a[1] <=> b[1]}.reverse # Needs to be descending.
                    .map {|f_data| f_data[0]}

        edge_vert = EdgeVertex.new(edge, vertex, primary_edges.include?(edge), faces)
        edge_vert.set_corner(self)

        # DB.draw_edge_vert(edge_vert) { |pt| DB.draw_pt_text(pt, e_i)}
        edge_verts << edge_vert
      end

      edge_verts
    end

  end



end