
module DM::Bevel

  class EdgeVertex
    include GeometryMixin

    # @return [Sketchup::Edge]
    attr_reader :edge
    # @return [Sketchup::Vertex]
    attr_reader :vertex
    # @return [EdgeVertex] A reference to the EdgeVertex opposite to this one. Nil, if there isn't one (meaning the opposite is not part of a Corner)
    attr_reader :opposite_edge_vert
    # @return [Corner]
    attr_reader :corner

    # @return [Array<FaceCorner>] A reference to the FaceCorners this edge_vertex belongs to.
    def face_corners
      @face_corners.clone
    end

    # @param edge [Sketchup::Edge]
    # @param vertex [Sketchup::Vertex]
    # @param is_primary [Boolean]
    # @param edge_faces [Array<Sketchup::Face>] Provides a way to sort the faces (for cases where edge.faces.length > 2)
    def initialize(edge, vertex, is_primary = false, edge_faces = nil)
      raise UnexpectedStateError, "edge does not contain vertex" unless edge.vertices.include?(vertex)
      # raise UnexpectedStateError, "edge must have one or two faces" if edge.faces.length < 1 || edge.faces.length > 2

      @edge = edge
      @vertex = vertex
      @is_primary = is_primary

      if edge_faces.nil?
        @faces = edge.faces
      else
        raise UnexpectedStateError.new("edge_faces.length do not match edge.faces", self) unless edge.faces.length == edge_faces.length
        edge.faces.each do |face|
          raise UnexpectedStateError.new("edge does not contain that face", self) unless edge_faces.include?(face)
        end
        @faces = edge_faces.clone
      end

      @face_corners = []
      @opposite_edge_vert = nil
      @corner = nil
      @cache_manager = CacheManager.new
    end

    # @param face_corner [FaceCorner]
    def add_face_corner(face_corner)
      raise UnexpectedStateError.new("face_corner does not contain this EdgeVertex", self, face_corner) if self != face_corner.edge_vert1 && self != face_corner.edge_vert2

      if self == face_corner.edge_vert1
        @face_corners[0] = face_corner
      elsif self == face_corner.edge_vert2
        @face_corners[1] = face_corner
      end
    end

    # @param opposite_edge_vert [EdgeVertex]
    def set_opposite_edge_vert(opposite_edge_vert)
      raise UnexpectedStateError.new("opposite_edge_vert cannot be nil",self, opposite_edge_vert) if opposite_edge_vert.nil?
      raise UnexpectedStateError.new("opposite_edge_vert is not opposite to this EdgeVertex",self, opposite_edge_vert)  unless self.opposite?(opposite_edge_vert)

      @opposite_edge_vert = opposite_edge_vert
    end

    # @param corner [Corner]
    def set_corner(corner)
      raise UnexpectedStateError.new("this edge_vert does not share the same vertex as the corner", self, corner) unless self.vertex == corner.vertex

      @corner = corner
    end

    # @param fake_face [FakeFace]
    def add_fake_face(fake_face)
      raise UnexpectedTypeError, "fake_face must be of type FakeFace" unless fake_face.instance_of?(FakeFace)
      raise UnexpectedStateError.new("This EdgeVertex already has two Faces.", self) if self.faces.length == 2

      @faces << fake_face
    end

    # @return [Boolean]
    def primary?
      @is_primary
    end

    # @return [Sketchup::Vertex]
    def opposite_vertex
      self.edge.other_vertex(self.vertex)
    end

    # @return [Geom::Vector3d]
    def edge_vector
      val = @cache_manager.retrieve(:edge_vector) do
        @vertex.position.vector_to(self.opposite_vertex.position).normalize
      end
      val.clone
    end

    # @return [Float]
    def edge_length
      val = @cache_manager.retrieve(:edge_length) do
        @vertex.position.distance(self.opposite_vertex.position).to_f # Never work with Sketchup Length's as they are slow
      end
      val
    end

    # @return [Array<Geom::Point3d, Geom::Vector3d>]
    def edge_line
      [@vertex.position, self.edge_vector]
    end

    # @return [Array<Sketchup::Face, FakeFace>]
    def faces
      @faces.clone
    end

    # @return[Array<Sketchup::Vertex>]
    def vertices
      [self.vertex, self.opposite_vertex]
    end

    # @return [Array<Geom::Point3d>]
    def points
      [self.start_pt, self.end_pt]
    end

    # @return [Geom::Point3d]
    def start_pt
      # Start point should always be from the context of @vertex.
      @vertex == @edge.start ? @edge.start.position : @edge.end.position
    end

    # @return [Geom::Point3d]
    def end_pt
      @vertex == @edge.start ? @edge.end.position : @edge.start.position
    end

    # @param edge_vert [EdgeVertex]
    def adjacent?(edge_vert)
      return false unless @vertex == edge_vert.vertex

      same_faces = (self.faces & edge_vert.faces)
      return false if same_faces.empty? # If they don't share a face, then they're not adjacent
      return true if same_faces.first.instance_of?(FakeFace)

      same_faces.each do |same_face|
        this_edge_use = self.edgeuses(same_face).first

        # If the next or previous edge is the edge_vert.edge, then they are adjacent.
        return true if [this_edge_use.next.edge, this_edge_use.previous.edge].include?(edge_vert.edge)
      end
      false
    end

    # Get the EdgeUses for this edge. EdgeUse contains the loop and face, so can get that info from there.
    # @param at_face [Sketchup::Face, Nil] Pass in nil if you want to get all EdgeUses for this EdgeVertex.
    # @return [Array<Sketchup::EdgeUse>]
    def edgeuses(at_face = nil)
      edge_uses = []
      self.faces.each do |face|
        next if !at_face.nil? && face != at_face

        loop = face.loops.length == 1 ? face.loops.first : face.loops.find {|loop| loop.edges.include?(@edge)}
        edge_uses << loop.edgeuses.find {|edge_use| edge_use.edge == @edge }
      end
      edge_uses
    end

    # @param edge_vert [EdgeVertex]
    def opposite?(edge_vert)
      self.opposite_vertex == edge_vert.vertex && @edge == edge_vert.edge
    end

    # https://forums.sketchup.com/t/concavity-problem-and-need-help/70589
    # @return [Boolean]
    def concave?
      val = @cache_manager.retrieve(:concave) do
        # TODO: Reversed faces will mess this function up.
        faces = self.faces
        vector = faces[0].normal * faces[1].normal
        vector.samedirection?(@edge.line[1]) == @edge.reversed_in?(faces[0])
      end
      val
    end

    # https://forums.sketchup.com/t/orienting-faces-and-the-api/83762/11 (At the very bottom, comment by CAUL)
    # @return [Boolean]
    def reversed_faces?
      return false if self.contains_fake_face?

      faces = self.faces
      self.edge.reversed_in?(faces[0]) == self.edge.reversed_in?(faces[1])
    end

    # @return [Boolean]
    def coplanar?
      val = @cache_manager.retrieve(:coplanar) do
        faces = self.faces
        self.coplanar_faces?(faces[0], faces[1])
      end
      val
    end

    # @return [Boolean]
    def quadface_diagonal?
      val = @cache_manager.retrieve(:quadface_diagonal) do
        !self.coplanar? && QuadFace.is_quadface_diagonal?(@edge) && QuadFace.is_quadface?(@edge.faces[0]) && QuadFace.is_quadface?(@edge.faces[1])
      end
      val
    end

    # @param other_edge_vert [EdgeVertex]
    # @return [Boolean]
    def parallel?(other_edge_vert)
      self.edge_vector.parallel?(other_edge_vert.edge_vector)
    end

    # @return [Boolean]
    def contains_fake_face?
      self.faces.each do |face|
        return true if face.instance_of?(FakeFace)
      end
      false
    end

    # @param other_edge_vert [EdgeVertex]
    # @return [Boolean]
    def shares_fake_face?(other_edge_vert)
      shared_face = (self.faces & other_edge_vert.faces).first
      return false if shared_face.nil?
      return false unless shared_face.instance_of?(FakeFace)

      true
    end

    # @param other_edge_vert [EdgeVertex]
    # @return [FaceCorner, Nil]
    def get_shared_face_corner(other_edge_vert)
      (self.face_corners & other_edge_vert.face_corners).first
    end

    # @param face_corner [FaceCorner]
    # @return [FaceCorner]
    def get_other_face_corner(face_corner)
      raise UnexpectedStateError.new("face_corner does not belong to this edge_vert", self, face_corner) unless self.face_corners.include?(face_corner)

      (self.face_corners - [face_corner]).first
    end

    # @param face [Sketchup::Face]
    # @return [Sketchup::Loop, Nil]
    def get_related_loop(face)
      return nil unless self.faces.include?(face)

      face.loops.each do |loop|
        return loop if loop.edges.include?(self.edge)
      end
      nil
    end

    # @param other_edge_vert [EdgeVertex]
    def ==(other_edge_vert)
      return false unless other_edge_vert.instance_of?(EdgeVertex)

      @edge == other_edge_vert.edge && @vertex == other_edge_vert.vertex
    end

    alias eql? :==

    def hash
      @edge.hash + @vertex.hash
    end

    def inspect
      self.to_s
    end

  end

end