
module DM::Bevel

  # edge_vert1 and edge_vert2 are the 'boundaries' (IE: they are not coplanar)
  # This class will contain a collection of face_corners that contain the coplanar things.
  class CoplanarFaceCorner < FaceCorner

    # @return [Array<FaceCorner>]
    def inner_face_corners
      @inner_face_corners.clone
    end

    # @param edge_vert1 [EdgeVertex]
    # @param edge_vert2 [EdgeVertex]
    # @param inner_coplanar_face_corners [Array<FaceCorner>]
    def initialize(edge_vert1, edge_vert2, inner_coplanar_face_corners)
      # Only allow coplanar start/end edge_verts if those edge_verts are adjacent to a QuadFaceCorner.
      if (edge_vert1.face_corners + edge_vert2.face_corners).find {|e_fc| e_fc.instance_of?(QuadFaceCorner)}.nil?
        raise UnexpectedStateError, "edge_vert1 cannot be coplanar" if edge_vert1.coplanar?
        raise UnexpectedStateError, "edge_vert2 cannot be coplanar" if edge_vert2.coplanar?
      end

      raise UnexpectedStateError, "inner_coplanar_face_corners cannot be empty" if inner_coplanar_face_corners.empty?
      super(edge_vert1, edge_vert2)

      normal = nil
      inner_coplanar_face_corners.each do |face_corner|
        normal = face_corner.face.normal if normal.nil?

        raise UnexpectedStateError, "inner_coplanar_face_corners does not contain a coplanar edge_vert" unless face_corner.contains_coplanar?
        raise UnexpectedStateError, "inner_coplanar_face_corners are not on the same face" if !normal.nil? && !normal.parallel?(face_corner.face.normal)
      end

      # @type [Array<FaceCorner>]
      @inner_face_corners = inner_coplanar_face_corners.clone
    end

    def contains_coplanar?
      true
    end

    # @return [Sketchup::Face]
    def face
      val = @cache_manager.retrieve(:face) do
        @inner_face_corners.find{|coplanar_face_corner| !coplanar_face_corner.instance_of?(FakeFaceCorner)}.face
      end
      val
    end

    # @return [Array<Sketchup::Face>]
    def faces
      val = @cache_manager.retrieve(:faces) do
        @inner_face_corners.map {|face_corner| face_corner.face}
      end
      val.clone
    end

    def face_between?
      val = @cache_manager.retrieve(:face_between) do
        normal = self.face.normal
        angle = angle_in_plane(self.edge_vert1.edge_vector, self.edge_vert2.edge_vector, normal)

        coplanar_angle = 0.0
        all_edge_verts = self.all_edge_verts
        (all_edge_verts.length - 1).times do |i|
          cur_edge_vert = all_edge_verts[i]
          next_edge_vert = all_edge_verts[i + 1]

          sub_angle = self.angle_in_plane(cur_edge_vert.edge_vector, next_edge_vert.edge_vector, normal)
          coplanar_angle += sub_angle.abs
        end

        angle.abs.round(PLUGIN::FLOAT_PRECISION) == coplanar_angle.abs.round(PLUGIN::FLOAT_PRECISION)
      end
      val
    end

    def mid_vector
      if !self.edge_verts_parallel?
        super
      else
        # case where edge_vert1 and edge_vert2 are parallel
        face_vec =  self.edge_vert2.edge_vector * self.face.normal
        # Make sure the vector is in the shared face direction.
        pt = self.vertex.position.offset(face_vec, 0.01)
        face_vec.reverse! unless is_pt_on_coplanar_faces?(pt)
        face_vec
      end
    end

    # @param edge_vert [EdgeVertex]
    # @return [EdgeVertex]
    def closest_coplanar_edge_vert(edge_vert)
      raise UnexpectedStateError.new("edge_vert does not belong to this FaceCorner", self, edge_vert) unless self.edge_verts.include?(edge_vert)

      is_edge_vert1 = self.edge_vert1 == edge_vert
      if is_edge_vert1
        closest_edge_vert = self.inner_face_corners.first.edge_vert2
      else
        closest_edge_vert = self.inner_face_corners.last.edge_vert1
      end
      closest_edge_vert
    end

    # @param edge_vert [EdgeVertex]
    # @return [FaceCorner]
    def closest_coplanar_face_corner(edge_vert)
      raise UnexpectedStateError.new("edge_vert does not belong to this FaceCorner", self, edge_vert) unless self.edge_verts.include?(edge_vert)

      if self.edge_vert1 == edge_vert
        closest_face_corner = self.inner_face_corners.first
      else
        closest_face_corner = self.inner_face_corners.last
      end
      closest_face_corner
    end

    # An ordered list of inner edge_verts. Does not include the boundary edge_verts.
    # @return [Array<EdgeVertex>]
    def inner_edge_verts
      # Since inner_face_corners is ordered, so will the edge_verts.
      edge_verts = []
      inner_face_corners = self.inner_face_corners
      (inner_face_corners.length - 1).times do |fc_i|
        edge_verts << inner_face_corners[fc_i].edge_vert2
      end
      edge_verts
    end

    # All edge_verts, including the 'boundary' edge_verts
    # @return [Array<EdgeVertex>]
    def all_edge_verts
      edge_verts = [self.edge_vert1]
      self.inner_face_corners.each do |inner_face_corner|
        edge_verts << inner_face_corner.edge_vert2
      end
      edge_verts
    end

    # @param other_face_corner [FaceCorner]
    # @return [Boolean]
    def ==(other_face_corner)
      return false unless super

      self.face == other_face_corner.face
    end

    alias eql? :==

    def hash
      new_hash = @edge_vert1.hash + @edge_vert2.hash
      @inner_face_corners.each do |face_corner|
        new_hash += face_corner.hash
      end
      new_hash
    end

    def inspect
      self.to_s
    end

    private

    # @param pt [Geom::Point3d]
    # @return [Boolean]
    def is_pt_on_coplanar_faces?(pt)
      self.inner_face_corners.each do |inner_face_corner|
        next if inner_face_corner.instance_of?(FakeFaceCorner)

        return true if self.point_on_face?(inner_face_corner.face, pt)
      end
      false
    end

  end

end