
module DM::Bevel

  class FaceCorner
    include GeometryMixin

    # @return [EdgeVertex]
    attr_reader :edge_vert1
    # @return [EdgeVertex]
    attr_reader :edge_vert2
    # @return [Corner] A reference to the Corner this FaceCorner belongs to.
    attr_reader :corner

    # @return [Array<FaceCorner, FaceCorner>]
    #   - Index 0: Adjacent to edge_vert1
    #   - Index 1: Adjacent to edge_vert2
    def adjacent_face_corners
      @adjacent_face_corners.clone
    end

    # @return [Array<FaceCorner, FaceCorner>]
    #   - Index 0: Next to edge_vert1
    #   - Index 1: Next to edge_vert2
    def next_face_corners
      @next_face_corners.clone
    end

    class State
      ALL_PRIMARY = 1
      PRIMARY_SECONDARY = 2
      ALL_SECONDARY = 3
    end

    # @param edge_vert1 [EdgeVertex]
    # @param edge_vert2 [EdgeVertex]
    def initialize(edge_vert1, edge_vert2)
      raise UnexpectedStateError, "edge_verts are not connected by the same vertex" unless edge_vert1.vertex == edge_vert2.vertex
      raise UnexpectedStateError, "edge_verts cannot be the same" if edge_vert1 == edge_vert2

      @edge_vert1 = edge_vert1
      @edge_vert2 = edge_vert2

      @corner = nil
      @adjacent_face_corners = [nil, nil]
      @next_face_corners = [nil, nil]

      @cache_manager = CacheManager.new
    end

    # @param corner [Corner]
    def set_corner(corner)
      raise UnexpectedStateError.new("corner does not share the same vertex as this FaceCorner", self, corner) unless corner.vertex == self.vertex

      @corner = corner
    end

    # @param prev_face_corner [FaceCorner]
    # @param next_face_corner [FaceCorner]
    def set_adjacent_face_corners(prev_face_corner, next_face_corner)
      raise UnexpectedStateError.new("prev_face_corner.edge_vert2 is not equal to self.edge_vert1", self, prev_face_corner) if !prev_face_corner.nil? && self.edge_vert1 != prev_face_corner.edge_vert2
      raise UnexpectedStateError.new("next_face_corner.edge_vert1 is not equal to self.edge_vert2", self, next_face_corner) if !next_face_corner.nil? && self.edge_vert2 != next_face_corner.edge_vert1

      @adjacent_face_corners = [prev_face_corner, next_face_corner]
    end

    # @param next_face_corner1 [FaceCorner]
    # @param next_face_corner2 [FaceCorner]
    def set_next_face_corners(next_face_corner1, next_face_corner2)
      raise UnexpectedStateError.new("self.edge_vert1 is not opposite to the edge_verts in next_face_corner1", self, next_face_corner1) if !next_face_corner1.nil? && !self.edge_vert1.opposite?(next_face_corner1.edge_vert1) && !self.edge_vert1.opposite?(next_face_corner1.edge_vert2)
      raise UnexpectedStateError.new("self.edge_vert2 is not opposite to the edge_verts in next_face_corner2", self, next_face_corner2) if !next_face_corner2.nil? && !self.edge_vert2.opposite?(next_face_corner2.edge_vert1) && !self.edge_vert2.opposite?(next_face_corner2.edge_vert2)

      @next_face_corners = [next_face_corner1, next_face_corner2]
    end

    # @return [Array<EdgeVertex>]
    def edge_verts
      [@edge_vert1, @edge_vert2]
    end

    # @return [Sketchup::Vertex]
    def vertex
      # FYI @edge_vert1.vertex == @edge_vert2.vertex
      @edge_vert1.vertex
    end

    # @return [Sketchup::Face]
    def face
      val = @cache_manager.retrieve(:face) do
        self.faces.first
      end
      val
    end

    # @return [Array<Sketchup::Edge>]
    def edges
      [self.edge_vert1.edge, self.edge_vert2.edge]
    end

    # @return [Sketchup::Loop]
    def loop
      self.edge_vert1.get_related_loop(self.face)
    end

    # Should always return one face unless it is a CoplanarFaceCorner
    # @return [Array<Sketchup::Face>]
    def faces
      val = @cache_manager.retrieve(:faces) do
        related_faces = (@edge_vert1.edge.faces & @edge_vert2.edge.faces)
        next related_faces if related_faces.length == 1

        # Pick the face that is between the edge_verts
        mid_vec = Geom.linear_combination(0.5, self.edge_vert1.edge_vector, 0.5, self.edge_vert2.edge_vector).normalize
        pt = self.vertex.position.offset(mid_vec, 0.01)
        related_faces.select! do |related_face|
          self.point_on_face?(related_face, pt)
        end

        raise UnexpectedStateError.new("faces should never be empty.", self) if related_faces.empty? # TODO: Logic check.
        related_faces
      end
      val.clone
    end

    # @return [Boolean] True if the edges in this FaceCorner are going from concave to convex, or vice-versa.
    def transition?
      @edge_vert1.concave? != @edge_vert2.concave?
    end

    # @return [Boolean]
    def three_point_corner?
      return false unless self.all_primary?
      return false if self.edge_verts_parallel?
      return false unless self.corner.mix_curve?
      return false if self.transition?
      self.adjacent_face_corners.each do |adjacent_face_corner|
        return false unless adjacent_face_corner.all_primary?
      end
      return false if self.face_between?

      true
    end

    # @return [Boolean]
    def contains_coplanar?
      self.edge_vert1.coplanar? || self.edge_vert2.coplanar?
    end

    # @param other_face_corner [FaceCorner]
    # @return [Boolean]
    def coplanar?(other_face_corner)
      return false unless self.vertex == other_face_corner.vertex

      self.coplanar_faces?(self.face, other_face_corner.face)
    end

    # @return [Boolean]
    def edge_verts_parallel?
      val = @cache_manager.retrieve(:edge_verts_parallel) do
        self.edge_vert1.edge_vector.parallel?(self.edge_vert2.edge_vector)
      end
      val
    end

    # If the below pt is classified as PointOutside, then that means the face is 'around' the face_corner (and not between).
    # @return [Boolean]
    def face_between?
      val = @cache_manager.retrieve(:face_between) do
        mid_vec = Geom.linear_combination(0.5, self.edge_vert1.edge_vector, 0.5, self.edge_vert2.edge_vector).normalize
        pt = self.vertex.position.offset(mid_vec, 0.01)
        self.point_on_face?(self.face, pt)
      end
      val
    end

    # Recursively check if the faces are the same.
    # @param other_face_corner [FaceCorner]
    # @return [Boolean]
    def same_face?(other_face_corner)
      if self.respond_to?(:inner_face_corners)
        self.inner_face_corners.each do |inner_face_corner|
          return true if inner_face_corner.same_face?(other_face_corner)
        end
      elsif other_face_corner.respond_to?(:inner_face_corners)
        other_face_corner.inner_face_corners.each do |inner_face_corner|
          return true if inner_face_corner.same_face?(self)
        end
      else
        return true if self.face == other_face_corner.face
      end
      return false
    end

    # @return [Boolean]
    def contains_edge_vert?(edge_vert)
      @edge_vert1 == edge_vert || @edge_vert2 == edge_vert
    end

    # The next face_corner at the end of the edge_vert.
    # @param face_corner [FaceCorner]
    # @return [Boolean]
    def next?(face_corner)
      # must be on the same face to be considered adjacent
      return false unless self.same_face?(face_corner)
      self.edge_verts.each do |edge_vert|
        face_corner.edge_verts.each do |other_edge_vert|
          return true if edge_vert.opposite?(other_edge_vert)
        end
      end
      false
    end

    # @param face_corner [FaceCorner]
    # @return [Boolean]
    def adjacent?(face_corner)
      return false unless self.vertex == face_corner.vertex

      # Note, if self == face_corner, then adjacent == true.
      # If there is a shared edge_vert, that means they're adjacent.
      !(self.edge_verts & face_corner.edge_verts).empty?
    end

    # @param face_corner [FaceCorner]
    # @return [EdgeVertex or Nil] Nil if there is no shared edge.
    def get_shared_edge_vert(face_corner)
      if adjacent?(face_corner)
        return face_corner.edge_verts.include?(self.edge_vert1) ? self.edge_vert1 : self.edge_vert2
      elsif next?(face_corner)
        if face_corner.edge_vert1.opposite?(self.edge_vert1) || face_corner.edge_vert2.opposite?(self.edge_vert1)
          return self.edge_vert1
        else
          return self.edge_vert2
        end
      end
      nil
    end

    # Gets the opposite EdgeVertex (IE: if edge_vert1 is provided, then this will return edge_vert2).
    # Nil if edge_vert does not belong to this FaceCorner.
    # @param edge_vert [EdgeVertex]
    # @return [EdgeVertex or Nil]
    def get_other_edge_vert(edge_vert)
      return nil unless self.edge_verts.include?(edge_vert)

      edge_vert == @edge_vert1 ? @edge_vert2 : @edge_vert1
    end

    # @return [Geom::Vector3d]
    def mid_vector
      val = @cache_manager.retrieve(:mid_vector) do
        vec = Geom.linear_combination(0.5, @edge_vert1.edge_vector, 0.5, @edge_vert2.edge_vector)
        vec.reverse! unless face_between?
        vec.normalize
      end
      val.clone
    end

    # @return [Array<EdgeVertex, EdgeVertex>] First index is the primary edge, second index is the secondary edge.
    def primary_secondary
      # TODO: This exception is more of a "you're using it wrong" check.
      raise UnexpectedStateError.new("FaceCorner not a PRIMARY_SECONDARY", self) unless self.state == State::PRIMARY_SECONDARY

      if @edge_vert1.primary?
        [@edge_vert1, @edge_vert2]
      else
        [@edge_vert2, @edge_vert1]
      end
    end

    # @return [Boolean]
    def all_primary?
      self.state == State::ALL_PRIMARY
    end

    # @return [Boolean]
    def all_secondary?
      self.state == State::ALL_SECONDARY
    end

    # @return [Boolean]
    def primary_secondary?
      self.state == State::PRIMARY_SECONDARY
    end

    # @return [State]
    def state
      if @edge_vert1.primary? && @edge_vert2.primary?
        State::ALL_PRIMARY
      elsif @edge_vert1.primary? != @edge_vert2.primary?
        State::PRIMARY_SECONDARY
      else
        State::ALL_SECONDARY
      end
    end

    # @param other_face_corner [FaceCorner]
    # @return [Boolean]
    def ==(other_face_corner)
      return false unless other_face_corner.is_a?(FaceCorner)

      return false unless self.hash == other_face_corner.hash
      true
    end

    alias eql? :==

    def hash
      @edge_vert1.hash + @edge_vert2.hash
    end

    def inspect
      self.to_s
    end

  end

end