module DM::Bevel

  # This class represents a QuadFace according to the standard developed by Thomthom which is:
  # If an edge is soft, smooth, and does not cast shadows, it is considered to be the shared edge
  # of two triangles that form a non-planar Quad.  In addition, planar quads without a diagonal always
  # considered to be Quadfaces
  class QuadFace

    # Create a new quadface in the model within the given entities collection. As a helpful side-effect,
    # it also sets the order of the new quad loop vertices to be the same as the order of the points that was passed
    # @param [Array<Geom::Point3d>] points array of 4 points that will form the quad in ccw order
    # @param [Sketchup::Entities] ents the entities collection to create the quad within
    # @param [Bool] force_non_planar forces a diagonal to be created even if the points are planar
    # @return [Quadface] returns a reference to the quad that was created
    def self.create(points, ents, force_non_planar = false)
      raise ArgumentError.new(LH["Must be exactly 4 points in array"]) if points.length != 4
      is_planar_quad = Common.quad_points_planar?(points)
      num_tries = 0
      begin
        num_tries += 1
        if is_planar_quad && !force_non_planar
          f = ents.add_face(points)
          check_and_flip_face(f, points)
          return QuadFace.get(f)
        else
          f1 = ents.add_face(points[0], points[1], points[3]) # could raise ArgumentError
          f2 = ents.add_face(points[1], points[2], points[3])
          check_and_flip_face(f1, points)
          check_and_flip_face(f2, points)
          shared_edge = (f1.edges & f2.edges)[0]
          if shared_edge
            set_diagonal(shared_edge)
          end
          return QuadFace.get(f1)
        end
      rescue ArgumentError # thrown if trying to create a triangle face with all points in a line
        force_non_planar = false
        is_planar_quad = true
        if num_tries < 2
          retry
        else
          #
        end
      end
    end

    # Finds all quads from the given collection
    # @param [Enumerable<Sketchup::Entity>] ents a collection of entities
    # @return [Array<QuadFace>] returns an array of quads
    def self.find_all(ents)
      faces = ents.grep(Sketchup::Face)
      qf_subfaces = Set.new
      quads = []
      faces.each do |f|
        next if qf_subfaces.include?(f)
        qf = QuadFace.get(f)
        if qf
          quads << qf
          qf_subfaces.merge(qf.sub_faces)
        end
      end
      quads

    end

    # Sets the properties for a quadface diagonal
    def self.set_diagonal(edge)
      edge.soft = true
      edge.smooth = true
      edge.casts_shadows = false
    end

    # Gets a QuadFace object from the given SketchUp face or nil if this face is not part of a QuadFace
    # @param [Sketchup::Face] su_face the SketchUP face that will be referenced by this Quadface
    # @param [Bool] store_all_entities for performance, this is set to false so that it will only be calculated as needed
    # @return [QuadFace]
    def self.get(su_face, store_all_entities = false)
      if is_quadface?(su_face)
        qf = QuadFace.new(su_face)
        if store_all_entities
          qf.vertices
          qf.edges
        end
        return qf
      end
      false
    end

    # @param edge [Sketchup::Edge]
    # @return [QuadFace]
    def self.get_from_diagonal(edge)
      return nil unless is_quadface_diagonal?(edge)

      QuadFace.new(edge.faces.first, edge)
    end

    # Returns the Quadfaces that are shared by the given edge
    # @param [Sketchup::Edge] edge the edge we are checking
    # @return [Array<QuadFace] returns an array of quadfaces (may be empty)
    def self.get_edge_quads(edge)
      if QuadFace.is_quadface_diagonal?(edge)
        return [QuadFace.get(edge.faces.first)]
      end
      edge_faces = edge.faces
      quads = []
      edge_faces.each do |ef|
        qf = QuadFace.get(ef)
        quads << qf if qf
      end
      quads
    end


    # Returns the unique Quadfaces that are shared by the given vertex
    # @param [Sketchup::Vertex] vert the vertex we are checking
    # @return [Array<QuadFace] returns an array of quadfaces (may be empty)
    def self.get_vertex_quads(vert)
      vert_faces = vert.faces
      quads = []
      vert_faces.each do |vf|
        qf = QuadFace.get(vf)
        # Check if this quad is already in the array which can happen if this vertex
        # is the termination of a quadface diagonal
        if qf
          match = false
          quads.each { |quad|
            match = true if quad == qf
          }
          quads << qf unless match
        end
      end
      quads
    end

    # Check whether this face is a QuadFace (since SketchUp does not support non-planar quads)
    # @param [Sketchup::Face] su_face the SketchUp face to be checked
    def self.is_non_planar_quadface?(su_face)
      num_verts = su_face.vertices.length
      if num_verts == 3
        diagonals = su_face.edges.find_all { |edge| is_quadface_diagonal?(edge) }
        # Only consider this a quad if there is exactly one of these diagonals for the given face
        if diagonals.length == 1
          return true
        end
      end
      false
    end

    # Check whether this face is a QuadFace (since SketchUp does not support non-planar quads, this should
    # always return true if the SketchUp face has 4 vertices)
    # @param [Sketchup::Face] su_face the SketchUp face to be checked
    def self.is_quadface?(su_face)
      num_verts = su_face.vertices.length
      return true if num_verts == 4
      if num_verts == 3
        diagonals = su_face.edges.find_all { |edge| is_quadface_diagonal?(edge) }
        # Only consider this a quad if there is exactly one of these diagonals for the given face
        if diagonals.length == 1
              return true
            end
          end
      false
    end

    # Check whether the given edge is considered to be the diagonal of a non-planar QuadFace
    # @param [Sketchup::Edge] edge
    def self.is_quadface_diagonal?(edge)
      # Check the casts shadows property first since it will almost always be set to true unless
      # set by Ruby QuadFace logic
      return false unless !edge.casts_shadows? && edge.smooth? && edge.soft?

      faces = edge.faces
      return false unless faces.length == 2

      face1 = faces[0]
      face2 = faces[1]
      return false unless face1.vertices.length == 3
      return false unless face2.vertices.length == 3

      # Don't allow multiple diagonals.
      (face1.edges + face2.edges).uniq.each do |f_edge|
        next if edge == f_edge

        return false if !f_edge.casts_shadows? && f_edge.smooth? && f_edge.soft? # Don't allow multiple diagonals.
      end

      # A quad with a diagonal can't be planar (based on QuadFace.new initialization)
      quad_pts = (face1.vertices + face2.vertices).uniq.map { |vert| vert.position }
      quad_plane = Geom.fit_plane_to_points(quad_pts)

      all_pts_fit = true
      quad_pts.each do |pt|
        next if pt.on_plane?(quad_plane)

        all_pts_fit = false
        break
      end

      return false if all_pts_fit

      true
    end

    # Check whether diagonal forms a valid quad - if not, set it so that it no longer is a quad-face diagonal
    def self.validate_diagonal(edge)
      if is_quadface_diagonal?(edge)
        # above is only true if the edge is considered a valid quad diagonal
      else
        edge.casts_shadows = true unless edge.casts_shadows?
      end
    end

    # Compares the normal of the SketchUp subface with the ordered array of quad (or triangle) points in order
    # to determine if the face should be flipped
    # @param [Sketchup::Face] face an actual SketchUp face
    # @param [Array<Point3D>] face_points ordered array (ccw) of points that form this outer loop of this quad
    def self.check_and_flip_face(face, face_points)
      v1 = face_points[1].vector_to(face_points[0])
      v2 = face_points[1].vector_to(face_points[2])
      quad_normal = v2.cross(v1)
      if quad_normal.dot(face.normal) < 0.0
        # sub face normal is in opposite direction of quad normal so flip the face
        face.reverse!
      end
    end

    # Create a QuadFace representation from the given SketchUp face
    # @param [Sketchup::Face] su_face the SketchUp face that forms half of the quad
    # @param [Sketchup::Edge] diagonal_edge optional edge on this face that will form the diagonal
    def initialize(su_face, diagonal_edge = nil)
      @vertices = nil
      @edges = nil
      num_verts = su_face.vertices.length
      case num_verts
      when 3
        diagonals = su_face.edges.find_all { |edge| QuadFace.is_quadface_diagonal?(edge) }
        # Only consider this a quad if there is exactly one of these diagonals for the given face
        if diagonals.length == 1
          @diagonal = diagonals[0]
          edge_faces = @diagonal.faces
          if edge_faces.length != 2
            raise ArgumentError.new(LH["QuadFace diagonal edge must have exactly two faces"])
          end
          @faces = edge_faces
          @is_planar = false
        else
          raise ArgumentError.new(LH["QuadFace has more than one diagonal"])
        end
      when 4
        @faces = [su_face]
        @is_planar = true
        @diagonal = nil
      else
        raise ArgumentError.new(LH["Face must have 3 or 4 vertices"])
      end
    end

    # Tests two quads for equality
    # @param [QuadFace] quad2
    def ==(quad2)
      return false unless quad2.is_a?(QuadFace)
      if self.is_planar?
        if quad2.is_planar?
          # If they are both planar, they will both contain one sub-face.  If this face
          # matches, the quads are the same
          return self.sub_faces[0] == quad2.sub_faces[0]
        else
          return false
        end
      else
        if quad2.is_planar?
          # this quad is not planar and quad2 is so they are not the same
          return false
        else
          # if both are non-planar with same diagonal, they must be the same quad
          return self.diagonal == quad2.diagonal
        end
      end
    end

    def sub_faces
      @faces.dup
    end

    def is_planar?
      @is_planar
    end

    def diagonal
      @diagonal
    end

    def vertices
      return @vertices if @vertices
      if @diagonal
        f0 = @faces[0]
        f1 = @faces[1]
        f0_loop_verts = f0.vertices
        f1_loop_verts = f1.vertices

        if @diagonal.reversed_in?(f0)
          end_vert_0 = @diagonal.start
          end_vert_1 = @diagonal.end
        else
          end_vert_0 = @diagonal.end
          end_vert_1 = @diagonal.start
        end

        far0 = f0_loop_verts[(f0_loop_verts.index(end_vert_0) + 1) % 3]
        far1 = f1_loop_verts[(f1_loop_verts.index(end_vert_1) + 1) % 3]

        # Determine the current winding order of the imaginary quad containing these two triangles
        @vertices = [far0, end_vert_1, far1, end_vert_0]
      else
        @vertices = @faces[0].outer_loop.vertices
      end
      @vertices
    end

    # Allows us to use the loops method to obtain an array of VertexLoops which have a similar interface to the
    # native SketchUp Loop class
    # @return [Array<VertexLoop>] this array will always have just one item which is the loop of the quad in ccw order
    def loops
      vl = VertexLoop.new(vertices)
      [vl]
    end

    # Returns an array of ordered edges around the perimeter of the quad
    def edges
      return @edges if @edges
      @edges = []
      vertices.each_index do |i|
        @edges << @vertices[i].common_edge(@vertices[(i + 1) % 4])
      end
      @edges
    end

    def all_connected
      @faces.first.all_connected
    end

    # Returns true if this quad has been deleted
    def deleted?
      is_deleted = @faces.find { |f| f.deleted? }
      is_deleted != nil
    end

    def valid?
      !deleted?
    end

    def select
      did_rebuild = false
      sel = Sketchup.active_model.selection
      begin
        @faces.each do |f|
          sel.add(f)
        end
      rescue TypeError
        # If the vertices of the quad-face were transformed, the stored edges and faces
        # will likely be changed to deleted entity types and will fail to be added to the selection
        unless did_rebuild
          rebuild_from_vertices
          did_rebuild = true
          retry
        end
      end
    end

    def deselect
      did_rebuild = false
      sel = Sketchup.active_model.selection
      begin
        @faces.each do |f|
          sel.remove(f)
        end
      rescue TypeError
        # If the vertices of the quad-face were transformed, the stored edges and faces
        # will likely be changed to deleted entity types and will fail to be added to the selection
        unless did_rebuild
          rebuild_from_vertices
          did_rebuild = true
          retry
        end
      end
    end

    def toggle_select
      sel = Sketchup.active_model.selection
      # QuadFace will only be considered selected if all sub-faces are selected
      not_in_selection = @faces.find { |f| !sel.contains?(f) }
      if not_in_selection
        select
      else
        deselect
      end
    end

    # If the vertices are still valid, rebuild the internals of the quad face
    def rebuild_from_vertices
      if @vertices
        # First ensure all vertices are valid
        invalid_vert = @vertices.find { |v| !v.valid? }
        raise LH["Invalid vertex"] if invalid_vert
        v0 = @vertices[0]
        v1 = @vertices[1]
        v2 = @vertices[2]
        v3 = @vertices[3]
        e0 = v0.common_edge(v1)
        e1 = v1.common_edge(v2)
        e2 = v2.common_edge(v3)
        e3 = v3.common_edge(v0)
        if e0 && e1 && e2 && e3
          f0 = e0.common_face(e1)
          f1 = e1.common_face(e2)
          f2 = e2.common_face(e3)
          f3 = e3.common_face(e0)
          common_faces = [f0, f1, f2, f3]
          common_faces.delete(nil)
          face_set = Set.new
          face_set.merge(common_faces)
          if face_set.length == 1
            # Check of all common edges faces reference the same face
            # If so, this must be a planar quad
            @is_planar = true
            @faces = [f0]
            @diagonal = nil
            @edges = nil # invalidate the edges so they will be recalculated next time they are needed
            return
          elsif face_set.length == 2
            @faces = face_set.to_a
            # Diagonal edge will be the intersection of the edges arrays of the two faces
            diagonal = (@faces[0].edges & @faces[1].edges).first
            unless QuadFace.is_quadface_diagonal?(diagonal)
              # Diagonal is not tagged correctly as a quad-face diagonal so fix it
              UI.messagebox(LH["Diagonal was not set"])
              QuadFace.set_diagonal(diagonal)
            end
            @diagonal = diagonal
            @edges = nil #invalidate the edges
            return
          elsif face_set.length == 0
            raise LH["Tried to rebuild quad face but no common faces were found"]
          else
            # More than 2 unique common faces were found for the given edges
            e0.material = "red"
            e1.material = "red"
            e2.material = "red"
            e3.material = "red"
            Sketchup.active_model.rendering_options["EdgeColorMode"] = 0
            raise LH["Tried to rebuild quad face but more than 2 common faces were found"]
          end
        else
          raise LH["Vertices did not have 4 common edges"]
        end
      else
        raise LH["Tried to rebuild quad face but vertices were nil"]
      end
    end

    # Sets the start vertex in the array of 4 perimeter vertices.
    def set_start_vertex!(vert)
      verts = self.vertices
      index = verts.index(vert)
      raise ArgumentError.new(LH["Vertex is not a part of this quads"]) if index == nil
      new_vertex_array = []
      4.times do |i|
        new_vertex_array << verts[(index + i) % 4]
      end
      @vertices = new_vertex_array
      @edges = nil # must invalidate the edges array so they will be recalculated if needed
    end

    # Find the opposite edge on the quad relative to the given edge
    def opposite_edge(this_edge)
      edges_arr = edges
      index_this_edge = edges_arr.index(this_edge)
      raise ArgumentError.new(LH["Edge is not a part of this quad"]) if index_this_edge == nil
      edges_arr[(index_this_edge + 2) % 4]
    end

    # Find the other quad used by this edge if it exists (and only one exists)
    # Quad will be created with the same orientation as this quad
    # @return [QuadFace or false] returns a quad face or false of the edge is not attached to another quadface
    def other_quad(edge)
      faces = edge.faces
      this_quad_faces = self.sub_faces
      uncommon_faces = faces - this_quad_faces
      this_quad_edge_index = self.edges.index(edge)
      if uncommon_faces.length == 1
        other_quad = QuadFace.get(uncommon_faces[0])
        return false unless other_quad
        other_quad_edge_index = other_quad.edges.index(edge)
        return false unless other_quad_edge_index

        # For consistent orientation, we want the index of the edge in the other quad
        # to be offset by 2 relative to this quad
        desired_edge_index = (this_quad_edge_index + 2) % 4
        diff = other_quad_edge_index - desired_edge_index

        # The difference is the value that we need to offset the current vertex order of the
        # other quad
        new_start_vertex = other_quad.vertices[diff]
        other_quad.set_start_vertex!(new_start_vertex)
        return other_quad
      end
      return false
    end

    # Check if the given edge is reversed with respect to this quad
    # @return [Bool] returns true only if the given edge is reversed on this quad
    def reversed_edge?(this_edge)
      edges_arr = edges
      index_this_edge = edges_arr.index(this_edge)
      raise ArgumentError.new(LH["Edge is not a part of this quad"]) if index_this_edge == nil
      faces = sub_faces
      faces.each do |f|
        if this_edge.reversed_in?(f)
          return true
        end
      end
      return false
    end

    def material
      @faces[0].material
    end

    def back_material
      @faces[0].back_material
    end

    def casts_shadows?
      @faces[0].casts_shadows?
    end

    def receives_shadows?
      @faces[0].receives_shadows?
    end

    def layer
      @faces[0].layer
    end

    def hidden?
      @faces[0].hidden?
    end

    def visible?
      @faces[0].visible?
    end

    # Gets the uvs for the array of points.  These points need not be on the quad face or sub-faces but it is
    # assumed that if this quads has two sub-faces, the UVs are the same at each shared vertex of the sub-faces (welded UVs)
    # Otherwise, strange results could occur.
    # @param [Array<Geom::Point3d>] points
    # @param [Boolean] is_front
    # @param [Sketchup::TextureWriter] tw
    def get_uvs(points, is_front, tw)
      # If this is a non-planar quad, use the average uv for the given point
      f_mats = @faces.collect { |f|
        if is_front
          f.material
        else
          f.back_material
        end
      }

      # Ensure the sub faces use the same material.  If not, use the first face and the first material only
      f_mats.uniq!
      if f_mats.length != 1
        # Multiple materials found for this quad so just take the first face only
        uv_faces = @faces[0]
      else
        # Only one material found for this quad
        # TODO: Should check that the end points of the quad diagonal both have the same UVs with respect
        # to each sub-face of the quad. Otherwise, strange results are likely to occur
        uv_faces = @faces.dup
      end
      uv_helpers = {}
      uv_faces.each { |f| uv_helpers[f] = f.get_UVHelper(is_front, !is_front, tw) }

      uvs = []
      points.each_index { |i|
        p = points[i]
        closest_face = get_best_face_for_uvs(uv_faces, p)
        uv_help = uv_helpers[closest_face]
        uv = is_front ? uv_help.get_front_UVQ(p) : uv_help.get_back_UVQ(p)
        # Divide by q to get rid of 'q' component
        q = uv.z
        uv.x = uv.x / q
        uv.y = uv.y / q
        uv.z = 1
        uvs << uv
      }
      uvs
    end

    def normal
      # Use the normal of the best-fit plane formed by the 4 vertices of this quad
      face_avg_normal = Geom::Vector3d.new
      @faces.each do |f|
        f_normal = f.normal
        face_avg_normal.x += f_normal.x
        face_avg_normal.y += f_normal.y
        face_avg_normal.z += f_normal.z
      end
      points = vertices.collect { |v| v.position }
      plane = Geom.fit_plane_to_points(points)
      plane_normal = Geom::Vector3d.new(plane[0], plane[1], plane[2])
      if plane_normal.dot(face_avg_normal) < 0.0
        plane_normal.reverse!
      end
      plane_normal.normalize
    end

    def reverse!
      @faces.each do |f|
        f.reverse!
      end
    end

    # Get the plane of the quad
    def plane
      [self.vertices[0].position, normal]
    end

    def erase!
      ents = Sketchup.active_model.active_entities
      ents.erase_entities(@faces)
      ents.erase_entities(@diagonal) if @diagonal && @diagonal.valid?
    end

    def hidden=(val)
      @faces.each { |f| f.hidden = val }
      @diagonal.hidden = val if @diagonal
    end

    def material=(mat)
      @faces.each { |f| f.material = mat }
    end

    def back_material=(mat)
      @faces.each { |f| f.back_material = mat }
    end

    def position_material(mat, pt_array, on_front)
      @faces.each do |f|
        f.position_material(mat, pt_array, on_front)
      end
    end

    # Equivalent of pushpull for a simple face but works for a quad
    # @param [Number] distance the distance in inches to pushpull (can be negative)
    # @param [Unused] flags
    # @return [Quadface] returns the extruded quadface
    def pushpull(distance, flags)
      offset = self.normal
      offset.length = distance
      points = self.vertices.collect { |v| v.position }
      erase!
      extruded_points = points.collect { |p| p.offset(offset) }
      f0 = [points[0], points[1], extruded_points[1], extruded_points[0]]
      f1 = [points[1], points[2], extruded_points[2], extruded_points[1]]
      f2 = [points[2], points[3], extruded_points[3], extruded_points[2]]
      f3 = [points[3], points[0], extruded_points[0], extruded_points[3]]
      side_faces = [f0, f1, f2, f3]

      ents = Sketchup.active_model.active_entities
      side_faces.each { |f_points| QuadFace.create(f_points, ents) }

      # Create and return the extruded quadface
      QuadFace.create(extruded_points, ents)
    end

    # Splits a planar quad (adds a diagonal to a planar quad) while maximizing the interior
    # angles between the edges
    def fold
      if @is_planar
        quad_points = vertices.collect { |v| v.position }

        # Find the minimum interior angle based on the diagonal being constructed from
        # index 1 to index 3
        first_face_angles = QuadFace.get_first_face_angles(quad_points)
        second_face_angles = QuadFace.get_second_face_angles(quad_points)

        if first_face_angles && second_face_angles
          all_angles = first_face_angles.concat(second_face_angles)
          min_angle_current = all_angles.min
        end

        # Find the minimum interior angle based on the diagonal being constructed from
        # index 0 to index 2
        first_face_angles_flipped = QuadFace.first_face_angles_flipped(quad_points)
        second_face_angles_flipped = QuadFace.second_face_angles_flipped(quad_points)

        if first_face_angles_flipped && second_face_angles_flipped
          all_angles_flipped = first_face_angles_flipped.concat(second_face_angles_flipped)
          min_angle_flipped = all_angles_flipped.min
        end

        ents = @faces[0].parent.entities
        points = [quad_points[1], quad_points[3]] # default diagonal points
        if min_angle_current && min_angle_flipped
          if min_angle_flipped.abs > min_angle_current.abs
            # diagonal from 0 to 2 has larger minimum angle so use that instead
            points = [quad_points[0], quad_points[2]]
          end
        end
        edge = ents.add_line(points[0], points[1])
        if edge
          QuadFace.set_diagonal(edge)
          @diagonal = edge
          @faces = edge.faces
          @is_planar = false
          if @faces.length != 2
            #
          end
        end

      end
    end

    # Removes the diagonal of a quad if it is found to be planar
    def unfold
      if @diagonal && @diagonal.valid?
        faces = @diagonal.faces
        current_edges = self.edges
        if faces[0].normal.angle_between(faces[1].normal).zero?
          ents = @diagonal.parent.entities
          # Erasing the diagonal should result in the creation of a new planar face
          ents.erase_entities([@diagonal])
          common_faces = []
          current_edges.each_index do |index|
            this_edge = current_edges[index]
            next_edge = current_edges[(index + 1) % 4 ]
            common_faces << this_edge.common_face(next_edge)
          end
          new_face = common_faces.first
          if new_face.is_a?(Sketchup::Face)
            @faces = [new_face]
            common_faces.delete(new_face) # should be empty array after this
            raise LH["More than one common face!"] if common_faces.length > 0
          else
            # This could happen if geometry was merged by a transform operation
            #raise LH["No common face between first two edges!"]
          end
          # Set vertices and edges to nil to force recalculation
          @vertices = nil
          @edges = nil
          @is_planar = true
          @diagonal = nil
        end
      end
    rescue TypeError
      #
    end

    # Quad sub-faces 'should' have shared UVs at the vertices.  However, due the way
    # UVs get projected on a triangle face, the UV Helper will return different results
    # for calculated UVs for the same input 3D point.  For best results, we want to use
    # the face that is closest to the given point when calculated the UVs noting that
    # the point does not need to be on any of the sub-faces for the UVs to be calculated
    def get_best_face_for_uvs(faces, point)
      return faces.first if faces.length == 1
      distances_from_edge = {}
      faces.each do |f|
        distances_from_edge[f] = []
        edges = f.edges
        edges.each { |edge|
          distances_from_edge[f] << point.distance_to_line(edge.line)
        }
      end

      # Get the minimum edge distance for each of the sub-faces
      min_distances = {}
      distances_from_edge.each do |f, distances|
        min_distances[f] = distances.min
      end

      # Get the overall minimum distance
      min_overall = min_distances.values.min
      # Return the face with the minimum distance from this point to any one
      # of the edges of the face
      min_distances.key(min_overall)
    end

    # Utility Methods for processing quad geometry
    # Given 4 points (quad), find the interior angles of the triangle that is formed from
    # the vertices at indecies 0, 1, 3.  The 'diagonal' of the quad is considered to be
    # from index 1 to index 3
    # @param [Array<Geom::Point3d>] quad an array of four points forming a quad
    # @return [Array<Number>] returns an array of angles in radians
    def self.get_first_face_angles(quad)
      # Get the vectors of the first face
      f0_edge_vec1 = quad[0].vector_to(quad[1])
      f0_edge_vec2 = quad[0].vector_to(quad[3])
      f0_edge_vec3 = quad[1].vector_to(quad[3])

      # Get the angles of the first face
      f0_angle1 = f0_edge_vec1.angle_between(f0_edge_vec2)
      f0_angle2 = f0_edge_vec1.reverse.angle_between(f0_edge_vec3)
      f0_angle3 = Math::PI - f0_angle1 - f0_angle2

      # We can't safely flip the edge if the angle between the existing diagonal
      # and the adjacent edges is greater than 90 degrees
      if f0_angle2 > 90.degrees || f0_angle3 > 90.degrees
        return false
      end
      return [f0_angle1, f0_angle2, f0_angle3]
    end

    # Given 4 points (quad), find the interior angles of the triangle that is formed from
    # the vertices at indecies 1, 2, 3.  The 'diagonal' of the quad is considered to be
    # from index 1 to index 3
    # @param [Array<Geom::Point3d>] quad an array of four points forming a quad
    # @return [Array<Number>] returns an array of angles in radians
    def self.get_second_face_angles(quad)
      # Get the vectors of the second face
      f1_edge_vec1 = quad[2].vector_to(quad[3])
      f1_edge_vec2 = quad[2].vector_to(quad[1])
      f1_edge_vec3 = quad[3].vector_to(quad[1])

      # Get the angles of the second face
      f1_angle1 = f1_edge_vec1.angle_between(f1_edge_vec2)
      f1_angle2 = f1_edge_vec1.reverse.angle_between(f1_edge_vec3)
      f1_angle3 = Math::PI - f1_angle1 - f1_angle2

      # We can't safely flip the edge if the angle between the existing diagonal
      # and the adjacent edges is greater than 90 degrees
      if f1_angle2 > 90.degrees || f1_angle3 > 90.degrees
        return false
      end
      return [f1_angle1, f1_angle2, f1_angle3]
    end

    # The methods below are identical to the two above except that in this condition, the
    # quad diagonal is assumed to be constructed from vertex 0 to 2
    # @param [Array<Geom::Point3d>] quad an array of four points forming a quad
    # @return [Array<Number>] returns an array of angles in radians
    def self.first_face_angles_flipped(quad)
      # Now with Flipped Edge Condition
      # Get the vectors of the first face
      f0_edge_vec1 = quad[0].vector_to(quad[1])
      f0_edge_vec2 = quad[0].vector_to(quad[2])
      f0_edge_vec3 = quad[1].vector_to(quad[2])

      # Get the angles of the first face
      f0_angle1 = f0_edge_vec1.angle_between(f0_edge_vec2)
      f0_angle2 = f0_edge_vec1.reverse.angle_between(f0_edge_vec3)
      f0_angle3 = Math::PI - f0_angle1 - f0_angle2

      [f0_angle1, f0_angle2, f0_angle3]
    end

    # @param [Array<Geom::Point3d>] quad an array of four points forming a quad
    # @return [Array<Number>] returns an array of angles in radians
    def self.second_face_angles_flipped(quad)
      # Get the vectors of the second face
      f1_edge_vec1 = quad[3].vector_to(quad[2])
      f1_edge_vec2 = quad[3].vector_to(quad[0])
      f1_edge_vec3 = quad[2].vector_to(quad[0])

      # Get the angles of the second face
      f1_angle1 = f1_edge_vec1.angle_between(f1_edge_vec2)
      f1_angle2 = f1_edge_vec1.reverse.angle_between(f1_edge_vec3)
      f1_angle3 = Math::PI - f1_angle1 - f1_angle2

      [f1_angle1, f1_angle2, f1_angle3]
    end

  end
end