
module DM::Bevel

  class FaceCornerOffsetData

    # @return [FaceCorner]
    attr_reader :face_corner

    # @return [Geom::Point3d]
    attr_reader :offset_pt

    # @param face_corner [FaceCorner]
    # @param offset_pt [Geom::Point3d]
    def initialize(face_corner, offset_pt)
      @face_corner = face_corner
      @offset_pt = offset_pt
    end

    # @param face_corner_offset [FaceCornerOffsetData]
    # @return [Boolean]
    def same_face?(face_corner_offset)
      self.face_corner.same_face?(face_corner_offset.face_corner)
    end

    # @param other_offset_data [FaceCornerOffsetData]
    def ==(other_offset_data)
      return false unless other_offset_data.instance_of?(FaceCornerOffsetData)
      return false unless @face_corner == other_offset_data.face_corner

      @offset_pt == other_offset_data.offset_pt
    end

    alias eql? :==

    def hash
      pt_hash = @offset_pt.x.hash + @offset_pt.y.hash + @offset_pt.z.hash
      @face_corner.hash + pt_hash
    end

    def inspect
      self.to_s
    end

  end

  # The part where the edge hits a corner and "transitions" into a patch/other edge/etc.
  class TransitionSegment
    # @return [FaceCornerOffsetData]
    attr_reader :face_corner_offset1
    # @return [FaceCornerOffsetData]
    attr_reader :face_corner_offset2
    # @return [BevelCurve]
    attr_reader :curve

    # @return [TransitionSegment]
    attr_reader :opposite_transition_segment
    # @return [Array<TransitionSegment>]
    def adjacent_transition_segments
      @adjacent_transition_segments.clone
    end

    # @param face_corner_offset1 [FaceCornerOffsetData]
    # @param face_corner_offset2 [FaceCornerOffsetData]
    # @param bevel_curve [BevelCurve]
    def initialize(face_corner_offset1, face_corner_offset2, bevel_curve)
      raise UnexpectedStateError, "face_corner_offsets must be at the same vertex" unless face_corner_offset1.face_corner.vertex == face_corner_offset2.face_corner.vertex
      raise UnexpectedStateError, "The face corners are not adjacent (they do not share a edge_vert" if face_corner_offset1.face_corner.get_shared_edge_vert(face_corner_offset2.face_corner).nil?

      @face_corner_offset1 = face_corner_offset1
      @face_corner_offset2 = face_corner_offset2
      @curve = bevel_curve
      @opposite_transition_segment = nil
      @adjacent_transition_segments = []
    end

    # @param opposite_transition_segment [TransitionSegment]
    def set_opposite_transition_segment(opposite_transition_segment)
      raise UnexpectedStateError.new("The segments must be opposite of each other", self, opposite_transition_segment) unless self.shared_edge_vert.opposite?(opposite_transition_segment.shared_edge_vert)

      @opposite_transition_segment = opposite_transition_segment
    end

    # @param adjacent_transition_segment [TransitionSegment]
    def add_adjacent_transition_segments(adjacent_transition_segment)
      unless connected?(adjacent_transition_segment)
        raise UnexpectedStateError.new("adjacent_transition_segment curve must share a start or end point with this transition_segment curve", self, adjacent_transition_segment)
      end

      @adjacent_transition_segments << adjacent_transition_segment unless @adjacent_transition_segments.include?(adjacent_transition_segment)
    end

    # @return [Array<FaceCornerOffsetData>]
    def face_corner_offsets
      [self.face_corner_offset1, self.face_corner_offset2]
    end

    # @return [Array<Sketchup::Face>]
    #   Index 0: The face for face_corner_offset1
    #   Index 1: The face for face_corner_offset2
    def faces
      face_array = []
      self.face_corner_offsets.each do |face_corner_offset|
        face_corner = face_corner_offset.face_corner
        if face_corner.instance_of?(CoplanarFaceCorner)
          # Always pick the closest inner_face_corner to the Shared edge_vert
          face = face_corner.closest_coplanar_face_corner(self.shared_edge_vert).face
        else
          face = face_corner.face
        end
        face_array << face
      end

      face_array
    end

    # @return [Boolean]
    def face_properties_different?
      face1 = self.faces[0]
      face2 = self.faces[1]
      return true if face1.instance_of?(FakeFace) || face2.instance_of?(FakeFace)
      return true unless face1.material == face2.material
      return true unless face1.back_material == face2.back_material
      return true unless face1.casts_shadows? == face2.casts_shadows?
      return true unless face1.receives_shadows? == face2.receives_shadows?
      return true unless face1.layer == face2.layer
      return true unless face1.hidden? == face2.hidden?
      return true unless face1.visible? == face2.visible?

      false
    end

    # @param other_transition_segment [TransitionSegment]
    # @return [Boolean]
    def connected?(other_transition_segment)
      start_end_array = [other_transition_segment.curve.start_pt, other_transition_segment.curve.end_pt]
      start_end_array.include?(self.curve.start_pt) || start_end_array.include?(self.curve.end_pt)
    end

    # @return [Boolean]
    def three_point_segment?
      return false unless self.face_corner_offset1.face_corner.face == self.face_corner_offset2.face_corner.face
      self.face_corner_offset1.face_corner.three_point_corner? && self.face_corner_offset2.face_corner.three_point_corner?
    end

    # @return [Boolean]
    def point_curve?
      self.curve.instance_of?(PointCurve)
    end

    # Case where the edge_offset is so large the corner to corners are touching.
    # @return [Boolean]
    def overlapping_with_opposite?
      self.loop_with_opposite do |pt1, pt2|
        return false unless pt1 == pt2
      end
      true
    end

    # @return [EdgeVertex]
    def shared_edge_vert
      @face_corner_offset1.face_corner.get_shared_edge_vert(@face_corner_offset2.face_corner)
    end

    # @param face_corner [FaceCorner]
    # @return [FaceCornerOffsetData, Nil]
    def get_offset_on_face_corner(face_corner)
      self.face_corner_offsets.each do |face_corner_offset|
        return face_corner_offset if face_corner == face_corner_offset.face_corner
      end
      nil
    end

    # @return [TransitionSegment]
    def adjacent_to_end
      self.adjacent_transition_segments.each do |adjacent_transition_segment|
        return adjacent_transition_segment if self.curve.end_pt == adjacent_transition_segment.curve.start_pt
      end
      nil
    end

    # @return [TransitionSegment]
    def adjacent_to_start
      (self.adjacent_transition_segments - [adjacent_to_end]).first
    end

    # Ensures we loop from the same face_corner (bezier_curve points direction could be opposite, so to avoid that use this method)
    # @yield [pt, opposite_pt]
    # @yieldparam pt [Geom::Point3d] The point on the bezier curve from this TransitionSegment's perspective
    # @yieldparam opposite_pt [Geom::Point3d] The point on the bezier curve from the opposite TransitionSegment
    def loop_with_opposite
      return if opposite_transition_segment.nil?

      start_on_same_face = @face_corner_offset1.same_face?(opposite_transition_segment.face_corner_offset1)
      @curve.length.times do |i|
        pt1 = @curve[i]
        opposite_i = start_on_same_face ? i : -1 - i
        pt2 = opposite_transition_segment.curve[opposite_i]
        yield(pt1, pt2) if block_given?
      end
    end

    # @return [Hash{Sketchup::Face => Array<Geom::Point3d>}]
    def face_to_opposite_start_end_pts
      return {} if opposite_transition_segment.nil?

      start_on_same_face = @face_corner_offset1.same_face?(opposite_transition_segment.face_corner_offset1)
      face_to_start_end_pts = { }
      face_to_start_end_pts[@face_corner_offset1.face_corner.face] = [@curve.start_pt, start_on_same_face ? opposite_transition_segment.curve.start_pt : opposite_transition_segment.curve.end_pt]

      raise UnexpectedStateError.new('', self) unless face_to_start_end_pts[@face_corner_offset2.face_corner.face].nil? # TODO: Logic check, made the assumption that the curve always starts/ends to two different faces

      face_to_start_end_pts[@face_corner_offset2.face_corner.face] = [@curve.end_pt, start_on_same_face ? opposite_transition_segment.curve.end_pt : opposite_transition_segment.curve.start_pt]
      face_to_start_end_pts
    end

    # @yield [pt]
    # @yieldparam pt [Geom::Point3d]
    # @return [Array<Point3d>] Does not include index 0 or the middle point.
    def inner_points_from_start
      half_pts = []
      half_length = (self.curve.length / 2) - 1
      (1..half_length).each do |i|
        pt = self.curve[i]
        yield(pt) if block_given?
        half_pts << pt
      end
      half_pts
    end

    # @yield [pt]
    # @yieldparam pt [Geom::Point3d]
    # @return [Array<Point3d>] Does not include index 0 or the middle point.
    def inner_points_from_end
      half_pts = []
      length = self.curve.length
      half_length = (length / 2) + 1
      (half_length..(length - 2)).each do |i|
        pt = self.curve[i]
        yield(pt) if block_given?
        half_pts << pt
      end
      half_pts.reverse
    end

    def inspect
      self.to_s
    end

  end


  class TransitionLoop

    # @return [Corner]
    attr_reader :corner

    # @return [Array<TransitionSegment>] An ordered list of TransitionSegments
    def transition_segments
      @transition_segments.clone
    end

    # @param corner [Corner]
    # @param transition_segments [Array<TransitionSegment>]
    def initialize(corner, transition_segments)
      raise UnexpectedStateError, "Must have at least 3 transition_segments to create a patch" if transition_segments.length < 3
      transition_segments.each do |transition_segment|
        shared_edge_vert = transition_segment.shared_edge_vert
        raise UnexpectedStateError.new("The transition_segments must all be at the same corner", corner, shared_edge_vert) unless shared_edge_vert.corner == corner
      end

      # Order the transition segments so we can iterate in order.
      cur_transition_segment = transition_segments.first
      ordered_transition_segments = [cur_transition_segment]
      i = 0
      while true
        i += 1
        raise UnexpectedStateError if i > 100 # TODO: Infinite loop check

        cur_transition_segment.adjacent_transition_segments.each do |adjacent_transition_segment|
          next if ordered_transition_segments.include?(adjacent_transition_segment)

          ordered_transition_segments << adjacent_transition_segment
          cur_transition_segment = adjacent_transition_segment
          break
        end

        break if ordered_transition_segments.length == transition_segments.length
      end

      # TODO: Logic check: Make sure we actually ordered them properly.
      ordered_transition_segments.length.times do |i|
        ts1 = ordered_transition_segments[i]
        ts2 = ordered_transition_segments[i + 1]
        ts2 = ordered_transition_segments[0] if ts2.nil?

        raise UnexpectedStateError.new('', transition_segments, ts1) unless transition_segments.include?(ts1)

        # Compare the start/end pts to make sure they are connected.
        num_equal = 0
        [ts1.curve.start_pt, ts1.curve.end_pt].each do |pt1|
          [ts2.curve.start_pt, ts2.curve.end_pt].each do |pt2|
            num_equal += 1 if pt1 == pt2
          end
        end
        raise UnexpectedStateError.new("The transition_loop must form a closed loop.", ts1, ts2) if num_equal == 0
      end

      # Make sure the following condition is true to ensure consistency of calculations.
      unless ordered_transition_segments.first.curve.start_pt == ordered_transition_segments.last.curve.end_pt
        ordered_transition_segments.reverse!
      end

      @corner = corner
      @transition_segments = ordered_transition_segments
    end

    # @param from_transition_segment [TransitionSegment]
    # @return [TransitionSegment]
    def get_transition_segment_across(from_transition_segment)
      cur_index = @transition_segments.index(from_transition_segment)
      return nil if cur_index.nil? # If this happens, that means from_transition_segment doesn't belong to this TransitionLoop

      length = @transition_segments.length
      half_i = length / 2
      across_index = cur_index + half_i
      i = across_index < length ? across_index : across_index - length
      @transition_segments[i]
    end

    # @param from_transition_segment [TransitionSegment]
    # @param from_start [Boolean]
    # @return [TransitionSegment]
    def get_next_transition_segment_over(from_transition_segment, from_start = true)
      cur_index = @transition_segments.index(from_transition_segment)
      return nil if cur_index.nil? # If this happens, that means from_transition_segment doesn't belong to this TransitionLoop

      length = @transition_segments.length
      next_index = cur_index + (from_start ? -2 : 2)
      next_index = next_index < length ? next_index : next_index - length
      @transition_segments[next_index]
    end

  end

end


