module DM::Bevel

  # @param entities [Array<Sketchup::Entity>]
  def self.clean_coplanar_collinear_edges(entities = Sketchup.active_model.selection.to_a)
    model = Sketchup.active_model
    begin
      if entities.empty?
        entities = Sketchup.active_model.active_entities.to_a
        no_selection = true
      end

      if CoreTool.active_tool.instance_of?(BevelTool)
        # Set to default select tool to avoid breaking things.
        Sketchup.active_model.select_tool(nil)
      end

      model.start_operation(LH["Clean Edges"], true)

      ents_to_select = []

      # Need to collect these first
      quad_cleaner = QuadFaceCleaner.new
      quad_cleaner.collect_edges(entities)

      coplanar_cleaner = CoplanarCleaner.new
      coplanar_cleaner.collect_edges(entities)

      # The logic to collect edges is the same in CollinearCleaner. So only need total for CoplanarCleaner.
      num_edges_processed = coplanar_cleaner.total_edges

      num_erased = 0
      unless coplanar_cleaner.entities_to_edges.empty?
        merged_faces_with_total = coplanar_cleaner.clean_coplanar
        num_erased = merged_faces_with_total[1]
        ents_to_select.concat(merged_faces_with_total[0])
      end

      collinear_cleaner = CollinearCleaner.new
      collinear_cleaner.collect_edges(entities)
      num_combined = 0
      unless collinear_cleaner.entities_to_grouped_edges.empty?
        new_ents_with_total = collinear_cleaner.clean_collinear
        num_combined = new_ents_with_total[1]
        ents_to_select.concat(new_ents_with_total[0]) # Add the new edges/faces to the selection.
      end

      if num_erased.zero? && num_combined.zero?
        model.abort_operation
      else
        # Now clean up the quads (if needed)
        quad_cleaner.clean_quads

        ents_to_select.select! { |ent| ent.valid? && ent.parent.entities == Sketchup.active_model.active_entities }
        Sketchup.active_model.selection.add(ents_to_select) unless ents_to_select.empty?

        model.commit_operation
      end

      self.notify_clean_complete(num_edges_processed, num_erased, num_combined, no_selection)
    rescue StandardError => err
      DM::Bevel::Licensing::Log.error(err.message)
      DM::Bevel::Licensing::Log.error(err.backtrace)
      model.abort_operation
      UI.messagebox(PLUGIN_NAME + ": " + LH["An error occurred. For help, please go to support.mindsightstudios.com"])
    end
  end

  def self.notify_clean_complete(num_edges_processed, num_erased, num_combined, no_selection = false)
    message = LH["Clean Edges"] + "\r\n\r\n"
    if no_selection
      message += LH["Processed All Edges in Context: "]
    else
      message += LH["Edges Selected: "]
    end
    message += "#{num_edges_processed}\r\n"
    if num_erased.zero? && num_combined.zero?
      message += LH["No Edges Found to Clean"]
    else
      message += LH["Coplanar Edges Deleted: "]
      message += "#{num_erased}\r\n"
      message += LH["Collinear Edges Combined: "]
      message += "#{num_combined}"
    end
    PLUGIN.notification(message, "bevel-clean_icon.png")
  end

  class EmptySelectionError < RuntimeError
  end

  module CleanerMixin

    # @param entities [Array<Sketchup::Entity>]
    # @return [Hash{Sketchup::Entities => Array<Sketchup::Edge>}]
    def collect_edges_from_entities(entities)
      entities_to_edges = Hash.new {|h,k| h[k] = []}
      entities.each do |entity|
        next unless entity.valid?

        edges = []
        if entity.instance_of?(Sketchup::Edge)
          edges << entity
        elsif entity.respond_to?(:edges)
          edges.concat(entity.edges)
        else
          next
        end

        next if edges.empty?

        cur_ents = entity.parent.entities
        edges.each do |edge|
          entities_to_edges[cur_ents] << edge
        end
      end
      entities_to_edges.values.each { |edges| edges.uniq! }
      entities_to_edges
    end

  end

  class QuadFaceCleaner
    include CleanerMixin

    # @return [Hash{Sketchup::Entities => Array<Sketchup::Edge>}]
    def entities_to_edges
      @entities_to_edges.clone
    end

    def initialize
      @entities_to_edges = Hash.new {|h,k| h[k] = []}
    end

    # @param entities [Array<Sketchup::Entity>]
    def collect_edges(entities)
      collect_quad_diagonals(entities).each_pair do |ents, edges|
        next if edges.empty?

        @entities_to_edges[ents].concat(edges)
        @entities_to_edges[ents].uniq!
      end
    end

    # @return [Array<Array<Sketchup::Face>, Integer>]
    #       - Index 0: List of edges after being converted back to normal.
    #       - Index 1: The number of quad diagonals fixed.
    def clean_quads
      # We have to collect them first because if we change one, then theres a chance it will affect the other quad_diagonal validation if they are adjacent.
      quad_diagonals_to_fix = []
      self.entities_to_edges.each_pair do |entities, edges|
        edges.each do |edge|
          next unless edge.valid?

          quad_diagonals_to_fix << edge unless QuadFace.is_quadface_diagonal?(edge)
        end
      end

      num_fixed = quad_diagonals_to_fix.length
      quad_diagonals_to_fix.each do |quad_diagonal_edge|
        quad_diagonal_edge.casts_shadows = true
      end

      [quad_diagonals_to_fix, num_fixed]
    end


    private

    # @param entities [Array<Sketchup::Entity>]
    # @return [Hash{Sketchup::Entities => Array<Sketchup::Edge>}]
    def collect_quad_diagonals(entities)
      entities_to_edges = Hash.new {|h,k| h[k] = []}
      collect_edges_from_entities(entities).each_pair do |ents, edges|
        edges.each do |edge|
          next unless QuadFace.is_quadface_diagonal?(edge)

          entities_to_edges[ents] << edge
        end
      end
      entities_to_edges
    end

  end

  class CoplanarCleaner
    include GeometryMixin
    include CleanerMixin

    # @return [Integer] The total edges collected (may not all be coplanar)
    attr_reader :total_edges

    # @return [Hash{Sketchup::Entities => Array<Sketchup::Edge>}]
    def entities_to_edges
      @entities_to_edges.clone
    end

    def initialize
      @entities_to_edges = Hash.new {|h,k| h[k] = []}
    end

    # @param entities [Array<Sketchup::Entity>]
    def collect_edges(entities)
      ents_to_edges_with_total = collect_coplanar_edges(entities)
      @total_edges = ents_to_edges_with_total[1]
      ents_to_edges_with_total[0].each_pair do |ents, edges|
        next if edges.empty?

        @entities_to_edges[ents].concat(edges)
        @entities_to_edges[ents].uniq!
      end
    end

    # @return [Array<Array<Sketchup::Face>, Integer>]
    #       - Index 0: List of Faces after merging.
    #       - Index 1: The number of combined collinear edges.
    def clean_coplanar
      num_erased = 0
      merged_faces = []
      self.entities_to_edges.each_pair do |entities, edges|
        edges = edges.select { |e| e.valid? }
        next if edges.empty?

        edges.each do |edge|
          edge.faces.each do |face|
            merged_faces << face
            face.hidden = false if face.hidden?
          end
        end

        num_erased += edges.length
        entities.erase_entities(edges)
      end
      merged_faces.select! {|face| face.valid? }
      [merged_faces, num_erased]
    end

    private

    # @param entities [Array<Sketchup::Entity>]
    # @return [Array<Hash{Sketchup::Entities => Array<Sketchup::Edge>}, Integer>]
    def collect_coplanar_edges(entities)
      entities_to_edges = Hash.new {|h,k| h[k] = []}
      total_edges = 0
      collect_edges_from_entities(entities).each_pair do |ents, edges|
        total_edges += edges.length
        edges.each do |edge|
          next unless valid_coplanar_edge?(edge)

          entities_to_edges[ents] << edge
        end
      end
      [entities_to_edges, total_edges]
    end

    # @param edge [Sketchup::Edge]
    # @return [Boolean]
    def valid_coplanar_edge?(edge)
      # Note: coplanar_edge? will also check if there are exactly two faces.
      return false unless coplanar_edge?(edge)

      face1 = edge.faces[0]
      face2 = edge.faces[1]

      return false unless face1.material == face2.material
      return false unless face1.back_material == face2.back_material
      # return false unless face1.casts_shadows? == face2.casts_shadows?
      # return false unless face1.receives_shadows? == face2.receives_shadows?
      # return false unless face1.layer == face2.layer
      # return false unless face1.hidden? == face2.hidden?
      # return false unless face1.visible? == face2.visible?
      true
    end

  end

  class CollinearCleaner
    include GeometryMixin
    include CleanerMixin

    # @return [Hash{Sketchup::Entities => Array<Array<Sketchup::Edge>>}]
    def entities_to_grouped_edges
      @entities_to_grouped_edges.clone
    end

    def initialize
      @entities_to_grouped_edges = Hash.new {|h,k| h[k] = []}

      @cache_manager = CacheManager.new
    end

    # @param entities [Array<Sketchup::Entity>]
    def collect_edges(entities)
      collect_collinear_edges(entities).each_pair do |ents, grouped_edges|
        grouped_edges.each do |edges|
          next if edges.empty?

          @entities_to_grouped_edges[ents] << edges
          @entities_to_grouped_edges[ents].uniq!
        end
      end
    end

    # @return [Array<Array<Sketchup::Entity>, Integer>]
    #       - Index 0: List of newly created edges/faces
    #       - Index 1: The number of combined collinear edges.
    def clean_collinear
      num_combined = 0
      new_ents = []
      self.entities_to_grouped_edges.each_pair do |ents, grouped_edges|
        face_to_grouped_edges = Hash.new { |h,k| h[k] = [] }
        grouped_edges.each do |edges|
          edges.each do |edge|
            edge.faces.each do |face|
              next if face_to_grouped_edges[face].include?(edges)

              face_to_grouped_edges[face] << edges
            end
          end
        end

        data_manager = BevelerDataManager.new(face_to_grouped_edges.keys)
        face_to_grouped_edges.each_pair do |face, all_edges|
          # @type [Sketchup::Face]
          face = face

          face_loop = FaceLoops.new(face)
          face_loop.loop_to_loop_vertices.each_pair do |loop, loop_vertices|
            # Find the start and end loop_vertices that are collinear, and that belong to this loop.
            collinear_start_end_loop_vertices = []
            all_edges.each do |collinear_edges|
              collinear_loop_vertices = loop_vertices.find_all { |loop_vertex| collinear_edges.include?(loop_vertex.next_edge) }
              next if collinear_loop_vertices.empty?

              collinear_start_end_loop_vertices << collinear_loop_vertices
            end

            edge_create_datas = []
            _handled_loop_vertices = {}
            loop_vertices.each do |loop_vertex|
              next unless _handled_loop_vertices[loop_vertex].nil?
              _handled_loop_vertices[loop_vertex] = true

              # Check if this loop_vertex is part of a collinear_edge.
              start_end_loop_vertices = nil
              collinear_start_end_loop_vertices.each do |collinear_loop_vertices|
                next unless collinear_loop_vertices.include?(loop_vertex)

                start_end_loop_vertices = collinear_loop_vertices
                break
              end

              if start_end_loop_vertices.nil?
                edge_create_datas << EdgeCreateData.new(loop_vertex.position, loop_vertex.next_vertex.position, edge: loop_vertex.next_edge)
              else
                # Find the further pts of the collinear edges.
                collinear_edges = start_end_loop_vertices.map {|coll_loop_vert| coll_loop_vert.next_edge }
                next if collinear_edges.length < 2

                furthest_start_end_vertices = find_furthest_vertices(collinear_edges)
                start_loop_vertex = loop_vertices.find { |coll_loop_vert| coll_loop_vert.vertex == furthest_start_end_vertices.first }
                end_loop_vertex = loop_vertices.find { |coll_loop_vert| coll_loop_vert.next_vertex == furthest_start_end_vertices.last}

                edge_create = EdgeCreateData.new(start_loop_vertex.position, end_loop_vertex.next_loop_vertex.position, edge: loop_vertex.next_edge)
                edge_create.reverse! unless start_loop_vertex.position.vector_to(end_loop_vertex.next_loop_vertex.position).normalize.samedirection?(loop_vertex.next_line[1])
                edge_create_datas << edge_create
                start_end_loop_vertices.each do |coll_loop_vertex|
                  _handled_loop_vertices[coll_loop_vertex] = true
                end
              end
            end

            next if edge_create_datas.length < 3

            data_manager.add_face_create_datas(FaceCreateData.new(edge_create_datas, face.normal, face, copy_dictionary: true, is_hole: face_loop.is_loop_a_hole?(loop)))
          end
        end

        # Create the new faces, delete the collinear edges, explode, and then reapply the edge properties
        grp = ents.add_group
        EntitiesGenerator.bulk_add_faces(grp.entities, data_manager)
        ents.erase_entities(grouped_edges.flatten)

        exploded_ents = grp.explode
        grouped_edges.each do |collinear_edges|
          num_combined += collinear_edges.length
        end

        if ents == Sketchup.active_model.active_entities
          exploded_ents.each do |ent|
            next unless ent.valid?
            next unless ent.instance_of?(Sketchup::Face) || ent.instance_of?(Sketchup::Edge)

            new_ents << ent
          end
        end
      end

      [new_ents, num_combined]
    end

    private

    # @param edges [Array<Sketchup::Edge>] A collection of connected collinear edges
    # @return [Array<Sketchup::Vertex, Sketchup::Vertex>] The start and end points of the collinear edges.
    def find_furthest_vertices(edges)
      raise UnexpectedStateError, "Must have at least two edges" unless edges.length > 1

      vertices = edges.flat_map{|edge| edge.vertices}.uniq #.map {|vert| vert.position }

      ref_vert = vertices.first
      ref_vec = vertices[0].position.vector_to(vertices[1].position).normalize
      vert_with_dist = [[ref_vert, 0.0]]
      vertices.each do |vert|
        next if ref_vert == vert

        vec = ref_vert.position.vector_to(vert.position)
        dist = vec.length.to_f
        dist *= -1 unless ref_vec.samedirection?(vec)
        vert_with_dist << [vert, dist.round(PLUGIN::FLOAT_PRECISION)]
      end
      vert_with_dist.sort! {|a,b| a[1] <=> b[1]}
      [vert_with_dist.first[0], vert_with_dist.last[0]]
    end

    # @param entities [Array<Sketchup::Entity>]
    # @return [Hash{Sketchup::Entities => Array<Array<Sketchup::Edge>>}]
    def collect_collinear_edges(entities)
      entities_to_edges = collect_edges_from_entities(entities)

      entities_to_collinear_grouped_edges = Hash.new {|h,k| h[k] = []}
      entities_to_edges.each_pair do |ents, edges|
        next if edges.length < 2

        # @type [Array<Sketchup::Edge>]
        edges = edges.uniq

        # Find all collinear edges
        _matched_edges = {}
        edges.each do |edge|
          next unless _matched_edges[edge].nil?

          collinear_grouped_edges = [edge]
          _matched_edges[edge] = true
          edges.each do |other_edge|
            next if edge == other_edge
            next unless _matched_edges[other_edge].nil?
            next unless collinear_edges?(edge, other_edge)

            collinear_grouped_edges << other_edge
            _matched_edges[other_edge] = true
          end

          next if collinear_grouped_edges.length < 2

          # Group them by connectivity now.
          group_by_connected(collinear_grouped_edges).each do |grouped_edges|
            next if grouped_edges.empty?

            entities_to_collinear_grouped_edges[ents] << grouped_edges
          end
        end
      end

      entities_to_collinear_grouped_edges
    end

    # @param edges [Array<Sketchup::Edge>] A collection of collinear edges
    # @return [Array<Array<Sketchup::Edge>>]
    def group_by_connected(edges)
      raise UnexpectedStateError, "Must have at least two edges" unless edges.length > 1

      grouped_edges = []
      _matched_edges = {}
      edges.each do |edge|
        next unless _matched_edges[edge].nil?

        edge_group = [edge]
        _matched_edges[edge] = true
        # Recursively find all the connected edges.
        edge_group.each do |cur_edge|
          edges.each do |other_edge|
            next if cur_edge == other_edge
            next unless _matched_edges[other_edge].nil?
            next if !cur_edge.start.edges.include?(other_edge) && !cur_edge.end.edges.include?(other_edge)

            edge_group << other_edge
            _matched_edges[other_edge] = true
          end
        end
        next if edge_group.length < 2

        grouped_edges << edge_group
      end

      grouped_edges
    end

    # @param edge [Sketchup::Edge]
    # @param other_edge [Sketchup::Edge]
    # @return [Boolean]
    def collinear_edges?(edge, other_edge)
      # they have to be on the same faces
      edge_faces = edge.faces
      other_edge_faces = other_edge.faces
      return false unless (edge_faces & other_edge_faces).length == edge_faces.length

      edge_line = edge.line
      start_pt = other_edge.start.position.to_a
      end_pt = other_edge.end.position.to_a

      return false unless start_pt.project_to_line(edge_line).distance(start_pt).round(3).zero?
      return false unless end_pt.project_to_line(edge_line).distance(end_pt).round(3).zero?
      true
    end

  end

end