
module DM::Bevel

  class FaceCreateData
    include GeometryMixin

    # @return [Array<EdgeCreateData>]
    attr_reader :edge_create_datas
    # @return [Geom::Vector3d]
    attr_reader :normal
    # @return  [Array<Geom::Point3d>]
    attr_reader :face_pts
    # @return [Sketchup::Face]
    attr_reader :face

    # @return [Boolean]
    def is_hole?
      @is_hole
    end

    # @return [Boolean]
    def copy_dictionary?
      @copy_dictionary
    end

    # @param edge_create_datas [Array<EdgeCreateData>]
    # @param normal [Geom::Vector3d] If set, this will force the normal of the created face to be in that direction
    # @param face [Sketchup::Face]
    # @param options [Hash]
    # @option options [Boolean] :is_hole Default false
    # @option options [Boolean] :copy_dictionary Default false
    def initialize(edge_create_datas, normal = nil, face = nil, **options)
      raise UnexpectedTypeError, "normal must be of type Geom::Vector3d" unless normal.instance_of?(Geom::Vector3d)
      raise UnexpectedTypeError, "face must be of type Sketchup::Face and must be valid" if !face.nil? && (!face.instance_of?(Sketchup::Face) || !face.valid?)

      raise BevelEntitiesError, "edge_create_datas cannot be empty" if edge_create_datas.empty?

      # Get an ordered list of points for the face.
      # Note that is is possible to have duplicate points due to coplanar edges.
      new_face_pts = edge_create_datas.map {|edge_create| edge_create.start_pt }
      new_face_pts << edge_create_datas.last.end_pt
      new_face_pts.pop if new_face_pts.first == new_face_pts.last

      raise BevelEntitiesError, "edge_create_datas not planar" unless self.points_planar?(new_face_pts)

      # @type [Array<EdgeCreateData>]
      @edge_create_datas = edge_create_datas
      # @type [Geom::Vector3d]
      @normal = normal
      # @type [Array<Geom::Point3d>]
      @face_pts = new_face_pts
      # @type [Sketchup::Face]
      @face = face

      # @type [Boolean]
      @is_hole = options[:is_hole].nil? ? false : (options[:is_hole] == true)
      # @type [Boolean]
      @copy_dictionary = options[:copy_dictionary].nil? ? false : (options[:copy_dictionary] == true)
    end

    def inspect
      self.to_s
    end

  end

  class EdgeCreateData

    # @return [Geom::Point3d]
    attr_reader :start_pt
    # @return [Geom::Point3d]
    attr_reader :end_pt

    # Shortcut to set both falgs
    # @param val [Boolean]
    def soften_smooth=(val)
      @properties.soft = val
      @properties.smooth = val
    end

    # @return [Sketchup::Edge]
    def edge
      @edge
    end

    # @return [EdgeProperties]
    def properties
      @properties
    end

    # @param start_pt [Geom::Point3d]
    # @param end_pt [Geom::Point3d]
    # @param options [Hash]
    # @option edge [Sketchup::Edge] If set, then the newly created edge will copy properties from that Edge.
    # @option :soft [Boolean] Default true
    # @option :smooth [Boolean] Default true
    # @option :copy_dictionary [Boolean] Default false
    # @option :dictionary_props [Hash] Default nil (won't add anything). Adds custom bevel specific values otherwise.
    def initialize(start_pt, end_pt, **options) # soft = true, smooth = true, hidden = false)
      raise UnexpectedTypeError, "start_pt is not a Geom::Point3d" unless start_pt.instance_of?(Geom::Point3d)
      raise UnexpectedTypeError, "end_pt is not a Geom::Point3d" unless end_pt.instance_of?(Geom::Point3d)
      raise UnexpectedStateError, "start_pt and end_pt are the same" if start_pt == end_pt

      @start_pt = start_pt.clone
      @end_pt = end_pt.clone

      if !options[:edge].nil?
        copy_dictionary = options[:copy_dictionary].nil? ? false : (options[:copy_dictionary] == true)
        @properties = EdgeProperties.from_edge(self, options[:edge], copy_dictionary)
      else
        @properties = EdgeProperties.new(self, options)
      end
    end

    # @return [Array<Geom::Point3d>]
    def points
      [@start_pt, @end_pt]
    end

    def reverse!
      temp = @start_pt
      @start_pt = @end_pt
      @end_pt = temp
    end

    def inspect
      self.to_s
    end


    class EdgeProperties

      # @param edge_create [EdgeCreateData]
      # @param edge [Sketchup::Edge]
      # @param copy_dictionary [Boolean]
      # @return [EdgeProperties]
      def self.from_edge(edge_create, edge, copy_dictionary = false)
        raise UnexpectedTypeError, "edge must be of type Sketchup::Edge" unless edge.instance_of?(Sketchup::Edge)

        options = {
          soft: edge.soft?,
          smooth: edge.smooth?,
          hidden: edge.hidden?,
          visible: edge.visible?,
          casts_shadows: edge.casts_shadows?,
          receives_shadows: edge.receives_shadows?,
          layer: edge.layer,
          material: edge.material,
          was_quad_diagonal: QuadFace.is_quadface_diagonal?(edge),
        }
        if copy_dictionary && !edge.attribute_dictionaries.nil?
          dictionary_hash = Hash.new { |h,k| h[k] = {}}
          edge.attribute_dictionaries.each do |dictionary|
            dictionary.each_pair do |key, value|
              dictionary_hash[dictionary.name][key] = value
            end
          end
          options[:dictionaries] = dictionary_hash
        end
        EdgeProperties.new(edge_create, options)
      end

      # @return [EdgeCreateData]
      attr_reader :edge_create
      # @return [Sketchup::Layer]
      attr_accessor :layer
      # @return [Sketchup::Material]
      attr_accessor :material

      # @return [Hash{String => Hash{String => Object}}]
      attr_reader :dictionaries

      # @return [Boolean]
      def soft?
        @soft
      end

      # @param val [Boolean]
      def soft=(val)
        @soft = val == true
      end

      # @return [Boolean]
      def smooth?
        @smooth
      end

      # @param val [Boolean]
      def smooth=(val)
        @smooth = val == true
      end

      # @return [Boolean]
      def hidden?
        @hidden
      end

      # @param val [Boolean]
      def hidden=(val)
        @hidden = val == true
      end

      # @return [Boolean]
      def visible?
        @visible
      end

      # @param val [Boolean]
      def visible=(val)
        @visible = val == true
      end

      # @return [Boolean]
      def casts_shadows?
        @casts_shadows
      end

      # @param val [Boolean]
      def casts_shadows=(val)
        @casts_shadows = val == true
      end

      # @return [Boolean]
      def receives_shadows?
        @receives_shadows
      end

      # @param val [Boolean]
      def receives_shadows=(val)
        @receives_shadows = val == true
      end

      # @return [Hash]
      def dictionary_props
        @dictionary_props
      end

      # @return [Boolean]
      def was_quad_diagonal?
        @was_quad_diagonal
      end

      # @param edge_create [EdgeCreateData]
      # @param options [Hash]
      def initialize(edge_create, **options)
        @edge_create = edge_create # A reference to the parent object.
        @soft = options[:soft].nil? ? true : (options[:soft] == true)
        @smooth = options[:smooth].nil? ? true : (options[:smooth] == true)
        @hidden = options[:hidden].nil? ? false : (options[:hidden] == true)
        @visible = options[:visible].nil? ? true : (options[:visible] == true)
        @casts_shadows = options[:casts_shadows].nil? ? true : (options[:casts_shadows] == true)
        @receives_shadows = options[:receives_shadows].nil? ? true : (options[:receives_shadows] == true)
        @layer = options[:layer].nil? ? nil : options[:layer]
        raise UnexpectedTypeError, "layer must be of type Sketchup::Layer" if !@layer.nil? && !@layer.instance_of?(Sketchup::Layer)
        @material = options[:material].nil? ? nil : options[:material]
        raise UnexpectedTypeError, "material must be of type Sketchup::Material" if !@material.nil? && !@material.instance_of?(Sketchup::Material)
        @dictionaries = options[:dictionaries].nil? ? nil : options[:dictionaries]
        @dictionary_props = {}
        @was_quad_diagonal = options[:was_quad_diagonal].nil? ? false : (options[:was_quad_diagonal] == true)
      end

      # @param edge [Sketchup::Edge]
      def apply_properties(edge)
        raise UnexpectedTypeError, "edge must be of type Sketch::Edge" unless edge.instance_of?(Sketchup::Edge)

        edge.soft = @soft unless @soft == edge.soft?
        edge.smooth = @smooth unless @smooth == edge.smooth?
        edge.hidden = @hidden unless @hidden == edge.hidden?
        edge.visible = @visible unless @visible == edge.visible?
        edge.casts_shadows = @casts_shadows unless @casts_shadows == edge.casts_shadows?
        edge.receives_shadows = @receives_shadows unless @receives_shadows == edge.receives_shadows?
        edge.layer = @layer if !@layer.nil? && @layer != edge.layer
        edge.material = @material if !@material.nil? && @material != edge.material

        unless @dictionaries.nil?
          @dictionaries.each_pair do |dict_name, key_value|
            key_value.each_pair do |key, value|
              edge.set_attribute(dict_name, key, value)
            end
          end
        end

        unless @dictionary_props.empty?
          @dictionary_props.each_pair do |key, value|
            edge.set_attribute(MAIN_DICTIONARY, key.to_s, value)
          end
        end
      end

      def make_quad_diagonal!
        @soft = true
        @smooth = true
        @casts_shadows = false
      end

      def inspect
        self.to_s
      end

    end

  end


end