
module DM::Bevel

  class BevelerEntities

    # @param entities_to_track [Array<Sketchup::Entity>] These entities are assumed to be all connected.
    def initialize(entities_to_track = [])
      # @type [Array<Sketchup::Entity>]
      if entities_to_track.empty?
        raise UnexpectedStateError, "No Entities to filter."
      else
        @original_entities = entities_to_track.to_a
      end

      @is_cached = false
      @cached_pt_to_edges = Hash.new{|h,k| h[k] = []}
      @cached_pt_to_faces = Hash.new{|h,k| h[k] = []}
      @cached_vert_to_pt_arr = { }
    end

    def clear_cache
      @is_cached = false
      @cached_pt_to_edges.clear
      @cached_pt_to_faces.clear
      @cached_vert_to_pt_arr.clear
    end

    # @param start_pt [Geom::Point3d]
    # @param end_pt [Geom::Point3d]
    # @return [Sketchup::Edge]
    def find_edge_at_pts(start_pt, end_pt)
      build_cache unless @is_cached

      start_edges = @cached_pt_to_edges[get_array_key(start_pt.to_a)]
      end_edges = @cached_pt_to_edges[get_array_key(end_pt.to_a)]
      same_edge = (start_edges & end_edges).first
      same_edge.nil? || !same_edge.valid? ? nil : same_edge
    end

    # Method will do an approximation of the pt position.
    # Very slow and should only be used if the other methods fail.
    # @param start_pt [Geom::Point3d]
    # @param end_pt [Geom::Point3d]
    # @return [Sketchup::Edge]
    def find_edge_at_pts_by_approximation(start_pt, end_pt)
      build_cache unless @is_cached

      # Note: array.sum does not exist in SU 2017
      start_arr = get_array_key(start_pt.to_a)
      start_sum = 0.0
      start_arr.each { |f| start_sum += f }

      end_arr = get_array_key(end_pt.to_a)
      end_sum = 0.0
      end_arr.each { |f| end_sum += f }

      all_start = []
      all_end = []
      precision = PLUGIN::FLOAT_PRECISION - 1 # Weird but this will get us a closer approximation.
      @cached_pt_to_edges.each_pair do |pt_arr, edges|
        pt_sum = 0.0
        pt_arr.each { |f| pt_sum += f }

        all_start.concat(edges) if (pt_sum - start_sum).round(precision).zero?
        all_end.concat(edges) if (pt_sum - end_sum).round(precision).zero?
      end

      matching_edges = (all_start & all_end)
      if matching_edges.empty?
        return nil
      elsif matching_edges.length == 1
        return matching_edges.first
      else
        pts = [start_pt, end_pt]
        matching_edges.each do |matching_edge|
          next unless pts.include?(matching_edge.start.position) && pts.include?(matching_edge.end.position)

          return matching_edge
        end
      end
      nil
    end

    # @param start_pt [Geom::Point3d]
    # @param end_pt [Geom::Point3d]
    # @return [Array<Sketchup::Edge>]
    def find_edges_by_pt_vec(start_pt, end_pt)
      build_cache unless @is_cached

      matching_edges = []
      vec = start_pt.vector_to(end_pt)
      [[start_pt, vec], [end_pt, vec.reverse]].each do |pt_vec|
        pt, vec = *pt_vec
        @cached_pt_to_edges[get_array_key(pt)].each do |edge|
          next unless edge.valid?

          edge_start_pt = @cached_vert_to_pt_arr[edge.start]
          if edge_start_pt.nil?
            edge_start_pt = edge.start.position
            @cached_vert_to_pt_arr[edge.start] = edge_start_pt
          end

          edge_end_pt = @cached_vert_to_pt_arr[edge.end]
          if edge_end_pt.nil?
            edge_end_pt = edge.end.position
            @cached_vert_to_pt_arr[edge.end] = edge_end_pt
          end

          if edge_start_pt == pt
            edge_vec = edge_start_pt.vector_to(edge_end_pt)
          else
            edge_vec = edge_end_pt.vector_to(edge_start_pt)
          end

          matching_edges << edge if edge_vec.samedirection?(vec)
        end
      end
      matching_edges
    end

    # @param face_pts [Array<Geom::Point3d>]
    # @return [Sketchup::Face, Nil]
    def find_face_by_pts(face_pts)
      build_cache unless @is_cached
      return nil if face_pts.length < 2

      found_faces = nil
      face_pts.each do |pt|
        faces = @cached_pt_to_faces[get_array_key(pt)]
        if found_faces.nil?
          found_faces = faces
        else
          found_faces = found_faces & faces
          return nil if found_faces.empty?
        end
      end

      found_faces.first
    end

    private

    def build_cache
      checked_ents = {} # hashes are faster
      @original_entities.each do |ent|
        next unless ent.valid?
        if ent.instance_of?(Sketchup::Edge)
          # Grab only the edge and the adjacent edges to their vertices.
          [ent.start, ent.end].each do |vertex|
            # @type [Sketchup::Vertex]
            vertex = vertex
            next unless checked_ents[vertex].nil?

            checked_ents[vertex] = true

            vertex.edges.each do |edge|
              next unless checked_ents[edge].nil?

              checked_ents[edge] = true

              @cached_pt_to_edges[get_array_key(edge.start.position)] << edge
              @cached_pt_to_edges[get_array_key(edge.end.position)] << edge
            end
          end
        elsif ent.instance_of?(Sketchup::Face)
          ent.vertices.each do |vert|
            cur_faces = @cached_pt_to_faces[get_array_key(vert.position)]
            vert.faces.each do |face|
              next if cur_faces.include?(face)

              cur_faces << face
            end
          end
        end
      end

      @is_cached = true
    end

    # @param pt [Geom::Point3d]
    # @return [Array<Float, Float, Float>]
    def get_array_key(pt)
      arr = pt.to_a
      [arr[0].round(PLUGIN::FLOAT_PRECISION), arr[1].round(PLUGIN::FLOAT_PRECISION), arr[2].round(PLUGIN::FLOAT_PRECISION)]
    end

  end



end


