
module DM::Bevel

  class EntitiesGenerator

    # @param entities [Sketchup::Entities]
    # @param data_manager [BevelerDataManager]
    # @return [Array<EdgeCreateData::EdgeProperties>] Returns the edge and the EdgeProperties used to create it.
    def self.bulk_add_faces(entities, data_manager)
      face_create_datas = data_manager.face_create_datas
      # If a face has holes, we have to use entities.add_faces. EntitiesBuilder and PolygonMesh DO NOT merge overlapping faces,
      # so we can't create a hole use those methods. entities.add_faces will merge the overlapping faces.
      other_face_creates = []
      face_to_face_creates = Hash.new {|h,k| h[k] = [] }
      face_create_datas.each do |face_create|
        if face_create.face.nil?
          other_face_creates << face_create
        else
          face_to_face_creates[face_create.face] << face_create
        end
      end

      # If a face has multiple loops, then there are holes.
      face_to_face_creates.each_pair do |face, face_creates|
        if face.loops.length == 1
          other_face_creates.concat(face_creates)
        else
          face_creates.each do |face_create|
            self.use_add_face(entities, face_create)
          end
        end
      end

      if entities.respond_to?(:build)
        self.use_entities_builder(entities, other_face_creates)
      else
        self.use_polygon_mesh(entities, other_face_creates)
      end

      # Apply edge properties now.
      self.apply_edge_properties(entities, data_manager)
    rescue ArgumentError => error
      # Occurs when non-planar points are added to add_face
      # Possibly other Sketchup Related functions.
      raise BevelEntitiesError, error.message, error.backtrace
    end

    # @param entities_to_modify [Array<Sketchup::Entity>]
    # @param edge_properties [Array<EdgeCreateData::EdgeProperties>]
    def self.cleanup_quads(entities_to_modify, edge_properties)
      beveler_ents = BevelerEntities.new(entities_to_modify.grep(Sketchup::Edge))
      invalid_quads = []
      edge_properties.each do |properties|
        next unless properties.was_quad_diagonal?

        edge = beveler_ents.find_edge_at_pts(properties.edge_create.start_pt, properties.edge_create.end_pt)
        next if edge.nil?

        invalid_quads << edge unless QuadFace.is_quadface_diagonal?(edge)
      end

      invalid_quads.each { |edge| edge.casts_shadows = true }
    end

    private

    # @param entities_to_modify [Array<Sketchup::Entity>] The entities that these properties apply to. (Speed Up, and controls which edges are affected.)
    # @param data_manager [BevelerDataManager]
    # @return [Array<EdgeCreateData::EdgeProperties>] Returns the edge and the EdgeProperties used to create it.
    def self.apply_edge_properties(entities_to_modify, data_manager)
      beveler_ents = BevelerEntities.new(entities_to_modify.grep(Sketchup::Edge))

      # TO avoid any external observers being called over and over, and for optimization,
      # only set the edge properties based on the very last edge_create_data.properties found.
      # @type [Hash {Sketchup::Edge => Array<EdgeCreateData::EdgeProperties>}]
      edge_to_edge_properties = Hash.new {|h,k| h[k] = []}
      data_manager.face_create_datas.each do |face_create|
        face_create.edge_create_datas.each do |edge_create|
          found_edge = beveler_ents.find_edge_at_pts(edge_create.start_pt, edge_create.end_pt)
          found_edge = beveler_ents.find_edge_at_pts_by_approximation(edge_create.start_pt, edge_create.end_pt) if found_edge.nil?
          next if found_edge.nil?

          edge_to_edge_properties[found_edge] << edge_create.properties
        end
      end

      # The last ones at the end will override any states before it.
      data_manager.force_smoothness_edge_create_datas.each do |edge_create|
        # Use an approximation of the edge position.
        beveler_ents.find_edges_by_pt_vec(edge_create.start_pt, edge_create.end_pt).each do |found_edge|
          edge_to_edge_properties[found_edge] << edge_create.properties
        end
      end

      edge_to_properties = {}
      # Apply the properties (if there are any differences)
      edge_to_edge_properties.each_pair do |edge, edge_properties|
        properties = edge_properties.last
        properties.apply_properties(edge)
        edge_to_properties[edge] = properties
      end

      edge_to_properties
    end

    # @param entities [Sketchup::Entities]
    # @param face_create_datas [Array<FaceCreateData>]
    def self.use_entities_builder(entities, face_create_datas)
      entities.build do |builder|
        face_create_datas.each do |face_create|
          face_pts = face_create.face_pts
          new_face = builder.add_face(face_pts)
          new_face.reverse! unless face_create.similar_direction?(new_face.normal, face_create.normal)
          self.copy_face_properties(face_create.face, new_face, face_create.copy_dictionary?) unless face_create.face.nil?
        end
      end
    end

    # @param entities [Sketchup::Entities]
    # @param face_create_datas [Array<FaceCreateData>]
    def self.use_polygon_mesh(entities, face_create_datas)
      # Apparently, setting multiple UVs at the same point with the same texture does not work.
      # So, we will have to search for the faces and apply the properties.
      # https://forums.sketchup.com/t/uv-mapping-importing-meshes-with-multiple-uvs-per-point/203120
      mesh = Geom::PolygonMesh.new
      face_create_datas.each do |face_create|
        face_pts = face_create.face_pts
        vec = Geom::Vector3d.new(Geom.fit_plane_to_points(face_pts)[0..2])
        face_pts = face_pts.reverse unless face_create.similar_direction?(vec, face_create.normal)
        mesh.add_polygon(face_pts)
      end
      entities.add_faces_from_mesh(mesh, Geom::PolygonMesh::NO_SMOOTH_OR_HIDE)

      # We have to search for the faces and apply any properties to them.
      beveler_ents = BevelerEntities.new(entities.grep(Sketchup::Face))
      face_create_datas.each do |face_create|
        next if face_create.face.nil?

        face = beveler_ents.find_face_by_pts(face_create.face_pts)
        next if face.nil? # Should be impossible

        self.copy_face_properties(face_create.face, face, face_create.copy_dictionary?)
      end
    end

    # @param entities [Sketchup::Entities]
    # @param face_create_datas [Array<FaceCreateData>]
    # @return [Array<Sketchup::Face>]
    def self.use_add_faces(entities, face_create_datas)
      faces = []
      face_create_datas.each do |face_create|
        faces << self.use_add_face(entities, face_create)
      end
      faces
    end

    # @param entities [Sketchup::Entities]
    # @param face_create [FaceCreateData]
    # @return [Sketchup::Face]
    def self.use_add_face(entities, face_create)
      face_pts = face_create.face_pts.clone
      new_face = entities.add_face(face_pts)
      new_face.reverse! unless face_create.similar_direction?(new_face.normal, face_create.normal)
      if face_create.is_hole?
        new_face.erase!
      else
        self.copy_face_properties(face_create.face, new_face, face_create.copy_dictionary?)
      end
      new_face
    end

    # @param old_face [Sketchup::Face]
    # @param new_face [Sketchup::Face]
    # @param copy_dictionary [Boolean]
    def self.copy_face_properties(old_face, new_face, copy_dictionary)
      raise UnexpectedStateError, "old_face is invalid." unless old_face.valid?
      raise UnexpectedStateError, "new_face is invalid." unless new_face.valid?

      self.copy_materials(old_face, new_face)
      new_face.casts_shadows = old_face.casts_shadows?
      new_face.receives_shadows = old_face.receives_shadows?
      new_face.layer = old_face.layer
      if copy_dictionary && !old_face.attribute_dictionaries.nil?
        old_face.attribute_dictionaries.each do |dictionary|
          dictionary.each_pair do |key, value|
            new_face.set_attribute(dictionary.name, key, value)
          end
        end
      end
    end

    # @param old_face [Sketchup::Face]
    # @param new_face [Sketchup::Face]
    def self.copy_materials(old_face, new_face)
      raise UnexpectedStateError, "old_face is invalid." unless old_face.valid?
      raise UnexpectedStateError, "new_face is invalid." unless new_face.valid?

      [true, false].each do |is_front|
        material = is_front ? old_face.material : old_face.back_material
        next if material.nil?

        if material.texture.nil?
          if is_front
            new_face.material = material
          else
            new_face.back_material = material
          end
        else
          texture = material.texture
          ref_pt = new_face.vertices.first.position
          face_trans = Geom::Transformation.new(ref_pt, new_face.normal)

          # It doesn't seem to matter what the size of the rectangle is. Just ensure it forms a valid rectangle.
          # Not sure if texture values can be nil or zero
          width = texture.image_width
          width = 1.0 if width.nil? || width < 1.0
          height = texture.image_height
          height = 1.0 if height.nil? || height < 1.0

          face_pts = [
            ref_pt,
            ref_pt.offset(face_trans.xaxis, width),
            ref_pt.offset(face_trans.xaxis, width).offset(face_trans.yaxis, height),
            ref_pt.offset(face_trans.yaxis, height),
          ]

          uv_helper = old_face.get_UVHelper(is_front, !is_front)
          uv_mapping_array = []
          face_pts.each do |pt|
            uv_mapping_array << pt
            if is_front
              uvq_pt = uv_helper.get_front_UVQ(pt)
            else
              uvq_pt = uv_helper.get_back_UVQ(pt)
            end
            # Need to normalize the uvq to uv coordinates:
            # https://ruby.sketchup.com/Sketchup/UVHelper.html#get_front_UVQ-instance_method
            uv_arr = uvq_pt.to_a # Faster to use array with floats, rather than a Point3d with Lengths.
            uv_pt = Geom::Point3d.new(uv_arr[0] / uv_arr[2], uv_arr[1] / uv_arr[2], 0.0)
            uv_mapping_array << uv_pt
          end

          new_face.position_material(material, uv_mapping_array, is_front)
        end
      end
    end
  end


end