module DM::Bevel

  class FaceLoops
    include GeometryMixin

    # @return [Sketchup::Face]
    attr_reader :face
    # @return [Hash {Sketchup::Loop => Array<LoopVertex>}]
    attr_reader :loop_to_loop_vertices
    # @return [Hash {LoopVertex => Array<Geom::Point3d>}]
    attr_reader :loop_vertex_to_override_pts

    # @param face [Sketchup::Face]
    def initialize(face)
      loop_to_loop_vertices = Hash.new {|h,k| h[k] = []}
      face.loops.each do |loop|
        loop.vertices.each do |vert|
          loop_to_loop_vertices[loop] << LoopVertex.new(loop, vert)
        end
      end

      loop_to_loop_vertices.each_value do |loop_vertices|
        loop_vertices.length.times do |v_i|
          prev_loop_vertex = loop_vertices[v_i - 1]
          cur_loop_vertex = loop_vertices[v_i]
          next_loop_vertex = loop_vertices[v_i + 1]
          next_loop_vertex = loop_vertices[0] if next_loop_vertex.nil?
          cur_loop_vertex.set_adjacent_loop_vertices(prev_loop_vertex, next_loop_vertex)
        end
      end

      @face = face
      @loop_to_loop_vertices = loop_to_loop_vertices
      @loop_vertex_to_override_pts = Hash.new {|h,k| h[k] = []}
    end

    # @return [Boolean]
    def is_override?
      self.loop_vertex_to_override_pts.each_pair do |_, override_pts|
        return true unless override_pts.empty?
      end
      false
    end

    # @param loop_vertex [LoopVertex]
    # @param override_pt_or_pts [Geom::Point3d, Array<Geom::Point3d>]
    def set_override_pts(loop_vertex, override_pt_or_pts)
      raise UnexpectedTypeError.new("loop_vertex must be of type LoopVertex", self) unless loop_vertex.instance_of?(LoopVertex)
      override_pt_or_pts = [override_pt_or_pts] unless override_pt_or_pts.instance_of?(Array)
      override_pt_or_pts.each { |pt| raise UnexpectedTypeError, "override_pt must be of type Geom::Point3d" unless pt.instance_of?(Geom::Point3d) }

      # The order of points is determined outside of this function as there are different rules on how they should be sorted.
      cur_override_pts = @loop_vertex_to_override_pts[loop_vertex]
      override_pt_or_pts.each do |pt|
        cur_override_pts << pt unless cur_override_pts.include?(pt)
      end
    end

    # @param loop_vertex [LoopVertex]
    def clear_override_pts(loop_vertex)
      raise UnexpectedTypeError.new("loop_vertex must be of type LoopVertex", self) unless loop_vertex.instance_of?(LoopVertex)
      raise UnexpectedStateError("loop_vertex does not belong to this face_loops") unless @loop_vertex_to_override_pts.key?(loop_vertex)

      @loop_vertex_to_override_pts[loop_vertex].clear
    end

    # If there is any moment where the inner loop does not contains a coplanar face, that is a 'hole'.
    # @param loop [Sketchup::Loop]
    # @return [Boolean]
    def is_loop_a_hole?(loop)
      raise UnexpectedStateError.new("loop does not belong to this FaceLoops", self, loop) if self.loop_to_loop_vertices[loop].nil?

      return false if loop.outer?

      loop.edges.each do |edge|
        other_faces = (edge.faces - [self.face])
        return true if other_faces.empty?

        other_faces.each do |other_face|
          return true unless self.face.normal.parallel?(other_face.normal)
        end
      end
      false
    end

    # @param face_corner_or_edge_vertex [FaceCorner, EdgeVertex]
    # @return [LoopVertex]
    def get_loop_vertex(face_corner_or_edge_vertex)
      raise UnexpectedTypeError, "Must be of type FaceCorner or EdgeVertex" if !face_corner_or_edge_vertex.is_a?(FaceCorner) && !face_corner_or_edge_vertex.instance_of?(EdgeVertex)

      vertex = face_corner_or_edge_vertex.vertex
      self.loop_to_loop_vertices.each_pair do |loop, loop_vertices|
        next unless loop.vertices.include?(vertex)

        loop_vertex = loop_vertices.find do |loop_vertex|
          if face_corner_or_edge_vertex.is_a?(FaceCorner)
            loop_vertex.vertex == vertex && (loop_vertex.edges & face_corner_or_edge_vertex.edges).length == 2
          else
            loop_vertex.vertex == vertex
          end
        end

        return loop_vertex
      end
      nil
    end

    # @return [Hash{Sketchup::Loop => Array<Geom::Point3d>}]
    def get_loop_to_new_face_pts
      loop_to_new_face_pts = Hash.new {|h,k| h[k] = []}
      self.loop_to_loop_vertices.each_pair do |loop, loop_vertices|
        loop_vertices.each_with_index do |loop_vertex, i|
          override_pts = self.loop_vertex_to_override_pts[loop_vertex]
          override_pts = [loop_vertex.position] if override_pts.empty?
          loop_to_new_face_pts[loop].concat(override_pts)
        end
      end
      loop_to_new_face_pts
    end

    # @return [Hash{Sketchup::Loop => Array<Sketchup::Edge>}]
    def get_loop_to_affected_edges
      loop_to_affected_edges = Hash.new {|h,k| h[k] = []}
      self.loop_to_loop_vertices.each_pair do |loop, loop_vertices|
        loop_vertices.length.times do |lv_i|
          cur_loop_vertex = loop_vertices[lv_i]
          next_loop_vertex = loop_vertices[lv_i + 1]
          next_loop_vertex = loop_vertices[0] if next_loop_vertex.nil?

          cur_override_pts = self.loop_vertex_to_override_pts[cur_loop_vertex]
          next_override_pts = self.loop_vertex_to_override_pts[next_loop_vertex]

          # If both are empty, then this edge did not change.
          next if cur_override_pts.empty? && next_override_pts.empty?

          # one of the vertices changed, so add the edge.
          loop_to_affected_edges[loop] << cur_loop_vertex.next_edge
        end
      end
      loop_to_affected_edges
    end


    def inspect
      self.to_s
    end

  end


end