module DM::Bevel

  class LoopVertex

    # @return [Sketchup::Loop]
    attr_reader :loop
    # @return [Sketchup::Vertex]
    attr_reader :vertex
    # @return [LoopVertex]
    attr_reader :next_loop_vertex
    # @return [LoopVertex]
    attr_reader :prev_loop_vertex

    # @return [Geom::Point3d]
    def position
      val = @cache_manager.retrieve(:position) do
        self.vertex.position
      end
      val
    end

    # @return [Sketchup::Vertex]
    def prev_vertex
      self.prev_loop_vertex.vertex
    end

    # @return [Sketchup::Vertex]
    def next_vertex
      self.next_loop_vertex.vertex
    end

    # @return [Sketchup::Edge]
    def prev_edge
      val = @cache_manager.retrieve(:prev_edge) do
        self.vertex.common_edge(self.prev_vertex)
      end
      val
    end

    # @return [Sketchup::Edge]
    def next_edge
      val = @cache_manager.retrieve(:next_edge) do
        self.vertex.common_edge(self.next_vertex)
      end
      val
    end

    # @return [Array<Geom::Point3d, Geom::Vector3d>]
    def next_line
      val = @cache_manager.retrieve(:next_line) do
        [self.position, self.position.vector_to(self.next_vertex.position).normalize]
      end
      val
    end

    # @return [Array<Geom::Point3d, Geom::Vector3d>]
    def prev_line
      val = @cache_manager.retrieve(:prev_line) do
        [self.position, self.position.vector_to(self.prev_vertex.position).normalize]
      end
      val
    end

    # @return [Float]
    def next_length
      val = @cache_manager.retrieve(:next_length) do
        self.position.distance(self.next_vertex.position).to_f
      end
      val
    end

    # @return [Float]
    def prev_length
      val = @cache_manager.retrieve(:prev_length) do
        self.position.distance(self.prev_vertex.position).to_f
      end
      val
    end

    # @param corner [Corner]
    # @return [EdgeVertex]
    def next_edge_vert(corner)
      raise UnexpectedStateError.new("corner not at the same vertex", self, corner) unless corner.vertex == self.vertex

      corner.edge_verts.find {|edge_vert| edge_vert.edge == self.next_edge}
    end

    # @param corner [Corner]
    # @return [EdgeVertex]
    def prev_edge_vert(corner)
      raise UnexpectedStateError.new("corner not at the same vertex", self, corner) unless corner.vertex == self.vertex

      corner.edge_verts.find {|edge_vert| edge_vert.edge == self.prev_edge}
    end

    # @return [Boolean]
    def parallel?
      self.next_line[1].parallel?(self.prev_line[1])
    end

    # @param loop [Sketchup::Loop]
    # @param vert [Sketchup::Vertex]
    def initialize(loop, vert)
      raise UnexpectedStateError.new("vert does not belong to loop", loop, vert) unless loop.vertices.include?(vert)

      @loop = loop
      @vertex = vert

      @cache_manager = CacheManager.new
    end

    # @return [Array<Sketchup::Edge>]
    def edges
      [self.prev_edge, self.next_edge]
    end

    # @param other_loop_vertex [LoopVertex]
    # @return [Sketchup::Edge, Nil]
    def get_shared_edge(other_loop_vertex)
      res = self.edges & other_loop_vertex.edges
      res.first
    end

    # @param prev_loop_vertex [LoopVertex]
    # @param next_loop_vertex [LoopVertex]
    def set_adjacent_loop_vertices(prev_loop_vertex, next_loop_vertex)
      raise UnexpectedStateError, "next_vert does not belong to loop" unless self.loop.vertices.include?(next_loop_vertex.vertex)
      raise UnexpectedStateError, "prev_vert does not belong to loop" unless self.loop.vertices.include?(prev_loop_vertex.vertex)
      raise UnexpectedStateError, "vert same as prev_vert" if self.vertex == prev_loop_vertex.vertex
      raise UnexpectedStateError, "vert same as next_vert" if self.vertex == next_loop_vertex.vertex
      raise UnexpectedStateError, "next_vert same as prev_vert" if next_loop_vertex.vertex == prev_loop_vertex.vertex
      raise UnexpectedStateError, "vert is not adjacent to next_vert" if self.vertex.common_edge(next_loop_vertex.vertex).nil?
      raise UnexpectedStateError, "vert is not adjacent to prev_vert" if self.vertex.common_edge(prev_loop_vertex.vertex).nil?

      @next_loop_vertex = next_loop_vertex
      @prev_loop_vertex = prev_loop_vertex
    end

    # @param line [Array<Geom::Point3d, Geom::Vector3d>]
    # @param pt [Geom::Point3d]
    # @return [Boolean]
    def point_in_same_direction?(line, pt)
      return false unless pt.on_line?(line)
      return true if pt == line[0]

      line[1].samedirection?(self.position.vector_to(pt))
    end

    # @param pts [Array<Geom::Point3d>]
    # @return [Boolean]
    def are_start_end_ordered?(pts)
      raise UnexpectedStateError.new('', self, pts) if pts.length < 2

      start_pt = pts.first
      end_pt = pts.last

      start_on_prev_line = self.point_in_same_direction?(self.prev_line, start_pt)
      end_on_prev_line = self.point_in_same_direction?(self.prev_line, end_pt)
      start_on_next_line = self.point_in_same_direction?(self.next_line, start_pt)
      end_on_next_line = self.point_in_same_direction?(self.next_line, end_pt)

      # next and prev edge are parallel cases.
      if start_on_prev_line && start_on_next_line
        if !end_on_prev_line && !end_on_next_line
          # Assume it's in order.
          return true
        end

      elsif end_on_prev_line && end_on_next_line
        if !start_on_prev_line && !start_on_next_line
          # Assume it's ordered if the start_pt isn't on a lnie.
          return true
        end
      end

      if start_on_prev_line
        if end_on_prev_line
          return start_pt.distance(self.vertex.position) > end_pt.distance(self.vertex.position)
        elsif end_on_next_line
          return true
        else
          return true # TODO: Assume true. Does this make sense?
        end
      elsif start_on_next_line
        if end_on_prev_line
          return false
        elsif end_on_next_line
          return start_pt.distance(self.vertex.position) < end_pt.distance(self.vertex.position)
        else
          return false # TODO: Assume false. Does this make sense?
        end
      elsif end_on_prev_line
        return false
      elsif end_on_next_line
        return true
      else
        # TODO: Nothing is touching the lines,
        # DB.draw_loop_vertex(self)
        # puts 'loop_vertex.are_start_end_ordered? nothing is touching lines?'
        # raise # TODO: Logic chekc
      end

      false
    end


    # @param other_loop_vertex [LoopVertex]
    # @return [Boolean]
    def ==(other_loop_vertex)
      return false unless other_loop_vertex.instance_of?(LoopVertex)
      return false unless @vertex == other_loop_vertex.vertex
      return false unless @loop == other_loop_vertex.loop
      true
    end

    alias eql? :==

    def hash
      @vertex.hash + @loop.hash + self.next_vertex.hash + self.prev_vertex.hash
    end

    def inspect
      self.to_s
    end

  end

end