module DM::Bevel
  module GeometryMixin

    # pts is an array of point3d objects
    # @param pts [Array<Geom::Point3d>]
    # @return [Geom::Point3d]
    def centroid(pts)
      avg_pt = [0.0, 0.0, 0.0]
      pts.each do |p|
        p_arr = p.to_a
        avg_pt[0] += p_arr[0]
        avg_pt[1] += p_arr[1]
        avg_pt[2] += p_arr[2]
      end
      num = pts.length.to_f
      Geom::Point3d.new((avg_pt[0] / num), (avg_pt[1] / num), (avg_pt[2] / num))
    end

    # A wrapper method for Geom.fit_plane_to_points to return a Point and Vector, rather than 4 float values.
    # @param points [Geom::Point3d>]
    # @return [Array<Geom::Point3d, Geom::Vector3d>]
    def get_plane_from_points(points)
      abcd = Geom.fit_plane_to_points(points)
      plane_normal = Geom::Vector3d.new(abcd[0], abcd[1], abcd[2]).normalize
      point_on_plane = Geom::Point3d.new(0, 0, 0).project_to_plane(abcd)
      plane = [point_on_plane, plane_normal]
      plane
    end

    # Determines if the points are planar or not
    # must be sent an array of 3 or more points
    # @param points [Array<Geom::Point3d>]
    # @return [Boolean]
    def points_planar?(points)
      return false if points.length < 3
      return true if points.length == 3

      plane = Geom.fit_plane_to_points(points)
      points.each{ |p| return false unless p.on_plane?(plane) }
      true
    end

    # True if the edge makes 2 or more faces and all faces are coplanar
    # @param edge [Sketchup::Edge]
    # @return [Boolean]
    def coplanar_edge?(edge)
      # https://forums.sketchup.com/t/check-if-a-face-is-situated-on-a-specific-face/133859/9
      faces = edge.faces
      return false unless faces.length == 2

      coplanar_faces?(faces[0], faces[1])
    end

    # @param face1 [Sketchup::Face]
    # @param face2 [Sketchup::Face]
    # @return [Boolean]
    def coplanar_faces?(face1, face2)
      return true if face1 == face2 # Somehow, it's possible for edges.faces to return have two faces that are the same.

      pts = []
      (face1.vertices + face2.vertices).uniq.each do |vert|
        pts << vert.position.to_a
      end
      face_plane = Geom.fit_plane_to_points(pts)
      pts.each do |pt|
        return false unless pt.distance(pt.project_to_plane(face_plane)).round(3).zero?
      end
      true
    end

    # @param points [Array<Geom::Point3d>]
    # @param plane [Array<Geom::Point3d, Geom::Vector3d>]
    # @return [Boolean]
    def points_on_plane?(points, plane)
      points.each do |pt|
        return false unless pt.on_plane?(plane)
      end
      true
    end

    # True if the point is on the line specified by the start_pt and end_pt
    # @param start_pt [Geom::Point3d, Array<Float>]
    # @param end_pt [Geom::Point3d, Array<Float>]
    # @param pt [Geom::Point3d, Array<Float>]
    # @return [Boolean]
    def point_on_line?(start_pt, end_pt, pt)
      vec =  start_pt.vector_to(end_pt)
      line = [start_pt, vec]
      return false unless pt.on_line?(line)

      dist = start_pt.distance(pt) + end_pt.distance(pt)
      vec.length == dist
    end

    # vector_to will return a Geom::Vector3d even if you use arrays. This method avoids that.
    # @param start_arr [Array<Float>]
    # @param end_arr [Array<Float>]
    # @return [Array<Float>]
    def vector_to_arr(start_arr, end_arr)
      [end_arr[0] - start_arr[0], end_arr[1] - start_arr[1], end_arr[2] - start_arr[2]]
    end

    # For some reason, array.distance still makes calls to Length. This method avoids that.
    # @param arr1 [Array<Float>]
    # @param arr2 [Array<Float>]
    # @return [Float]
    def distance_arr(arr1, arr2)
      v = (arr2[0] - arr1[0])**2 + (arr2[1] - arr1[1])**2 + (arr2[2] - arr1[2])**2
      Math.sqrt(v)
    end

    # @param vec_arr [Array<Float>]
    # @return [Float]
    def vec_length(vec_arr)
      v = vec_arr[0]**2 + vec_arr[1]**2 + vec_arr[2]**2
      Math.sqrt(v)
    end

    # Compare the values in the array and return the unique ones. FYI, array.uniq will do a hash comparision. This will force an == comparison.
    # @param array [Array]
    # @return [Array]
    def unique_array(array)
      return [] if array.empty?
      return [array.first] if array.length == 1

      indices_to_ignore = []
      uniq_objs = []
      array.each_with_index do |obj, i|
        next if indices_to_ignore.include?(i)

        uniq_objs << obj
        ((i + 1)..array.length - 1).each do |next_i|
          next_obj = array[next_i]
          if obj == next_obj
            indices_to_ignore << next_i
          end
        end
      end
      uniq_objs
    end

    # @param face [Sketchup::Face]
    # @param point [Geom::Point3d]
    # @param include_edge [Boolean]
    # @return [Boolean]
    def point_on_face?(face, point, include_edge = true)
      invalid_ids = Sketchup::Face::PointOutside | Sketchup::Face::PointNotOnPlane
      invalid_ids |= Sketchup::Face::PointOnEdge unless include_edge
      (face.classify_point(point) & invalid_ids) == 0
    end


    # @param face [Sketchup::Face]
    # @param point [Geom::Point3d]
    # @param include_edge [Boolean]
    # @return [Boolean]
    def points_on_face?(face, points, include_edge = true)
      all_pts_on_face = true
      points.each do |pt|
        next if point_on_face?(face, pt, include_edge)

        all_pts_on_face = false
        break
      end
      all_pts_on_face
    end

    # @param line1 [Array<Geom::Point3d, Geom::Vector3d>]
    # @param line2 [Array<Geom::Point3d, Geom::Vector3d>]
    # @param default_pt_if_parallel [Geom::Point3d]
    # @return [Geom::Point3d, Nil]
    def intersect_line_line(line1, line2, default_pt_if_parallel = nil)
      int_pt = Geom.intersect_line_line(line1, line2)
      int_pt.nil? ? default_pt_if_parallel : int_pt
    end

    # @param start_pt [Geom::Point3d]
    # @param end_pt [Geom::Point3d]
    # @param num_pts [Integer]
    # @return [Array<Geom::Point3d>]
    def fill_points_between(start_pt, end_pt, num_pts)
      # Below looks weird, but we need to maintain the references of the start/end Geom::Point3d objects.
      vec = start_pt.vector_to(end_pt)
      new_pts = [start_pt]
      (num_pts - 1).times do |s_i|
        new_pts << (start_pt.offset(vec, vec.length * ((s_i + 1).to_f / num_pts.to_f)))
      end
      new_pts << end_pt
      new_pts
    end


    # https://stackoverflow.com/questions/451426/how-do-i-calculate-the-area-of-a-2d-polygon
    # @param face_pts [Array<Geom::Point3d>]
    # @param normal [Geom::Vector3d]
    # @return [Boolean]
    def pts_form_valid_face?(face_pts, normal = nil)
      raise UnexpectedStateError.new("Must have at least 3 points", face_pts) if face_pts.length < 3

      normal = get_plane_from_points(face_pts).last if normal.nil?

      sum = 0.0
      (face_pts.length - 1).times do |pt_i|
        pt0 = face_pts[pt_i - 1]
        pt1 = face_pts[pt_i]
        pt2 = face_pts[pt_i + 1]

        vec1 = pt0.vector_to(pt1)
        next unless vec1.valid?

        vec2 = pt1.vector_to(pt2)
        next unless vec2.valid?

        cross_vec = vec1.cross(vec2)
        next unless cross_vec.valid?

        amt = cross_vec.length.to_f
        amt *= -1.0 unless cross_vec.samedirection?(normal)

        sum += amt
      end

      !sum.round(3).zero?
    end

    # A wrapper to Geom.linear_combination
    # @param pt_or_vec1 [Geom::Point3d, Geom::Vector3d]
    # @param pt_or_vec2 [Geom::Point3d, Geom::Vector3d]
    # @param weight [Float]
    # @return [Geom::Point3d, Geom::Vector3d]
    def linear_combination(pt_or_vec1, pt_or_vec2, weight = 0.5)
      raise UnexpectedTypeError, "pt_or_vec1 must be an instance of Point3d or Vector3d" if !pt_or_vec1.instance_of?(Geom::Point3d) && !pt_or_vec1.instance_of?(Geom::Vector3d)
      raise UnexpectedTypeError, "pt_or_vec2 must be an instance of Point3d or Vector3d" if !pt_or_vec2.instance_of?(Geom::Point3d) && !pt_or_vec2.instance_of?(Geom::Vector3d)
      raise UnexpectedStateError, "weight must be in the range of 0 to 1.0" if weight < 0.0 || weight > 1.0

      weight2 = (1.0 - weight).abs
      Geom.linear_combination(weight, pt_or_vec1, weight2, pt_or_vec2)
    end

    # @param vec [Geom::Vector3d]
    # @param other_vec [Geom::Vector3d]
    # @return [Boolean]
    def similar_direction?(vec, other_vec)
      ang = vec.angle_between(other_vec)
      ang <= Math::PI / 2.0
    end

    # https://forums.sketchup.com/t/ruby-angle-between-issue/104322
    # Counter-clockwise angle from vector2 to vector1, as seen from normal.
    # @param vector1 [Geom::Vector3d]
    # @param vector2 [Geom::Vector3d]
    # @param normal [Geom::Vector3d]
    # @return [Float]
    def angle_in_plane(vector1, vector2, normal = Z_AXIS)
      Math.atan2((vector2 * vector1) % normal, vector1 % vector2)
    end

    # @param plane [Array<Geom::Point3d, Geom::Vector3d>]
    # @param pt [Geom::Point3d]
    # @return [Boolean]
    def point_behind?(plane, pt)
      return false if pt.on_plane?(plane)

      proj_pt = pt.project_to_plane(plane)
      vec = pt.vector_to(proj_pt).normalize

      plane[1].samedirection?(vec)
    end

    # The perpendicular vector that always points 'into' the face.
    # @param edge [Sketchup::Edge]
    # @param face [Sketchup::Face]
    # @return [Geom::Vector3d]
    def perpendicular_vector(edge, face)
      edge_vec = edge.start.position.vector_to(edge.end.position).normalize
      perp_vec = face.normal * edge_vec
      # Grab the mid_pt of the edge. If we use the start/end pts, it's possible perp vec would offset the pt off the face (due to the angle it forms at that point.)
      mid_pt = Geom.linear_combination(0.5, edge.start.position, 0.5, edge.end.position)
      mid_pt.offset!(perp_vec, 0.01)
      perp_vec.reverse! unless point_on_face?(face, mid_pt)
      perp_vec
    end

    # https://forums.sketchup.com/t/orienting-faces-and-the-api/83762/11 (At the very bottom, comment by CAUL)
    # @param edge [Sketchup::Edge]
    # @return [Boolean]
    def reversed_faces?(edge)
      faces = edge.faces
      edge.reversed_in?(faces[0]) == edge.reversed_in?(faces[1])
    end


  end
end