module DM::Bevel

  class ObserverManager

    def self.model_observer
      @model_observer
    end

    def self.add_model_observer(model)
      if @model_observer
        model.remove_observer(@model_observer) # remove it if it was added previously
      end
      @model_observer = BevelModelObserver.new
      model.add_observer(@model_observer)
    end

  end

  # This observer watches the model for changes (transactions) and updates the
  # bevel proxy subsurface whenever a change is detected.
  class BevelModelObserver < Sketchup::ModelObserver

    def initialize
      reset
    end

    def reset
      @transaction_committed = false
      @transaction_started = false
      @needs_update = true
      # Keeps track of the active entities that is being edited at the time of the last txn commit
      # Allows us to detect if user has switched the active editing path
      @last_txn_active_entities = Sketchup.active_model.active_entities
      @proxy_edit_mode_active = false
      @proxy_did_commit_change = false
      @new_proxy_transaction = false
    end

    def needs_update=(value)
      @needs_update = value
    end

    # This method is used to notify the observer that a new proxy was just created which
    # can affect the flow of the logic when that transaction eventually gets committed
    def new_proxy_transaction
      @new_proxy_transaction = true
    end

    # These two methods allow the proxy settings to test whether
    # a change has been made to the proxy
    def start_watching_proxy_changes
      @proxy_did_commit_change = false
    end

    def proxy_did_change?
      @proxy_did_commit_change
    end

    def onTransactionStart(model)
      @transaction_started = true
    end

    def onTransactionCommit(model)
      @transaction_committed = true
      proxy = ProxySettings.active_proxy
      if model.active_entities != @last_txn_active_entities
        # Settings this immediately after ensures the code below will only execute once per transaction
        @last_txn_active_entities = model.active_entities
        # Using the outliner, it is possible to jump from inside one
        # proxy to another. Below, we make sure that the previous
        # proxy mode is cancelled in this situation as well
        # as the standard double-clicking to enter an object
        cancel_edit_proxy_mode(model) if @proxy_edit_mode_active
        if proxy
          # This means that the user closed Component edit mode or opened edit mode
          begin_edit_proxy_mode(proxy, model)
        end
      else
        # Active entities has not changed for this transaction so some edit was made to
        # the active entities
        if proxy
          if proxy != @proxy_settings.coarse_mesh_instance # Used outliner to change to a new instance of the same component
            @proxy_settings.init_proxy(proxy)
            refresh_proxy_refs(proxy)
          else
            # User is currently editing a proxy
            if @transaction_started && @transaction_committed
              @proxy_did_commit_change = true
              if @needs_update
                # No need to update if user is only viewing proxy only
                if @proxy_settings.show_bevel_surf != ProxySettings::ShowBevelsurf::PROXY
                  update_bevel_surface
                end
              end
            end
          end
        end
      end
    end

    def onTransactionUndo(model)
      on_undo_redo(model)
    end

    def onTransactionRedo(model)
      on_undo_redo(model)
    end

    def on_undo_redo(model)
      proxy = ProxySettings.active_proxy
      if model.active_entities != @last_txn_active_entities
        @last_txn_active_entities = model.active_entities
        if proxy
          setup_bevel_editing_variables(proxy)
          @proxy_settings.set_rendering_options
        else
          # If proxy edit mode was active prior to hitting undo but is not anymore
          if @proxy_edit_mode_active
            @proxy_settings.disable_edge_coloring
            cancel_bevel_editing_variables
          end
        end
      else
        if proxy
          if @proxy_settings.valid_bevel_surf?
            # In this case, a proxy was and is still being edited so we need to refresh
            refresh_proxy_refs(proxy)
          else
            # Here, the bevel proxy has become an invalid entity which could happen if the
            # user performs an undo immediately after committing a subsurface
            setup_bevel_editing_variables(proxy)
            @proxy_settings.set_rendering_options
          end
        end
      end
    end

    def begin_edit_proxy_mode(inst, model)
      # Combine this operation with the 'open' operation
      @needs_update = false # we don't want to trigger an update of the subsurface in this case
      model.start_operation(LH["Show proxy"], true, false, true)
      inst.make_unique if inst.is_a?(Sketchup::Group)
      setup_bevel_editing_variables(inst)
      @proxy_settings.get_bevel_surf_instance.make_unique # needed to prevent copied groups referencing the same subsurf
      @proxy_settings.set_rendering_options
      model.commit_operation
      @needs_update = true
    end

    # Setup key references and show the settings dialog
    def setup_bevel_editing_variables(proxy_inst)
      @proxy_settings = ProxySettings.instance
      @proxy_settings.init_proxy(proxy_inst)
      if @new_proxy_transaction
        # If a proxy was just created during this transaction, the proxySettings.active_bevel
        # object has already been setup properly (inside the Subsmooth method) so there is no need
        # to refresh the proxy references
        @new_proxy_transaction = false
      else
        refresh_proxy_refs(proxy_inst)
      end
      @proxy_settings.show
      @proxy_edit_mode_active = true
    end

    def disable_edge_coloring
      # Restore the user's previously used rendering options
      ro = Sketchup.active_model.rendering_options
      @stored_rendering_options.each do |key, val|
        ro[key] = val if val
      end
    end

    # Regenerate the proxy surface
    def refresh_proxy_refs(proxy)
      beveler = BevelNonDestructive.new(proxy)
      bevel_surf_inst = @proxy_settings.get_bevel_surf_instance
      beveler.bevel_surface_entities = bevel_surf_inst.definition.entities
      ProxySettings.active_bevel = beveler
      @proxy_settings.refresh_dialog(false) # Update the proxy settings dialog (needed for Undo / Redo)
    end

    # Remove references related to proxy editing
    # Nothing in this method should modify the model and potentially create an undoable operation
    def cancel_bevel_editing_variables
      @proxy_edit_mode_active = false
      ProxySettings.active_bevel = nil
      ProxySettings.instance.close
    end

    def cancel_edit_proxy_mode(model)
      # The operation below should be transparent and appended to the Close operation
      if @proxy_settings
        model.start_operation(LH["Hide Proxy"], true, false, true)
        @proxy_settings.make_proxysurf_visible(false)
        @proxy_settings.make_proxy_invisible
        @proxy_settings.restore_proxy_material
        @proxy_settings.disable_edge_coloring
        Sketchup.active_model.commit_operation
        # If the visibility was set to proxy only, it is possible that the subsurf will need to be updated
        # at this stage
        if @proxy_settings.show_bevel_surf == ProxySettings::ShowBevelsurf::PROXY && @proxy_did_commit_change
          update_bevel_surface
        end
      end
      cancel_bevel_editing_variables
    end

    def update_bevel_surface
      if ProxySettings.active_bevel
        @needs_update = false
        model = Sketchup.active_model
        model.start_operation(LH["Update Bevel"], true, false, true)
        ProxySettings.active_bevel.update_bevelsurf
        # the commit below will trigger this observer to fire another onTransactionCommit
        # event, but since the @needs_update variable is set to false above, the
        # update_subsurface method should only fire once
        model.commit_operation
        # Now, we set the @needs_update variable to true to prepare for the next potential
        # transaction. This logic assumes that the update_subsurface commit is the last
        # commit performed of the chained commits
        @needs_update = true
      end
    end

  end

  class AppObserver < Sketchup::AppObserver

    def initialize
      @has_started_observers = {}
      # This ensures observers are started when this extension is initially loaded and not just when SU is opened
      # or a new model is opened
      init_observers(Sketchup.active_model)
    end

    def expectsStartupModelNotifications
      true
    end

    def onOpenModel(m)
      init_observers(m)
    end

    def onNewModel(m)
      init_observers(m)
    end

    def onQuit
    end

    def init_observers(m)
      if @has_started_observers[m] && ON_MAC
        # do nothing
      else
        ProxySettings.instance.close if ProxySettings.instance.visible?
        ObserverManager.add_model_observer(m)
        @has_started_observers[m] = true
      end
    end

  end

  Sketchup.add_observer(AppObserver.new)
end