module DM::Bevel

  class ProxyMesh

    def initialize
      @polys = [] # array of ProxyPolys that make up the mesh
      @osd_vert_indices = {} # Hash of vertex indices indexed by vertex
      @osd_vert_positions = {} # Hash of 3D points in the mesh indexed by OSD vertex index
      @su_faces_with_osd_index = [] # array of SU faces where the index of the array matches a corresponding OSD face index
      @adj_polys = {}
      @dimension = []
      @face_meshes = []
      @mesh = Geom::PolygonMesh.new
      @osd_front_uvs = [] # array of front face UVs
      @osd_back_uvs = [] # array of back face UVs
      @adj_faces = [] # keep track of adjacent faces that should also be subdivided
      @front_uvs = []
      @back_uvs = []
      @trans = IDENTITY
      @num_polys = 0
      @num_indices = 0
      @creased_edges = []
      @creased_verts = []
      @has_materials = false # flag for whether this mesh has any materials applied to the faces
    end

    def transformation=(trans)
      @trans = trans
    end

    def transformation
      @trans
    end

    def is_mirrored?
      xa = @trans.xaxis
      ya = @trans.yaxis
      za = @trans.zaxis
      if xa.cross(ya).dot(za) < 0.0
        true
      else
        false
      end
    end

    def has_materials?
      @has_materials
    end

    # Adds a convex polygon to the mesh
    # @param [Sketchup::Face or Quadface] face
    # @param [Array<Sketchup::Vertex or Geom::Point3d>] verts
    # @param [Array<Geom::Point3d>] front_uvs
    # @param [Array<Geom::Point3d>] back_uvs
    def add_poly(face, verts, front_uvs = nil, back_uvs = nil)
      indices = add_verts_to_mesh(verts)
      @mesh.add_polygon(indices)
      new_poly = ProxyPoly.new(face)
      front_uv_indices = []
      back_uv_indices = []
      verts.each_index do |i|
        front_uv_indices << add_uv(front_uvs[i]) if front_uvs
        back_uv_indices << add_back_uv(back_uvs[i]) if back_uvs
      end
      new_poly.indices = indices
      new_poly.front_uv_indices = front_uv_indices
      new_poly.back_uv_indices = back_uv_indices
      @num_polys += 1
      @num_indices += verts.length
      @polys << new_poly
      @su_faces_with_osd_index << face
      if face.material || face.back_material
        @has_materials = true
      end
    end

    # @param [Sketchup::Face] face the coarse mesh parent face for this poly
    # @param [Array<Integer>] vert_indices the Proxymesh vertex indices for this polygon
    # @param [Array<Integer>] front_uv_indices
    def add_poly2(face, vert_indices, front_uv_indices)
      new_poly = ProxyPoly.new(face)
      new_poly.indices = vert_indices
      new_poly.front_uv_indices = front_uv_indices
      @polys << new_poly
      @num_polys += 1
      @num_indices += vert_indices.length
      if face.material || face.back_material
        @has_materials = true
      end
      begin
        add_polygon(vert_indices)
      rescue ArgumentError
        puts "Error adding polygon to mesh - index out of range"
      end
    end

    def add_creased_edge(edge, verts)
      indices = add_verts_to_mesh(verts)
      @creased_edges << CreasedEdge.new(edge, indices, BevelNonDestructive.get_crease_weight(edge))
    end

    def add_creased_vertex(vertex)
      indices = add_verts_to_mesh([vertex])
      @creased_verts << CreasedVertex.new(vertex, indices.first, BevelNonDestructive.get_crease_weight(vertex))
    end

    def add_verts_to_mesh(verts)
      indices = []
      verts.each do |v|
        if v.is_a?(Sketchup::Vertex)
          index = add_point(v.position.transform(@trans))
          @osd_vert_indices[v] = index - 1
        else
          index = add_point(v.transform(@trans))
        end
        indices << index
      end
      indices
    end

    def add_adj_poly(face, points, front_uvs = [], back_uvs = [])
      new_poly = ProxyPoly.new(face)
      indices = []
      points.each { |p| indices.push(self.add_point(p)) }
      new_poly.indices = indices
      new_poly.front_uvs = front_uvs
      new_poly.back_uvs = back_uvs
      @adj_polys[face] = [] if @adj_polys[face] == nil
      @adj_polys[face].push(new_poly)
    end

    # Updates the weights of the creased edges of this mesh without changing anything else
    # Used by the crease tool to perform live updates of the Proxysurf
    def update_creased_edges(edges)
      edges.each do |edge|
        weight = BevelNonDestructive.get_crease_weight(edge)
        # Check if this edge is already included in this mesh as a creased edge
        c_edge = @creased_edges.find { |ce| ce.edge == edge }
        if c_edge
          if weight != 0.0
            c_edge.weight = weight
          else
            @creased_edges.delete(c_edge)
          end
        else
          # Add a new creased edge to the mesh
          i0 = @mesh.point_index(edge.start.position.transform(@trans))
          i1 = @mesh.point_index(edge.end.position.transform(@trans))
          if i0 != 0 && i1 != 0
            # Both points should always be in the mesh
            @creased_edges << CreasedEdge.new(edge, [i0, i1], weight)
          end
        end
      end
    end

    # Updates the weights of the creased edges of this mesh without changing anything else
    # Used by the crease tool to perform live updates of the subsurf
    def update_creased_vertices(verts)
      verts.each do |v|
        weight = BevelNonDestructive.get_crease_weight(v)
        # Check if this vertex is already included in this mesh as a creased vertex
        c_vertex = @creased_verts.find { |cv| cv.vertex == v }
        if c_vertex
          if weight != 0.0
            c_vertex.weight = weight
          else
            @creased_verts.delete(c_vertex)
          end
        else
          # Add a new creased vertex to the mesh
          i0 = @mesh.point_index(v.position.transform(@trans))
          if i0 != 0
            # Point should always be in the mesh
            @creased_verts << CreasedVertex.new(v, i0, weight)
          end
        end
      end
    end

    def count_polys
      @num_polys
    end

    def count_face_indices
      @num_indices
    end

    def polygons
      @polys
    end

    def su_face_for_osd_index(index)
      @su_faces_with_osd_index[index]
    end

    def adj_faces
      return @adj_faces
    end

    def vertices
      @osd_vert_indices.keys
    end

    # Allows the  class to update the hash of vertices that correspond to osd indices
    def osd_vert_indices=(val)
      @osd_vert_indices = val
    end

    def osd_vert_indices
      @osd_vert_indices
    end

    def osd_vert_positions
      @osd_vert_positions
    end

    def creased_edges
      @creased_edges
    end

    def creased_verts
      @creased_verts
    end

    def face_meshes
      return @face_meshes
    end

    def add_point(p)
      return @mesh.add_point(p)
    end

    def add_uv(p)
      @front_uvs << p
      return @front_uvs.length - 1
    end

    def add_back_uv(p)
      @back_uvs << p
      return @back_uvs.length - 1
    end

    def count_uvs
      return @front_uvs.length
    end

    def front_uvs
      @front_uvs.dup
    end

    def front_uv_at(i)
      @front_uvs[i]
    end

    def back_uvs
      @back_uvs.dup
    end

    def back_uv_at(i)
      @back_uvs[i]
    end

    def polygon_front_uvs(poly)
      poly.front_uv_indices.collect { |i| front_uv_at(i) }
    end

    def polygon_back_uvs(poly)
      poly.back_uv_indices.collect { |i| back_uv_at(i) }
    end

    def add_polygon(p)
      return @mesh.add_polygon(p)
    end

    def count_points
      return @mesh.count_points
    end

    def count_edges
      @creased_edges.length
    end

    def count_corners
      @creased_verts.length
    end

    def point_at(i)
      return @mesh.point_at(i)
    end

    def point_index(p)
      return @mesh.point_index(p)
    end

    def points
      return @mesh.points
    end

    def set_point(i, p)
      return @mesh.set_point(i, p)
    end

    def mesh
      return @mesh
    end

    def add_face_mesh(face)
      #@face_meshes.push(face)
      fm = face.mesh
      polys = fm.polygons
      for poly in polys
        self.add_poly(face, [fm.point_at(poly[0]), fm.point_at(poly[1]), fm.point_at(poly[2])])
      end
    end

    # This method iterates through a ProxyMesh and adds the actual polygons to it so faces
    # can be generated from it in the model. Since mesh polygons are only needed to be added immediately
    # prior to using the ProxyMesh to create actual SketchUp faces, a separate function is provided here
    # to allow optimization rather than forcing polygons to be generated within the 'add_poly' method
    def generate_polygons
      mesh = self

      for poly in mesh.polygons
        num_verts = poly.length

        case num_verts
        when 3
          p0 = poly[0]
          p1 = poly[1]
          p2 = poly[2]
          mesh.add_polygon([p0, p1, p2])
        when 4
          p0 = mesh.point_at(poly[0])
          p1 = mesh.point_at(poly[1])
          p2 = mesh.point_at(poly[2])
          p3 = mesh.point_at(poly[3])

          if Common.quad_points_planar?([p0, p1, p2, p3])
            mesh.add_polygon(poly.indices)
          else
            mesh.add_polygon([poly[0], -poly[1], poly[3]])
            mesh.add_polygon([poly[1], poly[2], -poly[3]])
          end
        end
      end
      mesh
    end


    def calc_dimension
      dim = Hash.new(2)

      @polys.each { |face, polygons|
        #p "getting dimension for #{face}"
        polygons.each { |poly|
          if poly.length == 4
            dim[poly[0]] = [dim[poly[0]], 2].min
            dim[poly[1]] = [dim[poly[1]], 2].min
            dim[poly[2]] = [dim[poly[2]], 2].min
            dim[poly[3]] = [dim[poly[3]], 2].min
          elsif poly.length == 3
            dim[poly[0]] = [dim[poly[0]], 2].min
            dim[poly[1]] = [dim[poly[1]], 2].min
            dim[poly[2]] = [dim[poly[2]], 2].min
          elsif poly.length == 2 #we have an edge
            dim[poly[0]] = [dim[poly[0]], 1].min
            dim[poly[1]] = [dim[poly[1]], 1].min
          elsif poly.length == 1 #we have a vertex
            dim[poly[0]] = 0
          end

        }
      }

      dim
    end

  end #class end

  #####
  class ProxyPoly

    attr_reader :points, :indices, :front_uv_indices, :back_uv_indices, :face, :front_material, :back_material
    attr_writer :points, :indices, :front_uv_indices, :back_uv_indices

    def initialize(face)
      @face = face
      @points = []
      @indices = []
      @front_uv_indices = []
      @back_uv_indices = []
      @front_material = face.material
      @back_material = face.back_material if face.respond_to?(:back_material)
    end

    def length
      @indices.length
    end

    def [](index)
      @indices[index]
    end

    def front_material
      @front_material
    end

    def back_material
      @back_material
    end

    def front_uvs
      raise LH["Not Implemented Yet"]
    end

    def back_uvs
      raise LH["Not Implemented Yet"]
    end

  end #end class

  class CreasedEdge

    attr_reader :indices, :edge, :weight
    attr_writer :indices, :weight

    # @param [Sketchup::Edge] edge
    # @param [Array<Integer>] indices
    # @param [Numeric] weight
    def initialize(edge, indices, weight)
      @edge = edge
      @indices = indices
      @weight = weight
    end
  end

  class CreasedVertex

    attr_reader :index, :vertex, :weight
    attr_writer :index, :weight

    # @param [Sketchup::Vertex] vertex
    # @param [Array<Integer>] index
    # @param [Numeric] weight
    def initialize(vertex, index, weight)
      @vertex = vertex
      @index = index
      @weight = weight
    end

  end
end