
module DM::Bevel

  class BaseTool < CoreTool
    include GeometryMixin
    include BevelMixin
    include DrawMixin

    # @return [BevelParams]
    def bevel_params
      @bevel_params
    end

    def initialize
      # @type [BevelParams]
      @bevel_params = load_params
      # @type [Array<BevelData>]
      @bevel_datas = []
      # @type [Array<Sketchup::Edge>]
      @hovered_edges = []
      super
    end

    def load_params
      BevelParams.new
    end

    def reset
      @keys_down.clear
      @hovered_edges.clear
      @bevel_datas.clear
      super
    end

    def activate
      CoreTool.active_tool = self
      on_preselection
      set_status_text
      set_vcb
    rescue InvalidLicenseException
      Licensing.on_invalid_license_found
      Sketchup.active_model.select_tool(nil)
    end

    # @param view [Sketchup::View]
    # @param path_select_mode [Boolean]
    # @return [Array<Sketchup::Edge>]
    def get_edges_below_cursor(view, x, y, path_select_mode = false)
      ip = view.inputpoint(x, y)
      edges = []
      source = InvalidEdgeError::Source::EDGE
      if path_select_mode
        if ip.edge && (draw_hidden_geometry? || !ip.edge.soft?) # If hidden geometry is off, then don't allow selecting it.
          edge = collect_valid_edges([ip.edge]).first # This will raise an exception if the hovered edge is invalid.
          edges = collect_valid_edges(get_smart_path(edge))
        elsif ip.vertex
          source = InvalidEdgeError::Source::VERTEX
          pickray = view.pickray(x, y)
          edge_to_dist = {}
          _invalid_edges = []
          ip.vertex.edges.each do |edge|
            if bevelable_edge?(edge)
              edge_to_dist[edge] = Geom.closest_points(edge.line, pickray).first.distance(pickray[0]).to_f
            else
              _invalid_edges << edge
            end
          end

          if edge_to_dist.empty?
            collect_valid_edges(_invalid_edges) # This will raise an exception if the hovered edge is invalid.
          else
            closest_edge = edge_to_dist.min {|a, b| a[1] <=> b[1] }[0]
            edges = collect_valid_edges(get_smart_path(closest_edge))
          end
        end
      else
        if ip.vertex
          source = InvalidEdgeError::Source::VERTEX
          edges = collect_valid_edges([ip.vertex])
        elsif ip.edge && (draw_hidden_geometry? || !ip.edge.soft?) # If hidden geometry is off, then don't allow selecting it.
          edges = collect_valid_edges([ip.edge])
        elsif ip.face
          if !draw_hidden_geometry? && surface_face?(ip.face)
            source = InvalidEdgeError::Source::SURFACE
            # Can only select a surface if Hidden Geometry is off
            edges = collect_valid_edges(get_surface_face_edges(ip.face))
            collect_valid_edges([ip.face]) if edges.empty? # If there's nothing, then the entire thing is smooth. Show a warning this way.
          else
            source = InvalidEdgeError::Source::FACE
            edges = collect_valid_edges([ip.face])
          end
        else
          # No valid selection
        end
      end

      edges
    rescue InvalidEdgeError => err 
      err.source = source
      raise
    end

    # @param face [Sketchup::Face]
    # @return [Boolean]
    def surface_face?(face)
      face.edges.each do |edge|
        next unless edge.soft?
        next unless edge.faces.length == 2

        return true
      end
      false
    end

    # @param start_face [Sketchup::Face]
    # @return [Array<Sketchup::Edge>]
    def get_surface_face_edges(start_face)
      _checked_ents = {}
      valid_edges = []

      valid_surface_faces = [start_face]
      valid_surface_faces.each do |face|
        next unless _checked_ents[face].nil?

        _checked_ents[face] = true

        face.edges.each do |edge|
          next unless _checked_ents[edge].nil?

          _checked_ents[edge] = true
          if !edge.soft?
            valid_edges << edge
          else
            # its a soft edge so lets keep checking.
            edge.faces.each do |other_face|
              next unless _checked_ents[other_face].nil?

              valid_surface_faces << other_face
            end
          end
        end
      end
      valid_edges
    end

    # @return [Boolean]
    def draw_hidden_geometry?
      if Sketchup.version.to_i >= 20
        Sketchup.active_model.rendering_options["DrawHiddenGeometry"]
      else
        Sketchup.active_model.rendering_options["DrawHidden"]
      end
    end

    # @param start_edge [Sketchup::Edge]
    # @return [Array<Sketchup::Edge>]
    def get_smart_path(start_edge)
      return [] if start_edge.nil?

      # TODO: Will probably be configurable
      angle_threshold = 60.0.degrees.round(PLUGIN::FLOAT_PRECISION)

      _checked_ents = {}
      valid_edges = [start_edge]
      valid_edges.each do |edge|
        next unless _checked_ents[edge].nil?

        _checked_ents[edge] = true
        edge.vertices.each do |vert|
          next unless _checked_ents[vert].nil?

          _checked_ents[vert] = true

          edge_vec = edge.other_vertex(vert).position.vector_to(vert.position).normalize
          _possible_edge_to_angle = {}
          vert.edges.each do |other_edge|
            next unless _checked_ents[other_edge].nil?
            next unless bevelable_edge?(other_edge)

            other_edge_vec = vert.position.vector_to(other_edge.other_vertex(vert).position).normalize
            angle = edge_vec.angle_between(other_edge_vec).to_f.round(PLUGIN::FLOAT_PRECISION)
            next if angle > angle_threshold

            _possible_edge_to_angle[other_edge] = angle
          end
          next if _possible_edge_to_angle.empty?

          valid_edges << _possible_edge_to_angle.min {|a,b| a[1] <=> b[1] }[0]
        end
      end
      valid_edges
    end

    # @param edge [Sketchup::Edge]
    def bevelable_edge?(edge)
      validate_edges(edge) == BevelableEdgeState::VALID
    end

    # @param entities [Enumerable<Sketchup::Entity>]
    # @return [Array<Sketchup::Edge>]
    def collect_valid_edges(entities)
      entities = entities.to_a unless entities.is_a?(Array)
      active_entities = Sketchup.active_model.active_entities

      overall_edge_state = BevelableEdgeState::VALID

      # @type [Array<Sketchup::Edge>]
      edges = []
      entities.each do |ent|
        next unless ent.valid?
        next unless ent.parent.entities == active_entities # Only select edges/faces that are at the active_entities level

        if ent.instance_of?(Sketchup::Edge)
          next if edges.include?(ent)

          quad_face = QuadFace.get_from_diagonal(ent)
          if !quad_face.nil?
            entities.concat(quad_face.edges)
          elsif ent.curve.nil?
            edges << ent
          else
            edges += ent.curve.edges
          end
        elsif ent.instance_of?(Sketchup::Face) && QuadFace.is_non_planar_quadface?(ent)
          quad_face = QuadFace.get(ent)
          entities.concat(quad_face.edges)
        elsif ent.respond_to?(:edges)
          ent.edges.each do |edge|
            next if edges.include?(edge)

            if edge.curve.nil?
              edges << edge
            else
              edges += edge.curve.edges
            end
          end
        end
      end
      valid_edges = []

      overall_edge_state |= validate_edges(edges.uniq, valid_edges)

      # Only raise if there are no valid edges
      raise InvalidEdgeError.new(overall_edge_state) if valid_edges.empty? && overall_edge_state != BevelableEdgeState::VALID

      valid_edges
    end

    # @param invalid_edge_error [InvalidEdgeError]
    # @return [Array<String>]
    def get_invalid_edge_tooltips(invalid_edge_error)
      source_str = "     " # padding (doesn't seem to help on Mac)
      case invalid_edge_error.source
      when InvalidEdgeError::Source::FACE
        source_str+= LH["Invalid Face Edges:"]
      when InvalidEdgeError::Source::VERTEX
        source_str+= LH["Invalid Vertex Edges:"]
      when InvalidEdgeError::Source::SURFACE
        source_str+= LH["Invalid Surface Edges:"]
      else
        source_str+= LH["Invalid Edge(s):"]
      end

      msgs = []
      msgs << LH["Must be at active entities."] if invalid_edge_error.contains?(BevelableEdgeState::NOT_AT_ACTIVE_ENTS) # Unused as if Jul 10, 2023, but leaving for now
      msgs << source_str + ' ' + LH["Edge must have exactly two faces"] if invalid_edge_error.contains?(BevelableEdgeState::NOT_TWO_FACES)
      msgs << source_str + ' ' + LH["Faces are coplanar"] if invalid_edge_error.contains?(BevelableEdgeState::COPLANAR)
      msgs << source_str + ' ' + LH["Hidden"] if invalid_edge_error.contains?(BevelableEdgeState::HIDDEN)
      msgs << source_str + ' ' + LH["Soft"] if invalid_edge_error.contains?(BevelableEdgeState::SOFT)
      msgs << source_str + ' ' + LH["Smooth"] if invalid_edge_error.contains?(BevelableEdgeState::SMOOTH)
      msgs << source_str + ' ' + LH["Reversed faces"] if invalid_edge_error.contains?(BevelableEdgeState::REVERSED_FACES)
      msgs << source_str + ' ' + LH["Quad edge"] if invalid_edge_error.contains?(BevelableEdgeState::QUAD_EDGE)
      msgs
    end

    # Returns the platform specific keyboard key name
    # @param [String] win_name the name of the key used for Windows
    # @param [String] suffix optional text after the key name
    # @param [String] mac_name optional name of the key used for Mac (to over-ride the default behavior)
    def rep_key(win_name, suffix = nil, mac_name = nil)
      key = win_name.dup
      if ON_MAC
        key.gsub!("Ctrl", "Option")
        key.gsub!("CTRL", "OPTION")
        key.gsub!("Alt", "Command")
        key.gsub!("ALT", "COMMAND")
        key.replace(mac_name) if mac_name
      end
      if suffix
        return key + " #{suffix} "
      else
        return key
      end
    end

  end

end