
module DM::Bevel

  class CoreTool
    DELIMITER = Sketchup::RegionalSettings.decimal_separator == ',' ? ';' : ',' # Use ; if , is used as decimal_separator

    # Keep track of the active tool so we can use the cmd validation proc to display
    # that the tool is in an active state (needed for SU2017 and SU2018 support). In newer
    # SketchUp versions, we could get the active tool from active_model.tools.active_tool
    @@active_tool = {}
    @@last_active_tool = {}
    def self.active_tool=(value)
      model = Sketchup.active_model
      @@last_active_tool[model] = @@active_tool[model] unless @@active_tool[model].nil?
      @@active_tool[model] = value
    end

    # @return [BaseTool]
    def self.active_tool
      @@active_tool[Sketchup.active_model]
    end

    def self.last_active_tool
      @@last_active_tool[Sketchup.active_model]
    end

    def initialize
      @keys_down = []
      @x = @y = 0
      # Store the mouse position that the user clicked and released the left mouse button
      @xdown = 0
      @ydown = 0
      # Store state of left button so we can determine if user is click-dragging
      @left_button_down = false
      @dragging = false
      reset
    end

    def reset
      @keys_down.clear
      @xdown = 0
      @ydown = 0
      @left_button_down = false
      @dragging = false
    end

    def activate
      # Implement in other class
      CoreTool.active_tool = self
      set_vcb
      set_status_text
    end

    def on_preselection
      # Implement in other class
    end

    def deactivate(view)
      CoreTool.active_tool = nil
      view.lock_inference
      view.invalidate
    end

    def resume(view)
      set_vcb
      set_status_text
      view.invalidate
    end

    def suspend(view)
      set_vcb
      set_status_text
      view.invalidate
    end

    def set_status_text
      Sketchup::set_status_text("set_status_text Not implemented")
    end

    def set_vcb
      Sketchup.set_status_text("vcb not implemented", SB_VCB_LABEL)
    end

    def getExtents
      bb = Sketchup.active_model.bounds
      bb
    end

    # @param view [Sketchup::View]
    def onCancel(reason, view)
      reset
      set_status_text
      set_vcb
      view.lock_inference
      view.invalidate
    end

    # @param view [Sketchup::View]
    def onMouseMove(flags, x, y, view)
      @x = x
      @y = y
      @dragging = is_dragging?(x, y) unless @dragging # Only update the dragging state if user is not currently dragging
    end

    def onLButtonDown(flags, x, y, view)
      @left_button_down = true
      # Impossible to be dragging immediately after clicking the mouse
      # Dragging is only possible once button is clicked but mouse is moved
      @dragging = false
      @xdown = x
      @ydown = y
    end

    def onLButtonUp(flags, x, y, view)
      @left_button_down = false
      # Impossible to be dragging immediately after mouse button is released
      @dragging = false
    end

    def is_dragging?(x, y)
      return false unless @left_button_down
      pixel_dist = 10
      (x - @xdown).abs > pixel_dist || (y - @ydown).abs > pixel_dist
    end

    # Get the length a user enters in the VCB
    # @return [Length or nil]
    def get_length_from_vcb(text)
      begin
        value = PLUGIN.parse_length(text)
      rescue
        # Error parsing the text
        UI.beep
        value = nil
        Sketchup::set_status_text("", SB_VCB_VALUE)
      end
      value
    end

    def get_cursor(name, x, y)
      ext = ON_MAC ? ".pdf" : ".svg"
      icon = File.join(PATH, "icons", "#{name + ext}")
      UI.create_cursor(icon, x, y)
    end

    def onSetCursor
      set_cursor
    end

    def set_cursor
      # Intentionally blank - default behavior is to not provide any cursor
      # Sub-classes can over-write this method to set to an actual custom cursor
    end

  end

end