
module DM::Bevel

  #
  class EdgeExtruder
    include DrawMixin
    include GeometryMixin

    # @return [Geom::Point3d]
    attr_reader :extrude_edge_ref_point
    # @return [Sketchup::Edge]
    attr_reader :extrude_edge

    def initialize
      @extrude_edge_ref_point = nil
      @extrude_edge = nil
      @offset_vector = nil
    end

    # @return [Boolean]
    def valid?
      !@extrude_edge_ref_point.nil?
    end

    def clear
      @extrude_edge_ref_point = nil
      @extrude_edge = nil
      @offset_vector = nil
    end

    # @param view [Sketchup::View]
    # @param should_draw_offset_line [Boolean]
    def draw(view, should_draw_offset_line)
      return unless self.valid?

      view.draw_points(@extrude_edge_ref_point, 10, 2, 'white')
      view.draw_points(@extrude_edge_ref_point, 9, 2, 'red')

      if should_draw_offset_line
        mouse_on_line = @extrude_edge_ref_point + @offset_vector
        view.drawing_color = 'black'
        draw_2d_line(view, [@extrude_edge_ref_point, mouse_on_line], size: 1, stipple: '-')
      end
    end

    # Get the offset length in regards to offset line and view line
    # Determine the distances between the pick ray and offset line
    # @param view [Sketchup::View]
    # @return [Float]
    def get_offset_length(x, y, view)
      ray = view.pickray(x, y)
      offset_line = [@extrude_edge_ref_point, @offset_vector]
      pick_line = ray
      closest = Geom.closest_points(offset_line, pick_line)
      dist = @extrude_edge_ref_point.distance(closest[0])
      dist = 0.001 unless @offset_vector.samedirection?(@extrude_edge_ref_point.vector_to(closest[0])) # don't allow positive offset if going in the wrong direction
      @offset_vector.length = dist
      dist.to_f
    end

    # Returns a 3 element array if successful containing the point hit, entity hit, and container object hit
    # @param [Array<Geom::Point3d, Geom::Vector3d>] pickray
    # @return [Array<Geom::Point3d, Sketchup::Entity, Sketchup::ComponentInstance]
    def get_point_below_pickray(pickray, in_entities = Sketchup.active_model.active_entities)
      rt = Sketchup.active_model.raytest(pickray)
      while rt != nil
        if rt
          pos = rt[0]
          path = rt[1]
          ent = path.pop
          if ent.parent.entities == in_entities
            # Send back the position, face that was hit, and the component instance hit
            return [pos, ent, rt[1][-2]]
          end

          path.each do |group| # Check path for expected entities, group or component
            if group.definition.entities == in_entities
              # Send back the position, face that was hit, and the component instance hit
              return [pos, ent, rt[1][-2]]
            end
          end

          # recast using last position
          pickray = [pos, pickray[1]]
          rt = Sketchup.active_model.raytest(pickray)
        end
      end
      nil
    end

    # Get the point on the closest selected edge which will be used as the reference point
    # @param edges [Array<Sketchup::Edge>]
    # @param view [Sketchup::View]
    # @return [Array<Sketchup::Edge, Geom::Point3d>]
    #       - Index 0: The closest Edge
    #       - Index 1: The point on the Edge.
    def set_edge_reference_point(edges, x, y, view)
      mouse_pt = get_point_below_pickray(view.pickray(x, y))
      mouse_pt = mouse_pt.nil? ? view.guess_target : mouse_pt[0]

      distances = {}
      valid_edges = Set.new
      # Determine the distances between the pick ray and the edge lines
      edges.each do |edge|
        faces = edge.faces
        faces.each do |face|
          edge_line = edge.line
          if edge.reversed_in?(face)
            edge_line[1] = edge_line[1].reverse
          end
          pick_line = [mouse_pt, face.normal * edge_line[1]]
          closest = Geom.intersect_line_line(edge_line, pick_line)
          next if closest.nil?

          valid_edges.add(edge) if point_on_line?(edge.start.position, edge.end.position, closest) # Priority to closest[0] that is on the edge
          distance = mouse_pt.distance(closest)
          distances[edge] = { dist: distance, point: closest, vector: pick_line[1] } if distances[edge].nil? || distances[edge][:dist] > distance
        end
      end
      valid_distances = {}
      valid_edges.each do |edge|
        valid_distances[edge] = distances[edge]
      end
      distances = valid_distances unless valid_edges.empty?
      min_distance = distances.values.collect { |hash| hash[:dist] }.min
      # If any reference edges are parallel, the distances to the pickray will be identical so we need to now filter
      # the edges further using the centroid of the edge.  Pick the edge where the screen-distance between the
      # edge centroid and the mouse position is minimum.
      closest_edges = edges.find_all { |edge| distances[edge] && distances[edge][:dist] == min_distance }
      edge_centers = closest_edges.collect { |edge| centroid(edge.vertices.collect{ |v| v.position }) }
      screen_distances = edge_centers.collect { |p| view.screen_coords(p).distance(Geom::Point3d.new(x, y, 0)) }
      min_screen_distance = screen_distances.min
      min_index = screen_distances.index(min_screen_distance)
      return nil if min_index.nil?
      edge = closest_edges[min_index]
      # Get the closest point on this edge line
      closest = distances[edge][:point]
      @offset_vector = distances[edge][:vector]

      # Constrain point to be between the start and end points of the edge
      v0 = closest.vector_to(edge.end.position)
      v1 = closest.vector_to(edge.start.position)
      edge_vec = edge.line[1]
      if v0.dot(edge_vec) < 0.0
        closest = edge.end.position
      elsif v1.dot(edge_vec) > 0.0
        closest = edge.start.position
      end
      @extrude_edge_ref_point = closest
      @extrude_edge = edge

      [edge, closest]
    end

  end


end