module DM::Bevel

  class Window < UI::HtmlDialog
    DEFAULT_MATERIAL_NAME = LH["Default"]

    HTML_ROOT = File.join(PATH, "html")

    def initialize(options)
      super(options)
      # For windows that should not be resizable we make sure to set the size
      # after creating the dialog. Otherwise it might be remembering old values
      # and use that instead.
      set_size(options[:width], options[:height]) unless options[:resizable]

      @events = {}
      @state = {}
    end

    # Default way to load the html -- This can be overwritten for special cases
    def html_set
      set_file(File.join(HTML_ROOT, @dialog_filename))
    end

    def on(event, &block)
      return false if block.nil?

      @events[event] ||= []
      @events[event] << block
      true
    end

    # Platform neutral method that ensures that window stays on top of the main
    # window on both platforms. Also captures any blocks given and executes it
    # when the HTML DOM is ready.
    def show(&block)
      html_set
      add_callbacks unless visible?
      super { block.call if block }
    end

    # Note, since are using a Singleton approach for windows, we need to
    # re-attach the callbacks each time the HtmlDialog opens
    def add_callbacks
      add_action_callback("callback") { |_, callback_name|
        data = []
        trigger_events(callback_name, data)
      }

      add_action_callback("ready") { |_|
        localize
        #disable_context_menu # TODO: ensure context menus are disabled or overwritten
        refresh_dialog if self.respond_to?(:refresh_dialog)
        # After the dialog is displayed, we want to ensure that the viewport
        # remains in focus so that shortcut keys for commands will continue
        # to trigger even after the dialog pops up
        focus_sketchup
      }

      add_action_callback("showMessageBox") { |_, message|
        UI.messagebox(message)
      }

      add_action_callback("close") { |_|
        self.close
      }

      add_action_callback('ValueChanged') do |_, p|
        # p comes in as string in the format of "id|value"
        # Find the first | and split it from there. A value could have multiple |.
        pipe_idx = p.index('|')
        id = p[0..(pipe_idx - 1)]
        value = p[(pipe_idx + 1)..-1]
        id += '=' unless id.end_with?("=")
        on_value_changed(id, value)
      end

      add_action_callback("messageboxTooltip") { |_, p|
        # there is a bug in SU2017 HtmlDialogs on Windows where the tooltip will
        # not display on hover so instead, we display it with a messagebox when
        # the user clicks a label in the settings dialog
        if !ON_MAC && Sketchup.version.to_i == 17
          UI.messagebox(p)
        end
      }

      self.add_action_callback('button_clicked') do |_, id|
        on_button_clicked(id)
      end

      self.add_action_callback('focus') do |_, id|
        focus_sketchup(true)
      end
    end

    def on_value_changed(id, value)
      self.send(id, value)
    end

    def on_button_clicked(id)
      self.send(id)
    end

    # Over-ride this method in child classes
    def on_change_value(id, value)
      raise "on_change_value Not implemented"
    end

    # noinspection RubyResolve
    private

    def trigger_events(event, data = [])
      if @events[event]
        @events[event].each{ |callback|
          callback.call(self, data)
        }
        true
      else
        false
      end
    end

    def html_special_chars(text)
      val = ''
      if text
        val = text.dup
        val.gsub!(/'/, '&#39;')
        val.gsub!(/"/, '&#34;')
        val.gsub!(/\n/, '\\\\n')
      end
      val
    end

    def center_dlg_coordinates(width, height)
      view = Sketchup.active_model.active_view
      vp_width = view.vpwidth
      vp_height = view.vpheight
      left = (vp_width / 2.0) - (width / 2.0)
      top = (vp_height / 2.0) - (height / 2.0)
      pt = Geom::Point3d.new(left, top, 0)
      pt
    end

    def add_slashes(text)
      val = ''
      if text
        val = text.dup
        val.gsub!(/\\/, '\\\\\\')
        val.gsub!(/\n/, '\\\\n')
        val.gsub!(/"/, '\\\"')
      end
      val
    end

    def decode_param(p)
      p.gsub!(/PB2apos;/, "'")
      p.gsub!(/‚Äô/, "'")  #fix for an encoding issue on OSX that I cannot figure out
    end

    def set_checkbox(id, checked)
      checked = checked.to_s
      checked = 'false' if checked == ''
      cmd = "document.getElementById('#{id}').checked=#{checked}"
      execute_script(cmd)
      cmd = "document.getElementById('#{id}').value=#{checked}"
      execute_script(cmd)
    end

    def set_length(id, val)
      length = Sketchup.format_length(val)
      set_value(id, length)
    end

    # Sets the value of an input field in the web dialog
    def set_value(id, value, dlg = self)
      new_value = add_slashes(value.to_s)
      cmd = "document.getElementById('#{id}').value=\"#{new_value}\""
      dlg.execute_script(cmd)
    end

    def set_text(id, text)
      text = add_slashes(text)
      cmd = "document.getElementById('#{id}').innerHTML=\"#{text}\""
      execute_script(cmd)
    end

    def set_attribute(id, value, attr, dlg = self)
      new_value = add_slashes(value.to_s)
      cmd = "document.getElementById('#{id}').#{attr}=\"#{new_value}\""
      dlg.execute_script(cmd)
    end

    def set_select_widget(id, value)
      set_value(id, value)
      cmd = "$('##{id}').selectmenu( 'refresh' );"
      execute_script(cmd)
    end

    def set_multi_select(id, values)
      cmd = "setMultiSelect('#{id}',#{values});"
      execute_script(cmd)
    end

    def set_radio(name, value)
      cmd = "set_radio('#{name}','#{value}');"
      execute_script(cmd)
    end

    # gets an elements value and formats it as a length
    def format_length_to_float(id)
      length =  get_element_value(id)
      float = PLUGIN.parse_length(length).to_f
      float
    end

    # gets an elements value and formats it as an integer
    def format_input_to_integer(id)
      input =  get_element_value(id)
      integer = input.to_i
      integer
    end

    # Get model length unit
    def get_model_length_unit
      uo = Sketchup.active_model.options['UnitsOptions']
      uo['LengthUnit']
    end

    def store_dlg_size(dlg_name, width, height)
      Sketchup.write_default(dlg_name, 'dlgwidth' ,width)
      Sketchup.write_default(dlg_name, 'dlgheight' ,height)
    end

    def get_dlg_size(dlg_name)
      width = Sketchup.read_default(dlg_name, 'dlgwidth')
      height = Sketchup.read_default(dlg_name, 'dlgheight')
      [width.to_i, height.to_i]
    end

    def localize
      call_js("localize", LH.strings)
    end

    def viewport_height(factor = 1.0)
      Sketchup.active_model.active_view.vpheight * factor
    end

    def viewport_width(factor = 1.0)
      Sketchup.active_model.active_view.vpwidth * factor
    end

    # Simplifies calling the JavaScript in the webdialog by taking care of
    # generating the JS command needed to execute it.
    #
    # @param [String] function
    # @param [Mixed] *arguments
    #
    # @return [Nil]
    def call_js(function, *arguments)
      # Calling inspect will ensure strings are quoted and convert most basic
      # Ruby data types to JS types. This is a very naive version and will need
      # tweaking if more types are needed.
      js_args = arguments.map { |x| x.to_json }.join(", ")
      javascript = "#{function}(#{js_args});"
      execute_script(javascript)
    end

    def refresh_materials_list(id, include_default_mat = true, include_empty = false)
      mats = Sketchup.active_model.materials
      clear_list(id)
      create_option(id, '') if include_empty
      material_names = mats.collect { |mat| mat.name }
      material_names = material_names.sort
      material_names.unshift(DEFAULT_MATERIAL_NAME) if include_default_mat
      material_names.each { |name|
        create_option(id, name)
      }
    end

    def refresh_layers_list(id, include_default_layer = true, include_empty = false)
      layers = Sketchup.active_model.layers
      clear_list(id)
      create_option(id, '') if include_empty
      layer_names = layers.collect { |layer| layer.name }
      layer_names = layer_names.sort
      layer_names.each do |name|
        val = name
        if name == layers[0].name
          next unless include_default_layer

          name = LH['Untagged']
        end
        create_option(id, name, val)
      end
    end

    def clear_list(id)
      cmd = "clearList(\"#{id}\")"
      execute_script(cmd)
    end

    def create_option(id, name, value = name)
      name = add_slashes(name)
      value = add_slashes(value)
      cmd = "createOptionAndValue(\"#{id}\",\"#{name}\",\"#{value}\")"
      execute_script(cmd)
    end

    # Brings the main SketchUp Window into focus.
    def focus_sketchup(mac_only = false)
      if mac_only
        return unless ON_MAC
      end
      if Sketchup.respond_to?(:focus)
        Sketchup.focus # Added in 2021.1
      else
        return if ON_MAC # hack below does not work on Mac
        # Create, show, and close a dialog to allow SketchUp to take focus back from @dlg (does not work on MAC)
        # reference: https://forums.sketchup.com/t/how-to-set-focus-to-mainwindow/12383/11
        # Using WebDialog since doing this hack with HtmlDialog can actually cause MAC to crash
        dlg = UI::WebDialog.new(dialog_title = "", scrollable = false, pref_key = nil, width = 0, height = 0, left = 0, top = 0, resizable = false)
        dlg.show; dlg.close # give focus back to SketchUp
      end
    end

  end # class
end